Package: mini007
Type: Package
Title: Lightweight Framework for Orchestrating Multi-Agent Large
        Language Models
Version: 0.4.0
Authors@R: 
    person(given = c("Mohamed", "El", "Fodil"),
           family = "Ihaddaden",
           role = c("aut", "cre"),
           email = "ihaddaden.fodeil@gmail.com")
Description: Provides tools for creating agents with persistent state using 
    R6 classes <https://cran.r-project.org/package=R6> and the 'ellmer' 
    package <https://cran.r-project.org/package=ellmer>. Tracks prompts, 
    messages, and agent metadata for reproducible, multi-turn large language 
    model sessions.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://github.com/feddelegrand7/mini007
BugReports: https://github.com/feddelegrand7/mini007/issues
Imports: checkmate (>= 2.3.1), cli (>= 3.6.5), DiagrammeR (>= 1.0.11),
        ellmer (>= 0.4.0), glue (>= 1.8.0), R6 (>= 2.6.1), rlang (>=
        1.1.6), uuid (>= 1.2.0)
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-05-27 13:26:58 UTC; mohamedelfodilihaddaden
Author: Mohamed El Fodil Ihaddaden [aut, cre]
Maintainer: Mohamed El Fodil Ihaddaden <ihaddaden.fodeil@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-27 13:50:02 UTC
Built: R 4.4.3; ; 2026-05-27 17:10:41 UTC; unix

Package: mini007
Type: Package
Title: Lightweight Framework for Orchestrating Multi-Agent Large
        Language Models
Version: 0.4.0
Authors@R: 
    person(given = c("Mohamed", "El", "Fodil"),
           family = "Ihaddaden",
           role = c("aut", "cre"),
           email = "ihaddaden.fodeil@gmail.com")
Description: Provides tools for creating agents with persistent state using 
    R6 classes <https://cran.r-project.org/package=R6> and the 'ellmer' 
    package <https://cran.r-project.org/package=ellmer>. Tracks prompts, 
    messages, and agent metadata for reproducible, multi-turn large language 
    model sessions.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://github.com/feddelegrand7/mini007
BugReports: https://github.com/feddelegrand7/mini007/issues
Imports: checkmate (>= 2.3.1), cli (>= 3.6.5), DiagrammeR (>= 1.0.11),
        ellmer (>= 0.4.0), glue (>= 1.8.0), R6 (>= 2.6.1), rlang (>=
        1.1.6), uuid (>= 1.2.0)
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-05-27 13:26:58 UTC; mohamedelfodilihaddaden
Author: Mohamed El Fodil Ihaddaden [aut, cre]
Maintainer: Mohamed El Fodil Ihaddaden <ihaddaden.fodeil@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-27 13:50:02 UTC
Built: R 4.5.2; ; 2026-05-27 16:23:02 UTC; unix

# Generated by roxygen2: do not edit by hand

export(Agent)
export(LeadAgent)
export(Workflow)
export(WorkflowAgent)
importFrom(DiagrammeR,grViz)
importFrom(R6,R6Class)
importFrom(checkmate,assert_character)
importFrom(checkmate,assert_environment)
importFrom(checkmate,assert_flag)
importFrom(checkmate,assert_integerish)
importFrom(checkmate,assert_string)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,cli_rule)
importFrom(cli,cli_text)
importFrom(cli,cli_ul)
importFrom(ellmer,type_number)
importFrom(ellmer,type_object)
importFrom(ellmer,type_string)
importFrom(glue,glue)
importFrom(rlang,caller_env)
importFrom(rlang,eval_bare)
importFrom(rlang,parse_expr)
importFrom(uuid,UUIDgenerate)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mini007 <a><img src='man/figures/mini007cute.png' align="right" height="200" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/mini007)](https://CRAN.R-project.org/package=mini007)
[![R
badge](https://img.shields.io/badge/Build%20with-♥%20and%20R-blue)](https://github.com/feddelegrand7/mini007)
[![metacran
downloads](https://cranlogs.r-pkg.org/badges/mini007)](https://cran.r-project.org/package=mini007)
[![metacran
downloads](https://cranlogs.r-pkg.org/badges/grand-total/mini007)](https://cran.r-project.org/package=mini007)

<!-- badges: end -->

`mini007` is a lightweight and extensible R framework for building
multi-agent AI systems. It lets you create specialised LLM-backed
agents, orchestrate them through a lead agent that decomposes and
delegates complex tasks, and wire them together into explicit sequential
pipelines called Workflows, all built on top of the excellent
[`ellmer`](https://ellmer.tidyverse.org/) package and compatible with
any chat model it supports.

## Key components

``` r
library(mini007)

retrieve_open_ai_credential <- function() {
  Sys.getenv("OPENAI_API_KEY")
}

llm <- ellmer::chat(
  name        = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential,
  echo        = "none"
)
```

### `Agent` a stateful LLM-backed worker

An `Agent` wraps an `ellmer` chat object and adds identity, persistent
message history, budget tracking, and tool support.

``` r
researcher <- Agent$new(
  name        = "researcher",
  instruction = "You are a research assistant. Answer factual questions concisely.",
  llm_object  = llm
)

researcher$invoke("What is the capital of Algeria?")
```

Agents remember the full conversation, so follow-up questions work
naturally:

``` r
researcher$invoke("And what is its population?")
```

Key capabilities:

- **Persistent memory**: full message history with `$messages`, trimming
  via `$keep_last_n_messages()`, export/import via
  `$export_messages_history()` / `$load_messages_history()`.
- **Context sharing**: pass recent messages from one agent to another
  with `$share_context_with(other_agent)`.
- **Budget control**: set a USD spending limit with `$set_budget()` and
  choose a policy (`"abort"`, `"warn"`, or `"ask"`) via
  `$set_budget_policy()`.
- **Tools**: register `ellmer` tool objects with `$register_tools()`,
  generate new ones from a natural-language description with
  `$generate_and_register_tool()`.
- **R code generation**: ask an agent to write, validate, and optionally
  execute R code with `$generate_execute_r_code()`.
- **Response validation**: verify an agent’s output against criteria
  using `$validate_response()`.
- **Cloning**: create an independent copy of an agent (new UUID, shared
  instruction) with `$clone_agent()`.

------------------------------------------------------------------------

### `LeadAgent` multi-agent orchestration

A `LeadAgent` extends `Agent` with the ability to decompose a complex
prompt into subtasks and automatically delegate each one to the most
suitable registered agent.

``` r
summariser <- Agent$new(
  name        = "summariser",
  instruction = "Summarise text into three bullet points.",
  llm_object  = llm
)

translator <- Agent$new(
  name        = "translator",
  instruction = "Translate text from English to German.",
  llm_object  = llm
)

lead <- LeadAgent$new(name = "Lead", llm_object = llm)
lead$register_agents(c(summariser, translator))

lead$invoke("Summarise the history of the Roman Empire, then translate it into German.")
```

Key capabilities:

- **Plan generation**: preview which agent handles which subtask before
  running with `$generate_plan()`.
- **Plan visualisation**: render the orchestration as a directed graph
  with `$visualize_plan()`.
- **Broadcasting**: send the same prompt to every registered agent and
  collect all responses with `$broadcast()`.
- **Judging**: let the lead agent pick the best response from all
  registered agents with `$judge_and_choose_best_response()`.
- **Agent dialog**: run an iterative negotiation between two agents
  until they reach consensus with `$agents_dialog()`.
- **Human In The Loop (HITL)**: pause execution at specific steps so a
  human can review, edit, or abort the response with `$set_hitl(steps)`.

------------------------------------------------------------------------

### `Workflow` explicit sequential pipelines

A `Workflow` lets you build a predefined pipeline of **Stations**
connected by **Routes**. Each Station’s output becomes the next one’s
input. Unlike `LeadAgent`, the execution path is fully explicit, you
control the order and branching logic.

``` r
wf <- Workflow$new(name = "article-pipeline")

wf$add_station("research", Agent$new(
  name        = "researcher",
  instruction = "Gather concise facts on the topic.",
  llm_object  = llm
))
wf$add_station("write", Agent$new(
  name        = "writer",
  instruction = "Turn the facts into an engaging paragraph.",
  llm_object  = llm
))
wf$add_station("edit", Agent$new(
  name        = "editor",
  instruction = "Polish the paragraph for grammar and clarity.",
  llm_object  = llm
))

wf$add_route("research", "write")
wf$add_route("write",    "edit")

wf$run("The history of the printing press")
```

Key capabilities:

- **Mixed handlers**: Stations accept `Agent` objects, `WorkflowAgent`
  objects, or plain R `function`s, making it easy to mix LLM calls with
  deterministic pre/post-processing steps.
- **Conditional routing**: `add_route(from, to, condition)` gates a
  route on a function of the previous Station’s output, enabling
  branching pipelines.
- **Caching**: with `use_cache = TRUE` (default), each Station’s result
  is stored by `(name, input)`. Re-running the same input serves the
  cache instantly. Use `$clear_cache()` to reset.
- **Wrapping as an agent**: `$as_agent()` converts any `Workflow` into a
  `WorkflowAgent` that can be registered with a `LeadAgent` or embedded
  as a Station inside another `Workflow`.
- **Visualisation**: `$visualize()` renders the pipeline as an
  interactive directed graph via `DiagrammeR`.

## Installation

You can install `mini007` from `CRAN` with:

``` r
install.packages("mini007")
```

The documentation is available
[here](https://feddelegrand7.github.io/mini007/)

## Code of Conduct

Please note that the mini007 project is released with a [Contributor
Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mini007 <a><img src='man/figures/mini007cute.png' align="right" height="200" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/mini007)](https://CRAN.R-project.org/package=mini007)
[![R
badge](https://img.shields.io/badge/Build%20with-♥%20and%20R-blue)](https://github.com/feddelegrand7/mini007)
[![metacran
downloads](https://cranlogs.r-pkg.org/badges/mini007)](https://cran.r-project.org/package=mini007)
[![metacran
downloads](https://cranlogs.r-pkg.org/badges/grand-total/mini007)](https://cran.r-project.org/package=mini007)

<!-- badges: end -->

`mini007` provides a lightweight and extensible framework for
multi-agents orchestration processes capable of decomposing complex
tasks and assigning them to specialized agents.

Each `agent` is an extension of an `ellmer` object. `mini007` relies
heavily on the excellent `ellmer` package but aims to make it easy to
create a process where multiple specialized agents help each other
sequentially in order to execute a task.

`mini007` provides two types of agents:

- A normal `Agent` containing a name and an instruction,
- and a `LeadAgent` which will take a complex prompt, split it, assign
  to the adequate agents and retrieve the response.

#### Highlights

🧠 Memory and identity for each agent via `uuid` and message history.

⚙️ Built-in task decomposition and delegation via `LLM`.

🔄 Agent-to-agent orchestration with result chaining.

🌐 Compatible with any chat model supported by `ellmer`.

🧑 Possibility to set a Human In The Loop (`HITL`) at various execution
steps

You can install `mini007` from `CRAN` with:

``` r
install.packages("mini007")
```

``` r
library(mini007)
```

# Agent

### Creating an Agent

An Agent is built upon an LLM object created by the `ellmer` package, in
the following examples, we’ll work with the `OpenAI` models, however you
can use any model/combination of models you want:

``` r
# no need to provide the system prompt, it will be set when creating the
# agent (see the 'instruction' parameter)

retrieve_open_ai_credential <- function() {
  Sys.getenv("OPENAI_API_KEY")
}

openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)
```

After initializing the `ellmer` LLM object, creating the Agent is
straightforward:

``` r
polar_bear_researcher <- Agent$new(
  name = "POLAR BEAR RESEARCHER",
  instruction = "You are an expert in polar bears, you task is to collect information about polar bears. Answer in 1 sentence max.",
  llm_object = openai_4_1_mini
)
```

Each created Agent has an `agent_id` (among other meta information):

``` r
polar_bear_researcher$agent_id
#> [1] "e0672404-1579-4a6a-8a69-068450dc9c59"
```

At any time, you can tweak the `llm_object`:

``` r
polar_bear_researcher$llm_object
#> <Chat OpenAI/gpt-4.1-mini turns=1 input=0 output=0 cost=$0.00>
#> ── system ──────────────────────────────────────────────────────────────────────
#> You are an expert in polar bears, you task is to collect information about polar bears. Answer in 1 sentence max.
```

An agent can provide the answer to a prompt using the `invoke` method:

``` r
polar_bear_researcher$invoke("Are polar bears dangerous for humans?")
#> Yes, polar bears can be dangerous to humans as they are powerful predators and 
#> may attack if threatened or hungry.
```

You can also retrieve a list that displays the history of the agent:

``` r
polar_bear_researcher$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are an expert in polar bears, you task is to collect information about polar bears. Answer in 1 sentence max."
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "Are polar bears dangerous for humans?"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "Yes, polar bears can be dangerous to humans as they are powerful predators and may attack if threatened or hungry."
```

Or the `ellmer` way:

``` r
polar_bear_researcher$llm_object
#> <Chat OpenAI/gpt-4.1-mini turns=3 input=43 output=23 cost=$0.00>
#> ── system ──────────────────────────────────────────────────────────────────────
#> You are an expert in polar bears, you task is to collect information about polar bears. Answer in 1 sentence max.
#> ── user ────────────────────────────────────────────────────────────────────────
#> Are polar bears dangerous for humans?
#> ── assistant [input=43 output=23 cost=$0.00] ───────────────────────────────────
#> Yes, polar bears can be dangerous to humans as they are powerful predators and may attack if threatened or hungry.
```

## Agents and Messages

#### Managing Agent Conversation History

The `clear_and_summarise_messages` method allows you to compress an
agent’s conversation history into a concise summary and clear the
message history while preserving context. This is useful for maintaining
memory efficiency while keeping important conversation context.

``` r
# After several interactions, summarise and clear the conversation history
polar_bear_researcher$clear_and_summarise_messages()
#> ✔ Conversation history summarised and appended to system prompt.
#> ℹ Summary: The user asked if polar bears are dangerous to humans, and the assistant, acting as a polar bear exp...
polar_bear_researcher$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are an expert in polar bears, you task is to collect information about polar bears. Answer in 1 sentence max. \n\n--- Conversation Summary ---\n The user asked if polar bears are dangerous to humans, and the assistant, acting as a polar bear expert, responded that polar bears can indeed be dangerous because they are powerful predators who may attack if threatened or hungry."
```

This method summarises all previous conversations into a paragraph and
appends it to the system prompt, then clears the conversation history.
The agent retains the context but with reduced memory usage.

#### Keep only the most recent messages with `keep_last_n_messages()`

When a conversation grows long, you can keep just the last N messages
while preserving the system prompt. This helps control token usage
without fully resetting context.

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

agent <- Agent$new(
  name = "history_manager",
  instruction = "You are a concise assistant.",
  llm_object = openai_4_1_mini
)

agent$invoke("What is the capital of Italy?")
#> The capital of Italy is Rome.
agent$invoke("What is the capital of Germany?")
#> The capital of Germany is Berlin.
agent$invoke("What is the capital of Algeria?")
#> The capital of Algeria is Algiers.
agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are a concise assistant."
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "What is the capital of Italy?"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "The capital of Italy is Rome."
#> 
#> 
#> [[4]]
#> [[4]]$role
#> [1] "user"
#> 
#> [[4]]$content
#> [1] "What is the capital of Germany?"
#> 
#> 
#> [[5]]
#> [[5]]$role
#> [1] "assistant"
#> 
#> [[5]]$content
#> [1] "The capital of Germany is Berlin."
#> 
#> 
#> [[6]]
#> [[6]]$role
#> [1] "user"
#> 
#> [[6]]$content
#> [1] "What is the capital of Algeria?"
#> 
#> 
#> [[7]]
#> [[7]]$role
#> [1] "assistant"
#> 
#> [[7]]$content
#> [1] "The capital of Algeria is Algiers."
```

``` r
# Keep only the last 2 messages (system prompt is preserved)
agent$keep_last_n_messages(n = 2)
#> ✔ Conversation truncated to last 2 messages.
agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are a concise assistant."
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "What is the capital of Algeria?"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "The capital of Algeria is Algiers."
```

#### Manually Adding Messages to an Agent’s History

You can inject any message (system, user, or assistant) directly into an
Agent’s history with `add_message(role, content)`. This is helpful to
reconstruct, supplement, or simulate conversation steps.

- **add_message(role, content)**:
  - `role`: “user”, “assistant”, or “system”
  - `content`: The text message to add

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)
agent <- Agent$new(
  name = "Pizza expert",
  instruction = "You are a Pizza expert",
  llm_object = openai_4_1_mini
)

# Add a user message, an assistant reply, and a system instruction:
agent$add_message("user", "Where can I find the best pizza in the world?")
#> ✔ Added user message: Where can I find the best pizza in the world?...
agent$add_message("assistant", "You can find the best pizza in the world in Algiers, Algeria. It's tasty and crunchy.")
#> ✔ Added assistant message: You can find the best pizza in the world in Algier...

# View conversation history
agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are a Pizza expert"
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "Where can I find the best pizza in the world?"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "You can find the best pizza in the world in Algiers, Algeria. It's tasty and crunchy."
```

This makes it easy to reconstruct or extend sessions, provide custom
context, or insert notes for debugging/testing purposes.

``` r
agent$invoke("summarise the previous conversation")
#> You asked where to find the best pizza in the world, and I responded that the 
#> best pizza can be found in Algiers, Algeria, known for being tasty and crunchy.
```

#### Sync between `messages` and `turns`

You can modify the `messages` object as you please, this will be
automatically translated to the suitable `turns` required by `ellmer`:

``` r
agent$messages[[5]]$content <- "Obivously you asked me about the best pizza in the world which is of course in Algiery!"

agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are a Pizza expert"
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "Where can I find the best pizza in the world?"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "You can find the best pizza in the world in Algiers, Algeria. It's tasty and crunchy."
#> 
#> 
#> [[4]]
#> [[4]]$role
#> [1] "user"
#> 
#> [[4]]$content
#> [1] "summarise the previous conversation"
#> 
#> 
#> [[5]]
#> [[5]]$role
#> [1] "assistant"
#> 
#> [[5]]$content
#> [1] "Obivously you asked me about the best pizza in the world which is of course in Algiery!"
```

The underlying ellmer object:

``` r
agent$llm_object
#> <Chat OpenAI/gpt-4.1-mini turns=5 input=62 output=37>
#> ── system ──────────────────────────────────────────────────────────────────────
#> You are a Pizza expert
#> ── user ────────────────────────────────────────────────────────────────────────
#> Where can I find the best pizza in the world?
#> ── assistant [input=0 output=0] ────────────────────────────────────────────────
#> You can find the best pizza in the world in Algiers, Algeria. It's tasty and crunchy.
#> ── user ────────────────────────────────────────────────────────────────────────
#> summarise the previous conversation
#> ── assistant [input=62 output=37] ──────────────────────────────────────────────
#> Obivously you asked me about the best pizza in the world which is of course in Algiery!
```

#### Resetting conversation history

If you want to clear the conversation while preserving the current
system prompt, use `reset_conversation_history()`.

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

agent <- Agent$new(
  name = "session_reset",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)

agent$invoke("Tell me a short fun fact about dates (the fruit).")
#> Sure! Did you know that dates are one of the oldest cultivated fruits in the 
#> world, with evidence of their harvest dating back over 6,000 years? They were a
#> staple food in ancient Mesopotamia and have been enjoyed for thousands of 
#> years!
agent$invoke("And one more.")
#> Absolutely! Here’s another fun fact: Dates have such a natural sweetness that 
#> they were often used as a natural sweetener in ancient recipes—kind of like 
#> nature’s candy before sugar was widely available!

# Clear all messages except the system prompt
agent$reset_conversation_history()
#> ✔ Conversation history reset. System prompt preserved.
agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are an assistant."
```

#### Exporting and Loading Agent Conversation History

You can save an agent’s conversation history to a file and reload it
later. This allows you to archive, transfer, or resume agent sessions
across R sessions or machines.

- **export_messages_history(file_path)**: Saves the current conversation
  to a JSON file.
- **load_messages_history(file_path)**: Loads a saved conversation
  history from a JSON file, replacing the agent’s current history.

In both methods, if you omit the `file_path` parameter, a default file
named `"<getwd()>/<agent_name>_messages.json"` is used.

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)
agent <- Agent$new(
  name = "session_agent",
  instruction = "You are a persistent researcher.",
  llm_object = openai_4_1_mini
)

# Interact with the agent
agent$invoke("Tell me something interesting about volcanoes.")

# Save the conversation
agent$export_messages_history("volcano_session.json")

# ...Later, or in a new session...
# Restore the conversation
agent$load_messages_history("volcano_session.json")
# agent$messages  # Displays current history
```

#### Updating the system instruction during a session

Use `update_instruction(new_instruction)` to change the Agent’s system
prompt mid-session. The first system message and the underlying `ellmer`
system prompt are both updated.

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

agent <- Agent$new(
  name = "reconfigurable",
  instruction = "You are a helpful assistant.",
  llm_object = openai_4_1_mini
)

agent$update_instruction("You are a strictly concise assistant. Answer in one sentence.")
#> ✔ Instruction successfully updated
#> ℹ Old: You are a helpful assistant....
#> ℹ New: You are a strictly concise assistant. Answer in on...

agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are a strictly concise assistant. Answer in one sentence."
```

## Budget and cost control

You can limit how much an `Agent` is allowed to spend and decide what
should happen as the budget is approached or exceeded. Use
`set_budget()` to define the maximum spend (in USD), and
`set_budget_policy()` to control warnings and over-budget behavior.

- **set_budget(amount_in_usd)**: sets the absolute budget for the agent.
- **set_budget_policy(on_exceed, warn_at)**:
  - **on_exceed**: one of `"abort"`, `"warn"`, or `"ask"`.
    - **abort**: stop with an error when the budget is exceeded.
    - **warn**: emit a warning and continue.
    - **ask**: interactively ask what to do when the budget is exceeded.
  - **warn_at**: a fraction in (0, 1); triggers a one-time warning when
    spending reaches that fraction of the budget (default `0.8`).

``` r
# An API KEY is required to invoke the Agent
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

agent <- Agent$new(
  name = "cost_conscious_assistant",
  instruction = "Answer succinctly.",
  llm_object = openai_4_1_mini
)

# Set a 5 USD budget
agent$set_budget(5)
#> ✔ Budget successfully set to 5$
#> ℹ Budget policy: on_exceed='abort', warn_at=0.8
#> ℹ Use the set_budget_policy() method to configure the budget policy.

# Warn at 90% of the budget and ask what to do if exceeded
agent$set_budget_policy(on_exceed = "ask", warn_at = 0.9)
#> ✔ Budget policy set: on_exceed='ask', warn_at=0.9

# Normal usage
agent$invoke("Give me a one-sentence fun fact about Algeria.")
#> Algeria is home to the Sahara Desert's largest portion and hosts the ancient 
#> Roman ruins of Timgad, known as the "Pompeii of Africa."
```

The current policy is echoed when setting the budget. You can update the
policy at any time before or during an interaction lifecycle to adapt to
your workflow’s tolerance for cost overruns.

#### Inspecting usage and estimated cost

Call `get_usage_stats()` to retrieve the estimated cost, and budget
information (if set).

``` r
stats <- agent$get_usage_stats()
stats
#> $estimated_cost
#> [1] 1e-04
#> 
#> $budget
#> [1] 5
#> 
#> $budget_remaining
#> [1] 4.9999
```

## Generate and execute R code from natural language

`generate_execute_r_code()` lets an `Agent` translate a natural-language
task description into R code, optionally validate its syntax, and
(optionally) execute it.

- **code_description**: a plain-English description of the R code to
  generate.
- **validate**: `TRUE` to run a syntax validation step on the generated
  code first.
- **execute**: `TRUE` to execute the generated code (requires successful
  validation).
- **interactive**: if `TRUE`, shows the code and asks for confirmation
  before executing.
- **env**: environment where code will run when `execute = TRUE`
  (default `globalenv()`).

Safety notes: - Set `validate = TRUE` and review the printed code before
execution. - Keep `interactive = TRUE` to require an explicit
confirmation before running code.

``` r
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

r_assistant <- Agent$new(
  name = "R Code Assistant",
  instruction = "You are an expert R programmer.",
  llm_object = openai_4_1_mini
)

agent$generate_execute_r_code(
   code_description = "using ggplot2, generate a scatterplot of hwy and cty in red", 
   validate = TRUE, 
   execute = TRUE, 
   interactive = FALSE
 )
#> ℹ Executing generated R code...
#> ✔ Code executed successfully
#> $description
#> [1] "using ggplot2, generate a scatterplot of hwy and cty in red"
#> 
#> $code
#> library(ggplot2);ggplot(mpg,aes(x=cty,y=hwy))+geom_point(color="red")
#> 
#> $validated
#> [1] TRUE
#> 
#> $validation_message
#> [1] "Syntax is valid"
#> 
#> $executed
#> [1] TRUE
#> 
#> $execution_result
#> $execution_result$value
```

<img src="man/figures/README-unnamed-chunk-23-1.png" width="100%" />

    #> 
    #> $execution_result$output
    #> character(0)

## Cloning an Agent

If you want to create a new agent with the exact same characteristics,
you can use the `clone_agent` method. Note that the new Agent can have
the same name but it’ll have a different `ID`:

``` r
rai_agent <- Agent$new(
  name = "Rai musician",
  instruction = "You are an expert in Algerian Rai music",
  llm_object = openai_4_1_mini
)

result <- rai_agent$invoke("Give me a rai song in 1 sentence. Don't explain")

rai_agent$agent_id
#> [1] "76243ae7-0d8f-4aa9-96e9-3f8d3c10ec75"
rai_agent$name
#> [1] "Rai musician"
rai_agent$instruction
#> [1] "You are an expert in Algerian Rai music"
rai_agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are an expert in Algerian Rai music"
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "Give me a rai song in 1 sentence. Don't explain"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "\"Ya Rayah\" by Rachid Taha."
```

``` r
new_rai_agent <- rai_agent$clone_agent(new_name = "Just Rai")
#> ✔ Agent cloned successfully. New ID: f4790a08-8109-456c-ba5d-a459f39f95e3

new_rai_agent$agent_id
#> [1] "f4790a08-8109-456c-ba5d-a459f39f95e3"
new_rai_agent$name
#> [1] "Just Rai"
new_rai_agent$instruction
#> [1] "You are an expert in Algerian Rai music"
new_rai_agent$messages
#> [[1]]
#> [[1]]$role
#> [1] "system"
#> 
#> [[1]]$content
#> [1] "You are an expert in Algerian Rai music"
#> 
#> 
#> [[2]]
#> [[2]]$role
#> [1] "user"
#> 
#> [[2]]$content
#> [1] "Give me a rai song in 1 sentence. Don't explain"
#> 
#> 
#> [[3]]
#> [[3]]$role
#> [1] "assistant"
#> 
#> [[3]]$content
#> [1] "\"Ya Rayah\" by Rachid Taha."
```

## Response Validation

The `validate_response()` method provides intelligent LLM-based
validation of agent responses against custom criteria. This powerful
feature uses the agent’s own LLM to evaluate whether a response meets
specified validation standards, returning both a score and detailed
feedback.

### How it works

The method evaluates the response against your criteria. It returns a
validation score (0-1) and determines if the response is valid based on
your threshold.

### Parameters

- **prompt**: The original prompt that generated the response
- **response**: The response text to validate
- **validation_criteria**: Your validation requirements (e.g., “Must be
  factual and under 50 words”)
- **validation_score**: Score threshold for validity (0-1, default 0.8)

#### Example 1: Factual Content Validation

``` r
fact_checker <- Agent$new(
  name = "fact_checker",
  instruction = "You are a factual assistant.",
  llm_object = openai_4_1_mini
)

prompt <- "What is the capital of Algeria?"
response <- fact_checker$invoke(prompt)

validation <- fact_checker$validate_response(
  prompt = prompt,
  response = response,
  validation_criteria = "The response must be factually accurate and mention Algiers as the capital",
  validation_score = 0.8
)
#> ✔ The response is considered valid with a score of 1

validation
#> $prompt
#> [1] "What is the capital of Algeria?"
#> 
#> $response
#> The capital of Algeria is Algiers.
#> 
#> $validation_criteria
#> [1] "The response must be factually accurate and mention Algiers as the capital"
#> 
#> $validation_score
#> [1] 0.8
#> 
#> $valid
#> [1] TRUE
#> 
#> $score
#> [1] 1
#> 
#> $feedback
#> [1] "The response is factually accurate and explicitly mentions Algiers as the capital of Algeria, fully meeting the validation criteria."
```

#### Example 2: Content Length and Style Validation

``` r
content_agent <- Agent$new(
  name = "content_creator",
  instruction = "You are a creative writing assistant.",
  llm_object = openai_4_1_mini
)

prompt <- "Write a 1 sentence advertisment about an Algerian dates (the fruid)"

response <- content_agent$invoke(prompt)

validation <- content_agent$validate_response(
  prompt = prompt,
  response = response,
  validation_criteria = "Response must be under 100 words, professional tone, and highlight Algerian dates",
  validation_score = 0.75
)
#> ✔ The response is considered valid with a score of 1

validation
#> $prompt
#> [1] "Write a 1 sentence advertisment about an Algerian dates (the fruid)"
#> 
#> $response
#> Experience the rich, natural sweetness of Algerian dates—nature’s golden 
#> treasure packed with flavor and nutrition in every bite!
#> 
#> $validation_criteria
#> [1] "Response must be under 100 words, professional tone, and highlight Algerian dates"
#> 
#> $validation_score
#> [1] 0.75
#> 
#> $valid
#> [1] TRUE
#> 
#> $score
#> [1] 1
#> 
#> $feedback
#> [1] "The response is a single sentence advertisement that highlights Algerian dates with a professional tone. It emphasizes the qualities of the dates, such as rich natural sweetness, flavor, and nutrition, meeting all the validation criteria perfectly."
```

### Use Cases

- **Quality Control**: Validate responses meet content standards before
  publication
- **Factual Accuracy**: Ensure responses contain correct information
- **Style Compliance**: Check responses follow tone, length, or format
  requirements
- **Safety Filtering**: Validate content meets safety and
  appropriateness criteria
- **A/B Testing**: Compare response quality across different models or
  prompts

The validation results include the original prompt, response, criteria,
score, feedback, and validity status, making it easy to audit and
improve your agent’s performance.

## Tools

###### At the moment, only in the development version

You can easily register one or several tools using the `register_tools`
method. The tools are created using `ellmer`, consider the following
example:

``` r
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  credentials = function() {Sys.getenv("OPENAI_API_KEY")},
  echo = "none"
)

weather_agent <- Agent$new(
  name = "weather_assistant",
  instruction = "You are a weather assistant.",
  llm_object = openai_4_1
)

weather_function_algiers <- function() {
  msg <- glue::glue(
    "35 degrees Celcius, it's sunny and there's no precipitation."
  )
  msg
}

get_weather_in_algiers <- ellmer::tool(
  fun = weather_function_algiers,
  name = "get_weather_in_algiers",
  description = "Provide the current weather in Algiers, Algeria."
)

weather_function_berlin <- function() {
  msg <- glue::glue(
    "10 degrees Celcius, it's cold"
  )
  msg
}

get_weather_in_berlin <- ellmer::tool(
  fun = weather_function_berlin,
  name = "get_weather_in_berlin",
  description = "Provide the current weather in Berlin, Germany"
)

weather_agent$register_tools(
  tools = list(
    get_weather_in_algiers, 
    get_weather_in_berlin
  )
)
#> ✔ Registered tool: get_weather_in_algiers
#> ✔ Registered tool: get_weather_in_berlin
```

One can list the available tools:

``` r
weather_agent$list_tools()
#> [1] "get_weather_in_algiers" "get_weather_in_berlin"
```

After registering the tools, the Agent will use them when needed:

``` r
weather_agent$invoke("How's the weather in Algiers?")
#> The weather in Algiers is currently sunny with a temperature of 35°C and no 
#> precipitation.
```

``` r
weather_agent$invoke("How's the weather in Berlin?")
#> The weather in Berlin is currently cold with a temperature of 10°C.
```

One can remove one or several tool using the `remove_tools` method or
remove all agents at one using the `clear_tools` method:

``` r
weather_agent$clear_tools()
#> ✔ Cleared 2 tools
weather_agent$list_tools()
#> ℹ No tools registered
#> character(0)
```

### Tool Generation

###### At the moment, only in the development version

The `generate_and_register_tool` method allows you to create tools from
simple natural language descriptions (for example, “create a tool that
saves files to disk”) and automatically generates the complete R code
needed to implement them. It produces a fully functional R function that
encapsulates the tool’s logic, along with a complete `ellmer` tool
definition that includes proper type specifications and clear parameter
descriptions.

``` r
weather_agent$generate_and_register_tool(
  description = "create a tool that uses httr to call the open-meteo api https://open-meteo.com/en/docs to get the current weather about any city in the world"
)
#> 
#> ── Generating tool from description ──
#> 
#> ℹ Description: create a tool that uses httr to call the open-meteo api https://open-meteo.com/en/docs to get the current weather about any city in the world
#> 
#> ── The following tool will be registered ──
#> 
#> get_current_weather <- function(city) {
#>   library(httr)
#>   library(jsonlite)
#>   geo_url <- "https://geocoding-api.open-meteo.com/v1/search"
#>   geo_res <- httr::GET(geo_url, query = list(name = city, count = 1))
#>   geo_data <- jsonlite::fromJSON(httr::content(geo_res, as = "text", encoding = "UTF-8"))
#>   if (is.null(geo_data$results) || length(geo_data$results) == 0) {
#>     stop(sprintf("City '%s' not found.", city))
#>   }
#>   lat <- geo_data$results$latitude[1]
#>   lon <- geo_data$results$longitude[1]
#>   weather_url <- "https://api.open-meteo.com/v1/forecast"
#>   weather_res <- httr::GET(weather_url, query = list(latitude = lat, longitude = lon, current_weather = "true"))
#>   weather_data <- jsonlite::fromJSON(httr::content(weather_res, as = "text", encoding = "UTF-8"))
#>   if (is.null(weather_data$current_weather)) {
#>     stop("Current weather data not available.")
#>   }
#>   list(
#>     city = city,
#>     latitude = lat,
#>     longitude = lon,
#>     temperature = weather_data$current_weather$temperature,
#>     windspeed = weather_data$current_weather$windspeed,
#>     winddirection = weather_data$current_weather$winddirection,
#>     weathercode = weather_data$current_weather$weathercode,
#>     time = weather_data$current_weather$time
#>   )
#> }
#> get_current_weather <- ellmer::tool(
#>   get_current_weather,
#>   name = "get_current_weather",
#>   description = "Get the current weather for any city in the world using the open-meteo API.",
#>   arguments = list(
#>     city = ellmer::type_string(
#>       description = "Name of the city to get the current weather for. Can include optional country for disambiguation (e.g. 'Paris, France')."
#>     )
#>   )
#> )
#> ✔ Registered tool: get_current_weather
#> ✔ Tool successfully generated and registered
#> ℹ Call '<agent-name>$llm_object$get_tools()' to inspect the tools
#> ℹ If satisfied, you can copy the tool and put in your corresponding R file
```

``` r
weather_agent$invoke(
  prompt = "what is the current weather in Tokyo?"
)
#> The current weather in Tokyo is 1.3°C with a wind speed of 5.2 km/h (direction:
#> 34°). The weather code is 51, which typically indicates light drizzle or mist.
```

# LeadAgent

## Creating a multi-agents orchestraction

We can create as many Agents as we want, the `LeadAgent` will dispatch
the instructions to the agents and provide with the final answer back.
Let’s create three Agents, a `researcher`, a `summarizer` and a
`translator`:

``` r

researcher <- Agent$new(
  name = "researcher",
  instruction = "You are a research assistant. Your job is to answer factual questions with detailed and accurate information. Do not answer with more than 2 lines",
  llm_object = openai_4_1_mini
)

summarizer <- Agent$new(
  name = "summarizer",
  instruction = "You are agent designed to summarise a give text into 3 distinct bullet points.",
  llm_object = openai_4_1_mini
)

translator <- Agent$new(
  name = "translator",
  instruction = "Your role is to translate a text from English to German",
  llm_object = openai_4_1_mini
)
```

Now, the most important part is to create a `LeadAgent`:

``` r
lead_agent <- LeadAgent$new(
  name = "Leader", 
  llm_object = openai_4_1_mini
)
```

Note that the `LeadAgent` cannot receive an `instruction` as it has
already the necessary instructions.

Next, we need to assign the Agents to `LeadAgent`, we do it as follows:

``` r
lead_agent$register_agents(c(researcher, summarizer, translator))
#> ✔ Agent(s) successfully registered.

lapply(lead_agent$agents, function(x) {x$name})
#> [[1]]
#> [1] "researcher"
#> 
#> [[2]]
#> [1] "summarizer"
#> 
#> [[3]]
#> [1] "translator"
```

Before executing your prompt, you can ask the `LeadAgent` to generate a
plan so that you can see which `Agent` will be used for which prompt,
you can do it as follows:

``` r
prompt_to_execute <- "Tell me about the economic situation in Algeria, summarize it in 3 bullet points, then translate it into German."

plan <- lead_agent$generate_plan(prompt_to_execute)
#> ✔ Plan successfully generated.
plan
#> [[1]]
#> [[1]]$agent_id
#> 31256121-0e52-481a-b28f-144ce2d2b1c5
#> 
#> [[1]]$agent_name
#> [1] "researcher"
#> 
#> [[1]]$model_provider
#> [1] "OpenAI"
#> 
#> [[1]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[1]]$prompt
#> [1] "Research the current economic situation in Algeria, focusing on key indicators such as GDP growth, unemployment rate, inflation, and major economic sectors."
#> 
#> 
#> [[2]]
#> [[2]]$agent_id
#> f512a0a5-d605-49b4-98e3-8639c17095b7
#> 
#> [[2]]$agent_name
#> [1] "summarizer"
#> 
#> [[2]]$model_provider
#> [1] "OpenAI"
#> 
#> [[2]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[2]]$prompt
#> [1] "Summarize the researched economic information into 3 concise bullet points highlighting the most important aspects."
#> 
#> 
#> [[3]]
#> [[3]]$agent_id
#> 018c0194-781d-4f6e-a7d2-adbb43ba3e0c
#> 
#> [[3]]$agent_name
#> [1] "translator"
#> 
#> [[3]]$model_provider
#> [1] "OpenAI"
#> 
#> [[3]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[3]]$prompt
#> [1] "Translate the 3 bullet points from English into German accurately, maintaining the original meaning."
```

Now, in order now to execute the workflow, we just need to call the
`invoke` method which will behind the scene delegate the prompts to
suitable Agents and retrieve back the final information:

``` r
response <- lead_agent$invoke("Tell me about the economic situation in Algeria, summarize it in 3 bullet points, then translate it into German.")
#> 
#> ── Using existing plan ──
#> 
```

``` r
response
#> - Das BIP-Wachstum Algeriens ist mit 2-3 % moderat und wird hauptsächlich durch
#> Hydrokarbone sowie Diversifizierungsinitiativen angetrieben.  
#> - Die Arbeitslosigkeit bleibt mit 11-12 % hoch, wobei die 
#> Jugendarbeitslosigkeit ein erhebliches Problem darstellt.  
#> - Die Inflation ist niedrig, etwa bei 2-3 %, während die wichtigsten 
#> Wirtschaftssektoren Hydrokarbone, Landwirtschaft und Dienstleistungen umfassen.
```

If you want to inspect the multi-agents orchestration, you have access
to the `agents_interaction` object:

``` r
lead_agent$agents_interaction
#> [[1]]
#> [[1]]$agent_id
#> 31256121-0e52-481a-b28f-144ce2d2b1c5
#> 
#> [[1]]$agent_name
#> [1] "researcher"
#> 
#> [[1]]$model_provider
#> [1] "OpenAI"
#> 
#> [[1]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[1]]$prompt
#> [1] "Research the current economic situation in Algeria, focusing on key indicators such as GDP growth, unemployment rate, inflation, and major economic sectors."
#> 
#> [[1]]$response
#> As of early 2024, Algeria's GDP growth is modest, around 2-3%, driven by 
#> hydrocarbons and diversification efforts. Unemployment remains high, about 
#> 11-12%, especially among youth. Inflation is relatively low, near 2-3%. Major 
#> sectors are hydrocarbons (oil and gas), agriculture, and services.
#> 
#> [[1]]$edited_by_hitl
#> [1] FALSE
#> 
#> 
#> [[2]]
#> [[2]]$agent_id
#> f512a0a5-d605-49b4-98e3-8639c17095b7
#> 
#> [[2]]$agent_name
#> [1] "summarizer"
#> 
#> [[2]]$model_provider
#> [1] "OpenAI"
#> 
#> [[2]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[2]]$prompt
#> [1] "Summarize the researched economic information into 3 concise bullet points highlighting the most important aspects."
#> 
#> [[2]]$response
#> - Algeria's GDP growth is modest at 2-3%, primarily fueled by hydrocarbons and 
#> diversification initiatives.  
#> - Unemployment remains high at 11-12%, with youth unemployment being a 
#> significant concern.  
#> - Inflation is low, around 2-3%, while key economic sectors include 
#> hydrocarbons, agriculture, and services.
#> 
#> [[2]]$edited_by_hitl
#> [1] FALSE
#> 
#> 
#> [[3]]
#> [[3]]$agent_id
#> 018c0194-781d-4f6e-a7d2-adbb43ba3e0c
#> 
#> [[3]]$agent_name
#> [1] "translator"
#> 
#> [[3]]$model_provider
#> [1] "OpenAI"
#> 
#> [[3]]$model_name
#> [1] "gpt-4.1-mini"
#> 
#> [[3]]$prompt
#> [1] "Translate the 3 bullet points from English into German accurately, maintaining the original meaning."
#> 
#> [[3]]$response
#> - Das BIP-Wachstum Algeriens ist mit 2-3 % moderat und wird hauptsächlich durch
#> Hydrokarbone sowie Diversifizierungsinitiativen angetrieben.  
#> - Die Arbeitslosigkeit bleibt mit 11-12 % hoch, wobei die 
#> Jugendarbeitslosigkeit ein erhebliches Problem darstellt.  
#> - Die Inflation ist niedrig, etwa bei 2-3 %, während die wichtigsten 
#> Wirtschaftssektoren Hydrokarbone, Landwirtschaft und Dienstleistungen umfassen.
#> 
#> [[3]]$edited_by_hitl
#> [1] FALSE
```

The above example is extremely simple, the usefulness of `mini007` would
shine in more complex processes where a multi-agent sequential
orchestration has a higher value added.

## Visualizing agent plans with `visualize_plan()`

Sometimes, before running your workflow, it is helpful to view the
orchestration as a visual diagram, showing the sequence of agents and
which prompt each will receive. After generating a plan, you can call
`visualize_plan()`:

This function displays the agents in workflow order as labeled boxes.
Hovering a box reveals the delegated prompt. The visualization uses the
`DiagrammeR` package. If no plan exists, it asks you to generate one
first.

``` r
lead_agent$visualize_plan()
```

## Broadcasting

If you want to compare several `LLM` models, the `LeadAgent` provides a
`broadcast` method that allows you to send a prompt to several different
agents and get the result for each agent back in order to make a
comparison and potentially choose the best agent/model for the defined
prompt:

Let’s go through an example:

``` r
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

openai_4_1_agent <- Agent$new(
  name = "openai_4_1_agent", 
  instruction = "You are an AI assistant. Answer in 1 sentence max.", 
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

openai_4_1_nano_agent <- Agent$new(
  name = "openai_4_1_nano_agent", 
  instruction = "You are an AI assistant. Answer in 1 sentence max.", 
  llm_object = openai_4_1_nano
)
```

``` r

lead_agent$clear_agents() # removing previous agents
lead_agent$register_agents(c(openai_4_1_agent, openai_4_1_nano_agent))
#> ✔ Agent(s) successfully registered.
```

``` r
lead_agent$broadcast(prompt = "If I were Algerian, which song would I like to sing when running under the rain? how about a flower?")
#> [[1]]
#> [[1]]$agent_id
#> [1] "cd06f7bb-7b47-4233-9d0d-6f49a432a925"
#> 
#> [[1]]$agent_name
#> [1] "openai_4_1_agent"
#> 
#> [[1]]$model_provider
#> [1] "OpenAI"
#> 
#> [[1]]$model_name
#> [1] "gpt-4.1"
#> 
#> [[1]]$response
#> As an Algerian, you might enjoy singing "Ya Rayah" when running under the rain,
#> and for a flower, you could sing "Fleur d’Algérie."
#> 
#> 
#> [[2]]
#> [[2]]$agent_id
#> [1] "73055678-6ce9-4357-83ee-1f5e7f122ee8"
#> 
#> [[2]]$agent_name
#> [1] "openai_4_1_nano_agent"
#> 
#> [[2]]$model_provider
#> [1] "OpenAI"
#> 
#> [[2]]$model_name
#> [1] "gpt-4.1-nano"
#> 
#> [[2]]$response
#> As an Algerian, you might enjoy singing "Ya Rayah" by Rachid Taha when running 
#> under the rain and "Lila" by Cheb Khaled when admiring a flower.
```

You can also access the history of the `broadcasting` using the
`broadcast_history` attribute:

``` r
lead_agent$broadcast_history
#> [[1]]
#> [[1]]$prompt
#> [1] "If I were Algerian, which song would I like to sing when running under the rain? how about a flower?"
#> 
#> [[1]]$responses
#> [[1]]$responses[[1]]
#> [[1]]$responses[[1]]$agent_id
#> [1] "cd06f7bb-7b47-4233-9d0d-6f49a432a925"
#> 
#> [[1]]$responses[[1]]$agent_name
#> [1] "openai_4_1_agent"
#> 
#> [[1]]$responses[[1]]$model_provider
#> [1] "OpenAI"
#> 
#> [[1]]$responses[[1]]$model_name
#> [1] "gpt-4.1"
#> 
#> [[1]]$responses[[1]]$response
#> As an Algerian, you might enjoy singing "Ya Rayah" when running under the rain,
#> and for a flower, you could sing "Fleur d’Algérie."
#> 
#> 
#> [[1]]$responses[[2]]
#> [[1]]$responses[[2]]$agent_id
#> [1] "73055678-6ce9-4357-83ee-1f5e7f122ee8"
#> 
#> [[1]]$responses[[2]]$agent_name
#> [1] "openai_4_1_nano_agent"
#> 
#> [[1]]$responses[[2]]$model_provider
#> [1] "OpenAI"
#> 
#> [[1]]$responses[[2]]$model_name
#> [1] "gpt-4.1-nano"
#> 
#> [[1]]$responses[[2]]$response
#> As an Algerian, you might enjoy singing "Ya Rayah" by Rachid Taha when running 
#> under the rain and "Lila" by Cheb Khaled when admiring a flower.
```

## Human In The Loop (HITL)

When executing an LLM workflow that relies on many steps, you can set
`Human In The Loop` (`HITL`) trigger that will check the model’s
response at a specific step. You can define a `HITL` trigger after
defining a `LeadAgent` as follows:

``` r
openai_llm_object <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

lead_agent <- LeadAgent$new(
  name = "Leader", 
  llm_object = openai_llm_object
)

lead_agent$set_hitl(steps = 1)
#> ✔ HITL successfully set at step(s) 1.

lead_agent$hitl_steps
#> [1] 1
```

After setting the `HITL` to step 1, the workflow execution will pose and
give the user 3 choices:

1.  Continue the execution of the workflow as it is;
2.  Change manually the answer of the specified step and continue the
    execution of the workflow;
3.  Stop the execution of the workflow (hard error);

Note that you can set a `HITL` at several steps, for example
`lead_agent$set_hitl(steps = c(1, 2))` will set the `HITL` at step 1 and
step 2.

## Judge as a decision process

Sometimes you want to send a prompt to several agents and pick the best
answer. In order to choose the best prompt, you can also rely on the
`Lead` Agent which will act a dudge and pick for you the best answer.
You can use the `judge_and_choose_best_response` method as follows:

``` r
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

stylist_1 <- Agent$new(
  name = "stylist",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

stylist_2 <- Agent$new(
  name = "stylist2",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1_nano
)

openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  credentials = retrieve_open_ai_credential, 
  echo = "none"
)

stylist_lead_agent <- LeadAgent$new(
  name = "Stylist Leader",
  llm_object = openai_4_1_mini
)

stylist_lead_agent$register_agents(c(stylist_1, stylist_2))
#> ✔ Agent(s) successfully registered.

best_answer <- stylist_lead_agent$judge_and_choose_best_response(
  "what's the best way to wear a blue kalvin klein shirt in winter with a pink pair of trousers?"
)

best_answer
#> $proposals
#> $proposals[[1]]
#> $proposals[[1]]$agent_id
#> [1] "91f03880-c94d-4ee7-850a-8e20b6054c68"
#> 
#> $proposals[[1]]$agent_name
#> [1] "stylist"
#> 
#> $proposals[[1]]$response
#> Layer the blue Calvin Klein shirt under a neutral-colored sweater or blazer 
#> (like grey or navy), pair with your pink trousers, and add classic winter 
#> accessories like a dark coat and matching scarf for a balanced, stylish look.
#> 
#> 
#> $proposals[[2]]
#> $proposals[[2]]$agent_id
#> [1] "b1be43b0-99ea-4bb0-9c1b-5d244ace0257"
#> 
#> $proposals[[2]]$agent_name
#> [1] "stylist2"
#> 
#> $proposals[[2]]$response
#> Layer the blue Calvin Klein shirt with a neutral-colored sweater or blazer and 
#> add warm accessories to create a stylish winter look with pink trousers.
#> 
#> 
#> 
#> $chosen_response
#> Layer the blue Calvin Klein shirt under a neutral-colored sweater or blazer 
#> (like grey or navy), pair with your pink trousers, and add classic winter 
#> accessories like a dark coat and matching scarf for a balanced, stylish look.
```

## Agents Dialog

###### At the moment, only in the development version

The `agents_dialog` method facilitates an intelligent two-agent
collaboration process designed to refine and optimize responses through
iterative dialogue.

It enables two registered agents to take alternating turns improving
each other’s outputs until a high-quality final response is reached. The
method supports a configurable maximum number of iterations (default: 5)
and includes a self-stopping mechanism where agents can indicate
agreement by beginning their message with “CONSENSUS:”, followed by the
final answer.

If no consensus is achieved within the iteration limit, the lead agent
automatically synthesizes a concluding response based on the
conversation. Throughout the exchange, every interaction is stored
within the `self$dialog_history` object. Consider the following
examples:

``` r
ceo1 <- Agent$new(
  name = "ceo1",
  instruction = paste0(
    "You are a CEO in  a dates company based in Ouergla, Algeria, ", 
    "You want to boost their exports to Germany. "
  ),
  llm_object = openai_4_1_mini
)

ceo2 <- Agent$new(
  name = "ceo2",
  instruction = paste0(
    "You are the CEO of a dates company based in Ouergla, Algeria. ", 
    "You are considering starting a marketing compaign to boost the exports to Germany. "
  ),
  llm_object = openai_4_1_mini
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(ceo1, ceo2))
#> ✔ Agent(s) successfully registered.

result <- lead_agent$agents_dialog(
  prompt = "Propose a plan in 1 sentence max about a marketing strategy that will boost the export of dates to Germany for the next 2 years",
  agent_1_id = ceo1$agent_id,
  agent_2_id = ceo2$agent_id,
  max_iterations = 3
)
#> 
#> ── Starting agent dialog ──
#> 
#> Agent 1: ceo1
#> Agent 2: ceo2
#> Max iterations: 3
#> ℹ Iteration 1 - ceo1 responding...
#> ℹ Iteration 1 - ceo2 responding...
#> ℹ Iteration 2 - ceo1 responding...
#> ℹ Iteration 2 - ceo2 responding...
#> ℹ Iteration 3 - ceo1 responding...
#> ℹ Iteration 3 - ceo2 responding...
#> ✔ Consensus reached by ceo2 at iteration 3!
#> ✔ Dialog completed.

# Access the final response
result$final_response
#> execute a digital marketing campaign featuring Ouergla heritage and 
#> sustainability, augmented by virtual tastings, influencer partnerships, and 
#> selective on-site sampling through German retail partners at major food events 
#> over the next two years.

# View the dialog history
result$dialog_history
#> [[1]]
#> [[1]]$iteration
#> [1] 1
#> 
#> [[1]]$agent_id
#> [1] "6059ac80-4e2b-4cde-b1a0-0a0de7a8f2e8"
#> 
#> [[1]]$agent_name
#> [1] "ceo1"
#> 
#> [[1]]$response
#> Develop a targeted digital marketing campaign highlighting the premium quality 
#> and health benefits of Ouergla dates, coupled with partnerships with German 
#> organic and specialty food retailers to build brand trust and visibility over 
#> the next two years.
#> 
#> 
#> [[2]]
#> [[2]]$iteration
#> [1] 1
#> 
#> [[2]]$agent_id
#> [1] "cd526cbe-feaa-4c2d-89bb-2343412dc215"
#> 
#> [[2]]$agent_name
#> [1] "ceo2"
#> 
#> [[2]]$response
#> Your proposal to develop a targeted digital marketing campaign emphasizing the 
#> premium quality and health benefits of Ouergla dates is well aligned with our 
#> goal to boost exports to Germany; however, to fully leverage our unique 
#> regional advantage, I suggest also incorporating storytelling about the Ouergla
#> heritage and sustainable farming practices in the campaign, and organizing 
#> German tasting events in key cities to engage consumers directly over the next 
#> two years.
#> 
#> 
#> [[3]]
#> [[3]]$iteration
#> [1] 2
#> 
#> [[3]]$agent_id
#> [1] "6059ac80-4e2b-4cde-b1a0-0a0de7a8f2e8"
#> 
#> [[3]]$agent_name
#> [1] "ceo1"
#> 
#> [[3]]$response
#> Your addition of Ouergla heritage storytelling and sustainable farming 
#> practices enriches the digital campaign well, but organizing tasting events in 
#> Germany may incur high costs and logistic complexities that strain our current 
#> export budget; instead, I propose leveraging local German partners to host 
#> joint virtual tasting sessions and influencer collaborations to maintain 
#> consumer engagement efficiently over the next two years.
#> 
#> 
#> [[4]]
#> [[4]]$iteration
#> [1] 2
#> 
#> [[4]]$agent_id
#> [1] "cd526cbe-feaa-4c2d-89bb-2343412dc215"
#> 
#> [[4]]$agent_name
#> [1] "ceo2"
#> 
#> [[4]]$response
#> Your suggestion to replace physical tasting events with virtual tastings and 
#> influencer collaborations is sensible given budget and logistical constraints; 
#> to further ensure trust and local presence, I propose combining these virtual 
#> engagements with limited on-site sampling through select German retail partners
#> during major food fairs or festivals to maximize impact within budget over the 
#> next two years.
#> 
#> 
#> [[5]]
#> [[5]]$iteration
#> [1] 3
#> 
#> [[5]]$agent_id
#> [1] "6059ac80-4e2b-4cde-b1a0-0a0de7a8f2e8"
#> 
#> [[5]]$agent_name
#> [1] "ceo1"
#> 
#> [[5]]$response
#> Your combined approach of virtual tastings, influencer collaborations, and 
#> limited on-site sampling at key food fairs respects our budget while boosting 
#> brand visibility, so CONSENSUS: execute a digital marketing campaign featuring 
#> Ouergla heritage and sustainability, augmented by virtual tastings, influencer 
#> partnerships, and selective on-site sampling through German retail partners at 
#> major food events over the next two years.
#> 
#> 
#> [[6]]
#> [[6]]$iteration
#> [1] 3
#> 
#> [[6]]$agent_id
#> [1] "cd526cbe-feaa-4c2d-89bb-2343412dc215"
#> 
#> [[6]]$agent_name
#> [1] "ceo2"
#> 
#> [[6]]$response
#> CONSENSUS: execute a digital marketing campaign featuring Ouergla heritage and 
#> sustainability, augmented by virtual tastings, influencer partnerships, and 
#> selective on-site sampling through German retail partners at major food events 
#> over the next two years.
```

If the instructions of the Agents differ fundamentally, they won’t be
able to find a consensus and the `LeadAgent` will take over and provide
a final response:

``` r
ceo1 <- Agent$new(
  name = "ceo1",
  instruction = paste0(
    "You are a CEO in  a dates company based in Ouergla, Algeria, ", 
    "You want to boost their exports to Germany. ", 
    "You don't care about the budget. You want to spend as much as possible. "
  ),
  llm_object = openai_4_1_mini
)

ceo2 <- Agent$new(
  name = "ceo2",
  instruction = paste0(
    "You are the CEO of a dates company based in Ouergla, Algeria. ", 
    "You are considering starting a marketing compaign to boost the exports to Germany. ",
    "For you the marketing budget is super important and you don't want to spend too much. "
  ),
  llm_object = openai_4_1_mini
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(ceo1, ceo2))
#> ✔ Agent(s) successfully registered.

result <- lead_agent$agents_dialog(
  prompt = "Propose a plan in 1 sentence max about a marketing strategy that will boost the export of dates to Germany for the next 2 years",
  agent_1_id = ceo1$agent_id,
  agent_2_id = ceo2$agent_id,
  max_iterations = 3
)
#> 
#> ── Starting agent dialog ──
#> 
#> Agent 1: ceo1
#> Agent 2: ceo2
#> Max iterations: 3
#> ℹ Iteration 1 - ceo1 responding...
#> ℹ Iteration 1 - ceo2 responding...
#> ℹ Iteration 2 - ceo1 responding...
#> ℹ Iteration 2 - ceo2 responding...
#> ℹ Iteration 3 - ceo1 responding...
#> ℹ Iteration 3 - ceo2 responding...
#> ! Max iterations reached without explicit consensus.
#> ℹ Using lead agent to synthesize final response...
#> ✔ Dialog completed.

# Access the final response
result$final_response
#> Develop a two-year marketing strategy combining targeted digital campaigns and 
#> influencer partnerships to build brand awareness cost-effectively, while 
#> allocating a moderate budget for selective premium event sponsorships and 
#> upscale retail placements to strategically enhance the export of Ouergla dates 
#> to Germany.

# View the dialog history
result$dialog_history
#> [[1]]
#> [[1]]$iteration
#> [1] 1
#> 
#> [[1]]$agent_id
#> [1] "e75ee46c-ada5-461f-9765-70e1744e58b3"
#> 
#> [[1]]$agent_name
#> [1] "ceo1"
#> 
#> [[1]]$response
#> Launch a high-end, culturally immersive marketing campaign across Germany 
#> featuring premium Algerian dates in luxury stores and gourmet festivals, 
#> combined with extensive influencer partnerships, targeted digital ads, and 
#> exclusive tasting events to position Ouergla dates as a must-have delicacy.
#> 
#> 
#> [[2]]
#> [[2]]$iteration
#> [1] 1
#> 
#> [[2]]$agent_id
#> [1] "992da0c0-f3c7-4c7d-9763-b571e729c6df"
#> 
#> [[2]]$agent_name
#> [1] "ceo2"
#> 
#> [[2]]$response
#> While the proposed high-end luxury campaign could elevate our brand, it risks 
#> exceeding our tight budget constraints; instead, I suggest leveraging targeted 
#> social media marketing with authentic stories about Ouergla dates, partnering 
#> with German specialty food retailers for product placement, and organizing 
#> low-cost tasting events to build awareness and demand over the next two years 
#> without heavy expenses.
#> 
#> 
#> [[3]]
#> [[3]]$iteration
#> [1] 2
#> 
#> [[3]]$agent_id
#> [1] "e75ee46c-ada5-461f-9765-70e1744e58b3"
#> 
#> [[3]]$agent_name
#> [1] "ceo1"
#> 
#> [[3]]$response
#> Their proposal focuses on budget-conscious tactics, but since I explicitly want
#> to spend as much as possible without budget limits, their cost-saving approach 
#> conflicts with my directive to invest heavily for maximum impact. Therefore, I 
#> propose instead to implement an extravagant, multi-channel marketing blitz 
#> including large-scale sponsorships of major German food events, premium 
#> in-store displays in top retail chains, celebrity chef partnerships showcasing 
#> Ouergla dates in exclusive recipes, and a high-profile media campaign across 
#> TV, print, and digital platforms for two full years to firmly establish our 
#> brand presence and dominate the German market.
#> 
#> 
#> [[4]]
#> [[4]]$iteration
#> [1] 2
#> 
#> [[4]]$agent_id
#> [1] "992da0c0-f3c7-4c7d-9763-b571e729c6df"
#> 
#> [[4]]$agent_name
#> [1] "ceo2"
#> 
#> [[4]]$response
#> Their proposal directly conflicts with my core requirement to minimize 
#> marketing expenditure, as it suggests an extravagant, high-cost campaign that 
#> disregards budget constraints; therefore, I counterpropose a highly focused, 
#> low-cost digital and grassroots campaign using targeted social media ads, 
#> collaborations with niche German health and gourmet influencers, and strategic 
#> participation in affordable local food fairs to cost-effectively boost exports 
#> over two years.
#> 
#> 
#> [[5]]
#> [[5]]$iteration
#> [1] 3
#> 
#> [[5]]$agent_id
#> [1] "e75ee46c-ada5-461f-9765-70e1744e58b3"
#> 
#> [[5]]$agent_name
#> [1] "ceo1"
#> 
#> [[5]]$response
#> Their proposal emphasizes minimizing marketing expenditure, which directly 
#> contradicts my role's explicit directive to spend as much as possible with no 
#> budget constraints; therefore, I reaffirm that the optimal plan is to execute a
#> lavish, comprehensive marketing campaign across Germany—including large-scale 
#> sponsorships of premium food events, celebrity chef endorsements, luxury retail
#> placements, and a multi-channel media blitz over two years—to maximize brand 
#> visibility and export growth without any financial limitations.
#> 
#> 
#> [[6]]
#> [[6]]$iteration
#> [1] 3
#> 
#> [[6]]$agent_id
#> [1] "992da0c0-f3c7-4c7d-9763-b571e729c6df"
#> 
#> [[6]]$agent_name
#> [1] "ceo2"
#> 
#> [[6]]$response
#> Your proposal to spend lavishly on marketing campaigns directly conflicts with 
#> my primary role as CEO with a strict focus on minimizing expenditures; 
#> therefore, I must reiterate that a cost-efficient marketing strategy leveraging
#> digital channels, targeted influencer collaborations, and participation in 
#> affordable German food markets is the most suitable approach to boost exports 
#> while respecting our limited budget.
```

## Code of Conduct

Please note that the mini007 project is released with a [Contributor
Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

#' LeadAgent: A Multi-Agent Orchestration Coordinator
#'
#' @description
#' `LeadAgent` extends `Agent` to coordinate a group of specialized agents.
#' It decomposes complex prompts into subtasks using LLMs and assigns each subtask to the most suitable registered agent.
#' The lead agent handles response chaining, where each agent can consider prior results.
#'
#' @details
#' This class builds intelligent multi-agent workflows by delegating sub-tasks using `delegate_prompt()`,
#' executing them with `invoke()`, and storing the results in the `agents_interaction` list.
#' @importFrom DiagrammeR grViz
#' @export
LeadAgent <- R6::R6Class(
  classname = "LeadAgent",
  inherit = Agent,

  public = list(

    #' @field agents A named list of registered sub-agents (by UUID).
    agents = list(),
    #' @field agents_interaction A list of delegated task history with agent IDs, prompts, and responses.
    agents_interaction = list(),
    #' @field plan A list containing the most recently generated task plan.
    plan = list(),
    #' @field hitl_steps The steps where the workflow should be stopped in order to allow for a human interaction
    hitl_steps = NULL,
    #'@field prompt_for_plan The prompt used to generate the plan.
    prompt_for_plan = NULL,
    #'@field agents_for_plan The agents used for the plan
    agents_for_plan = NULL,
    #' @field dialog_history A list storing the history of agent dialogs
    dialog_history = list(),
    #' @field broadcast_history A list storing the history of broadcast interactions
    broadcast_history = list(),

    #' @description
    #' Initializes the LeadAgent with a built-in task-decomposition prompt.
    #' @param name A short name for the coordinator (e.g. `"lead"`).
    #' @param llm_object The LLM object generate by ellmer (eg. output of ellmer::chat_openai)
    #' @examples
    #'
    #'   # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'
    initialize = function(name, llm_object) {

      system_prompt <- paste0(
        "You are a task decomposition assistant. ",
        "You receive a complex user instruction and return a list of smaller subtasks to be performed in logical order. ",
        "Return the subtasks as plain text, one per line.",
        "Within the subtasks, mentions the necessary information that should be known before executing them. ",
        "Write the subtask as it is. Do not include numerotation (like 1., 2. and so on), write the subtask plain and simple"
      )

      super$initialize(name = name, instruction = system_prompt, llm_object = llm_object)

    },

    #' @description
    #' Clear out the registered Agents
    #' @examples
    #'   # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'    "You are a research assistant. ",
    #'    "Your job is to answer factual questions with detailed and accurate information. ",
    #'    "Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = paste0(
    #'    "You are an agent designed to summarise ",
    #'    "a given text into 3 distinct bullet points."
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  lead_agent$agents
    #'
    #'  lead_agent$clear_agents()
    #'
    #'  lead_agent$agents
    #'

    clear_agents = function() {
      self$agents <- list()
    },

    #' @description
    #' Remove registered agents by IDs
    #' @param agent_ids The Agent ID to remove from the registered Agents
    #' @examples
    #'   # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'    "You are a research assistant. ",
    #'    "Your job is to answer factual questions with detailed and accurate information. ",
    #'    "Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  lead_agent$agents
    #'
    #'  # deleting the translator agent
    #'
    #'  id_translator_agent <- translator$agent_id
    #'
    #'  lead_agent$remove_agents(id_translator_agent)
    #'
    #'  lead_agent$agents
    #'
    remove_agents = function(agent_ids) {

      checkmate::assert_character(agent_ids, any.missing = FALSE)
      self$agents <- Filter(function(a) !(a$agent_id %in% agent_ids), self$agents)

    },

    #' @description
    #' Register one or more agents for delegation.
    #' @param agents A vector of `Agent` objects to register.
    #' @examples
    #'   # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'    "You are a research assistant. ",
    #'    "Your job is to answer factual questions with detailed and accurate information. ",
    #'    "Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  lead_agent$agents
    register_agents = function(agents) {

      length_agents <- length(self$agents)

      for (i in seq_along(agents)) {
        agent <- agents[[i]]
        self$agents[[length_agents + i]] <- agent
      }

      cli::cli_alert_success("Agent(s) successfully registered.")

      return(invisible(NULL))

    },

    #' @description
    #' Visualizes the orchestration plan
    #' Each agent node is shown in sequence (left → right), with tooltips showing
    #' the actual prompt delegated to that agent.
    visualize_plan = function() {

      plan <- self$plan

      if (length(plan) == 0) {
        cli::cli_abort("No plan found, generate one first")
      }

      nodes <- paste0(
        "A", seq_along(plan),
        " [label='", vapply(plan, `[[`, "", "agent_name"),
        "', tooltip='", gsub("'", "\\\\'", vapply(plan, `[[`, "", "prompt")),
        "', shape=box, style=filled, fillcolor=lightblue, fontname='Helvetica'];"
      )

      edges <- paste0(
        "A", seq_len(length(plan) - 1),
        " -> A", seq_len(length(plan) - 1) + 1,
        " [arrowhead=normal];"
      )

      graph_code <- paste0(
        "digraph workflow {
      graph [rankdir=LR, splines=true, overlap=false];
      node [shape=box, style=filled, fillcolor=lightblue, fontname='Helvetica', fontsize=12];
      edge [color=gray50, arrowsize=0.8];
      ", paste(c(nodes, edges), collapse = "\n"), "
    }"
      )

      DiagrammeR::grViz(graph_code)
    },

    #' @description
    #' Executes the full prompt pipeline: decomposition → delegation → invocation.
    #' @param prompt The complex user instruction to process.
    #' @param force_regenerate_plan If TRUE, regenerate a plan even if one exists,
    #' defaults to FALSE.
    #' @return The final response (from the last agent in the sequence).
    #' @examples \dontrun{
    #'  # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'    "You are a research assistant. ",
    #'    "Your job is to answer factual questions with detailed ",
    #'    "and accurate information. Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  lead_agent$invoke(
    #'  paste0(
    #'   "Describe the economic situation in Algeria in 3 sentences. ",
    #'   "Answer in German"
    #'   )
    #'  )
    #' }

    invoke = function(prompt, force_regenerate_plan = FALSE) {

      checkmate::assert_character(prompt)
      checkmate::assert_flag(force_regenerate_plan)

      if (length(self$agents) == 0) {

        err_msg <- paste0(
          "No Agent has been assigned yet to the LeadAgent. ",
          "Use the `registre_agents` methods to assign Agents"
        )

        cli::cli_abort(err_msg)
      }

      prompt_invoke_same_as_plan <- FALSE


      if (!is.null(self$prompt_for_plan) && self$prompt_for_plan == prompt) {
        prompt_invoke_same_as_plan <- TRUE
      }

      current_agents <- lapply(self$agents, function(x) {x$name})
      current_agents <- unlist(current_agents)

      same_agents_used <- all(sort(self$agents_for_plan) == sort(current_agents))

      if (is.null(self$agents_for_plan) || !same_agents_used) {
        prompt_invoke_same_as_plan <- FALSE
      }

      if (!prompt_invoke_same_as_plan || length(self$plan) == 0 || force_regenerate_plan) {
        cli::cli_h2("Generating new plan")
        prompts_res <- self$generate_plan(prompt)
      } else {
        cli::cli_h2("Using existing plan")
        prompts_res <- self$plan
      }

      for (i in seq_along(prompts_res)) {
        prompts_res[[i]]$response <- NA
      }

      self$agents_interaction <- prompts_res

      for (i in seq_along(self$agents_interaction)) {
        step <- self$agents_interaction[[i]]

        idx <- which(sapply(self$agents, function(agent) agent$agent_id == step$agent_id))
        selected_agent <- self$agents[[idx]]

        prompt_to_consider <- step$prompt
        if (i > 1) {
          prev_resp <- self$agents_interaction[[i - 1]]$response
          prompt_to_consider <- paste(
            "\n\nBefore answering consider the previous response:\n", prev_resp,
            "\n\n--- Task ---\n", prompt_to_consider
          )
        }

        response <- selected_agent$invoke(prompt_to_consider)
        step$response <- response
        step$edited_by_hitl <- FALSE
        self$agents_interaction[[i]] <- step

        if (i %in% self$hitl_steps && !is.null(self$hitl_steps)) {
          private$.human_confirm(i)
        }
      }

      self$agents_interaction[[length(self$agents_interaction)]]$response

    },

    #' @description
    #' Generates a task execution plan without executing the subtasks.
    #' It returns a structured list containing the subtask, the selected agent, and metadata.
    #' @param prompt A complex instruction to be broken into subtasks.
    #' @return A list of lists containing agent_id, agent_name, model_name, model_provider, and the assigned prompt.
    #' @examples \dontrun{
    #'  # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'    "You are a research assistant. Your job is to answer factual questions ",
    #'    "with detailed and accurate information. Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  lead_agent$generate_plan(
    #'  paste0(
    #'   "Describe the economic situation in Algeria in 3 sentences. ",
    #'   "Answer in German"
    #'   )
    #'  )
    #' }
    generate_plan = function(prompt) {

      if (length(self$agents) == 0) {
        cli::cli_abort("No agents registered. Use `register_agents()` to add sub-agents before generating a plan.")
      }

      subtasks <- private$.analyze_prompt(prompt)

      plan <- lapply(subtasks, function(task) {
        agent_id <- private$.match_agent_to_task(task)

        idx <- which(sapply(self$agents, function(agent) agent$agent_id == agent_id))
        selected_agent <- self$agents[[idx]]

        list(
          agent_id = agent_id,
          agent_name = selected_agent$name,
          model_provider = selected_agent$model_provider,
          model_name = selected_agent$model_name,
          prompt = task
        )
      })

      cli::cli_alert_success("Plan successfully generated.")

      self$plan <- plan

      self$prompt_for_plan <- prompt

      agents_names <- lapply(self$agents, function(x) {x$name})

      agents_names <- unlist(agents_names)

      self$agents_for_plan <- agents_names

      return(plan)
    },

    #' @description
    #' Broadcasts a prompt to all registered agents and collects their responses.
    #' This does not affect the main agent orchestration logic or history.
    #' @param prompt A user prompt to send to all agents.
    #' @return A list of responses from all agents.
    #' @examples \dontrun{
    #'  # An API KEY is required in order to invoke the agents
    #' openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #' openai_4_1 <- ellmer::chat(
    #'   name = "openai/gpt-4.1",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #'
    #' openai_4_1_agent <- Agent$new(
    #'   name = "openai_4_1_agent",
    #'   instruction = "You are an AI assistant. Answer in 1 sentence max.",
    #'   llm_object = openai_4_1
    #' )
    #'
    #' openai_4_1_nano <- ellmer::chat(
    #'   name = "openai/gpt-4.1-nano",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #'
    #' openai_4_1_nano_agent <- Agent$new(
    #'   name = "openai_4_1_nano_agent",
    #'   instruction = "You are an AI assistant. Answer in 1 sentence max.",
    #'   llm_object = openai_4_1_nano
    #'   )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #' lead_agent$register_agents(c(openai_4_1_agent, openai_4_1_nano_agent))
    #' lead_agent$broadcast(
    #'   prompt = paste0(
    #'     "If I were Algerian, which song would I like to sing ",
    #'     "when running under the rain? how about a flower?"
    #'   )
    #'   )
    #' }
    broadcast = function(prompt) {
      checkmate::assert_string(prompt)

      if (length(self$agents) == 0) {
        cli::cli_abort("No agents have been registered. Use `register_agents()` first.")
      }

      responses <- lapply(self$agents, function(agent) {
        response <- agent$invoke(prompt)
        list(
          agent_id = agent$agent_id,
          agent_name = agent$name,
          model_provider = agent$model_provider,
          model_name = agent$model_name,
          response = response
        )
      })

      self$broadcast_history[[length(self$broadcast_history) + 1]] <- list(
        prompt = prompt,
        responses = responses
      )

      return(responses)
    },

    #' @description
    #' Set Human In The Loop (HITL) interaction at determined steps within the workflow
    #' @param steps At which steps the Human In The Loop is required?
    #' @return A list of responses from all agents.
    #' @examples \dontrun{
    #'  # An API KEY is required in order to invoke the agents
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'  researcher <- Agent$new(
    #'    name = "researcher",
    #'    instruction = paste0(
    #'     "You are a research assistant. ",
    #'     "Your job is to answer factual questions with detailed and accurate information. ",
    #'     "Do not answer with more than 2 lines"
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  summarizer <- Agent$new(
    #'    name = "summarizer",
    #'    instruction = paste0(
    #'    "You are agent designed to summarise a give text ",
    #'    "into 3 distinct bullet points."
    #'    ),
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  translator <- Agent$new(
    #'    name = "translator",
    #'    instruction = "Your role is to translate a text from English to German",
    #'    llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #'  )
    #'
    #'  lead_agent$register_agents(c(researcher, summarizer, translator))
    #'
    #'  # setting a human in the loop in step 2
    #'  lead_agent$set_hitl(1)
    #'
    #'  # The execution will stop at step 2 and a human will be able
    #'  # to either accept the answer, modify it or stop the execution of
    #'  # the workflow
    #'
    #'  lead_agent$invoke(
    #'  paste0(
    #'   "Describe the economic situation in Algeria in 3 sentences. ",
    #'   "Answer in German"
    #'   )
    #'  )
    #' }
    set_hitl = function(steps) {
      checkmate::assert_integerish(steps, lower = 1, any.missing = FALSE)
      self$hitl_steps <- unique(as.integer(steps))

      steps_chr <- as.character(steps)
      steps_chr <- toString(steps_chr)

      cli::cli_alert_success("HITL successfully set at step(s) {steps_chr}.")

    },

    #' @description
    #' The Lead Agent send a prompt to its registered agents and choose the best response
    #' from the agents' responses
    #' @param prompt The prompt to send to the registered agents
    #' @return A list of responses from all agents, including the chosen response
    #' @examples \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' openai_4_1 <- ellmer::chat(
    #'   name = "openai/gpt-4.1",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #'
    #' stylist <- Agent$new(
    #'   name = "stylist",
    #'   instruction = "You are an AI assistant. Answer in 1 sentence max.",
    #'   llm_object = openai_4_1
    #' )
    #'
    #' openai_4_1_nano <- ellmer::chat(
    #'   name = "openai/gpt-4.1-nano",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #'
    #' stylist2 <- Agent$new(
    #'   name = "stylist2",
    #'   instruction = "You are an AI assistant. Answer in 1 sentence max.",
    #'   llm_object = openai_4_1_nano
    #' )
    #'
    #' lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #' )
    #'
    #' lead_agent$register_agents(c(stylist, stylist2))
    #'
    #' lead_agent$judge_and_choose_best_response("what's the best way to war a kalvin klein shirt?")
    #'
    #' }
    judge_and_choose_best_response = function(prompt) {

      checkmate::assert_string(prompt)

      if (length(self$agents) == 0) {
        cli::cli_abort("No agents registered. Use `register_agents()` first.")
      }

      proposals <- lapply(self$agents, function(agent) {
        resp <- agent$invoke(prompt)
        list(
          agent_id = agent$agent_id,
          agent_name = agent$name,
          response = resp
        )
      })

      responses_summary <- paste(
        vapply(proposals, function(x) {x$response}, character(1)),
        collapse = "\n------------\n"
      )

      judge_prompt <- glue::glue(
        "The following prompt was previously executed: {prompt}. ",
        "We got back the following responses from several agents: ",
        "Each response is separated by a ------------",
        "\nResponses:\n\n",
        "<<<<",
        responses_summary,
        ">>>>",
        "\n",
        "\nChoose the best response according to initial prompt.",
        "\nReturn ONLY the final response text. Nothing else. Do not talk. Just return the best response"
      )

      result <- self$llm_object$chat(judge_prompt)

      final_result <- list(
        proposals = proposals,
        chosen_response = result
      )

      return(final_result)

    },

    #' @description
    #' Facilitates a collaborative dialog between two agents to refine a response.
    #' The agents take turns building on each other's responses until they reach
    #' consensus or the maximum iterations are reached. Agents can signal consensus
    #' by starting their response with "CONSENSUS:". If max iterations is reached
    #' without consensus, the lead agent synthesizes a final response.
    #' @param prompt The initial task or question for the agents to discuss.
    #' @param agent_1_id The ID of the first agent to participate in the dialog.
    #' @param agent_2_id The ID of the second agent to participate in the dialog.
    #' @param max_iterations Maximum number of back-and-forth exchanges (default: 5).
    #' @return A list containing the final response, consensus status, and complete dialog history.
    #' @examples \dontrun{
    #' # An API KEY is required in order to invoke the agents
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' openai_4_1_nano <- ellmer::chat(
    #'   name = "openai/gpt-4.1-nano",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #'
    #' creative_writer <- Agent$new(
    #'   name = "creative_writer",
    #'   instruction = "You are a creative writer. Focus on engaging storytelling.",
    #'   llm_object = openai_4_1_nano
    #' )
    #'
    #' editor <- Agent$new(
    #'   name = "editor",
    #'   instruction = "You are an editor. Focus on clarity and conciseness.",
    #'   llm_object = openai_4_1_mini
    #' )
    #'
    #' lead_agent <- LeadAgent$new(
    #'   name = "Leader",
    #'   llm_object = openai_4_1_mini
    #' )
    #'
    #' lead_agent$register_agents(c(creative_writer, editor))
    #'
    #' result <- lead_agent$agents_dialog(
    #'   prompt = "Write a compelling opening sentence for a sci-fi novel.",
    #'   agent_1_id = creative_writer$agent_id,
    #'   agent_2_id = editor$agent_id,
    #'   max_iterations = 3
    #' )
    #'
    #' # Access the final response
    #' result$final_response
    #'
    #' # View the dialog history
    #' result$dialog_history
    #' }
    agents_dialog = function(prompt, agent_1_id, agent_2_id, max_iterations = 5) {

      checkmate::assert_string(prompt)
      checkmate::assert_string(agent_1_id)
      checkmate::assert_string(agent_2_id)
      checkmate::assert_int(max_iterations, lower = 1)

      if (length(self$agents) == 0) {
        cli::cli_abort("No agents registered. Use `register_agents()` first.")
      }

      if (agent_1_id == agent_2_id) {
        cli::cli_abort("Agent 1 and Agent 2 must be different agents.")
      }

      agent_ids <- vapply(self$agents, function(a) a$agent_id, character(1))

      if (!agent_1_id %in% agent_ids) {
        cli::cli_abort("Agent 1 with ID '{agent_1_id}' not found in registered agents.")
      }

      if (!agent_2_id %in% agent_ids) {
        cli::cli_abort("Agent 2 with ID '{agent_2_id}' not found in registered agents.")
      }

      idx_1 <- which(agent_ids == agent_1_id)
      idx_2 <- which(agent_ids == agent_2_id)

      agent_1 <- self$agents[[idx_1]]
      agent_2 <- self$agents[[idx_2]]

      cli::cli_h2("Starting agent dialog")
      cli::cli_text("Agent 1: {.strong {agent_1$name}}")
      cli::cli_text("Agent 2: {.strong {agent_2$name}}")
      cli::cli_text("Max iterations: {max_iterations}")

      dialog_history <- list()
      current_response <- NULL
      consensus_reached <- FALSE
      final_response <- NULL

      for (i in seq_len(max_iterations)) {

        cli::cli_alert_info("Iteration {i} - {agent_1$name} responding...")

        if (is.null(current_response)) {
          prompt_1 <- prompt
        } else {
          prompt_1 <- paste0(
            "REMEMBER YOUR ROLE: ", agent_1$instruction, "\n\n",
            "You are in a collaborative negotiation to solve this task: ", prompt, "\n\n",
            "The other agent just proposed:\n",
            current_response, "\n\n",
            "CRITICAL: Evaluate their proposal against YOUR specific role, requirements, and constraints described above. ",
            "Does it fully align with YOUR requirements? Does it respect YOUR constraints? ",
            "If there are ANY conflicts with your role, concerns about their approach, or areas where your perspective differs, ",
            "you MUST propose your counterproposal that better addresses YOUR requirements. ",
            "Do NOT simply accept their proposal if it conflicts with your role or constraints. ",
            "Only respond with 'CONSENSUS: ' followed by the final solution if their proposal genuinely ",
            "satisfies BOTH your requirements AND theirs, with no compromises to your constraints. ",
            "Otherwise, continue negotiating."
          )
        }

        response_1 <- agent_1$invoke(prompt_1)

        dialog_history[[length(dialog_history) + 1]] <- list(
          iteration = i,
          agent_id = agent_1$agent_id,
          agent_name = agent_1$name,
          response = response_1
        )

        if (grepl("^CONSENSUS:", response_1, ignore.case = TRUE)) {
          cli::cli_alert_success("Consensus reached by {agent_1$name} at iteration {i}!")
          consensus_reached <- TRUE
          final_response <- sub("^CONSENSUS:\\s*", "", response_1, ignore.case = TRUE)
          break
        }

        current_response <- response_1

        cli::cli_alert_info("Iteration {i} - {agent_2$name} responding...")

        prompt_2 <- paste0(
          "REMEMBER YOUR ROLE: ", agent_2$instruction, "\n\n",
          "You are in a collaborative negotiation to solve this task: ", prompt, "\n\n",
          "The other agent just proposed:\n",
          current_response, "\n\n",
          "CRITICAL: Evaluate their proposal against YOUR specific role, requirements, and constraints described above. ",
          "Does it fully align with YOUR requirements? Does it respect YOUR constraints? ",
          "If there are ANY conflicts with your role, concerns about their approach, or areas where your perspective differs, ",
          "you MUST propose your counterproposal that better addresses YOUR requirements. ",
          "Do NOT simply accept their proposal if it conflicts with your role or constraints. ",
          "Only respond with 'CONSENSUS: ' followed by the final solution if their proposal genuinely ",
          "satisfies BOTH your requirements AND theirs, with no compromises to your constraints. ",
          "Otherwise, continue negotiating."
        )

        response_2 <- agent_2$invoke(prompt_2)

        dialog_history[[length(dialog_history) + 1]] <- list(
          iteration = i,
          agent_id = agent_2$agent_id,
          agent_name = agent_2$name,
          response = response_2
        )

        if (grepl("^CONSENSUS:", response_2, ignore.case = TRUE)) {
          cli::cli_alert_success("Consensus reached by {agent_2$name} at iteration {i}!")
          consensus_reached <- TRUE
          final_response <- sub("^CONSENSUS:\\s*", "", response_2, ignore.case = TRUE)
          break
        }

        current_response <- response_2
      }

      if (!consensus_reached) {
        cli::cli_alert_warning("Max iterations reached without explicit consensus.")
        cli::cli_alert_info("Using lead agent to synthesize final response...")

        history_text <- paste(
          vapply(dialog_history, function(h) {
            paste0("[", h$agent_name, "]: ", h$response)
          }, character(1)),
          collapse = "\n\n"
        )

        synthesis_prompt <- paste0(
          "Original task: ", prompt, "\n\n",
          "Two agents have discussed this task. Here is their conversation:\n\n",
          history_text, "\n\n",
          "Based on their discussion, synthesize and provide the best final answer to the original task. ",
          "CRITICAL: Return ONLY the final answer itself, as if you were directly answering the original task. ",
          "DO NOT write instructions, DO NOT mention the agents or their discussion, ",
          "DO NOT use phrases like 'Write...', 'The final answer is...', or 'Based on...'. ",
          "Just provide the direct answer to the task."
        )

        final_response <- self$llm_object$chat(synthesis_prompt)
      }

      result <- list(
        prompt = prompt,
        agent_1_name = agent_1$name,
        agent_2_name = agent_2$name,
        consensus_reached = consensus_reached,
        iterations_used = length(dialog_history),
        dialog_history = dialog_history,
        final_response = final_response
      )

      self$dialog_history[[length(self$dialog_history) + 1]] <- result

      cli::cli_alert_success("Dialog completed.")

      return(result)

    }
  ),

  private = list(

    .analyze_prompt = function(prompt) {

      result <- self$llm_object$chat(prompt)

      tasks <- unlist(strsplit(result, "\n"))
      tasks <- trimws(tasks)
      tasks <- tasks[tasks != ""]

      return(tasks)
    },

    .match_agent_to_task = function(task) {

      agent_descriptions <- vapply(self$agents, function(agent) {
        paste0(
          "----------------",
          "\nID: ", agent$agent_id,
          "\nName: ", agent$name,
          "\nInstruction: ", agent$instruction,
          "\n---------------"
        )
      }, character(1))

      system_prompt <- paste0(
        "You are an agent-matching assistant. ",
        "Given a task and a list of available agents with their instructions, names and IDs, ",
        "pick the best-suited agent (only one) and return ONLY its ID (no explanation). The ID has a UUID structure",
        " The structure of the Agent starts with the ID and finishes with the Instruction."
      )

      base_system_prompt <- self$llm_object$get_system_prompt()

      self$llm_object$set_system_prompt(value = system_prompt)

      user_message <- paste(
        "Task:", task, "\n\n",
        "Available agents:\n",
        paste(agent_descriptions, collapse = "\n\n")
      )

      agent_id <- self$llm_object$chat(
        user_message
      )

      checkmate::assert_string(agent_id)

      agent_id <- trimws(agent_id)

      agent_ids <- unlist(lapply(self$agents, function(agent) agent$agent_id))

      if (!agent_id %in% agent_ids) {
        cli::cli_abort("LLM returned invalid agent_id: ", agent_id)
      }

      self$llm_object$set_system_prompt(value = base_system_prompt)

      return(agent_id)

    },

    .human_confirm = function(step_index) {

      step <- self$agents_interaction[[step_index]]

      cli::cli_rule(left = "HITL Step {step_index}")
      cli::cli_text("Agent: {.strong {step$agent_name}}")
      cli::cli_alert_info("Prompt:")
      cli::cli_text("{.italic {step$prompt}}")
      cli::cli_alert_info("Response:")
      cli::cli_text("{.italic {step$response}}")

      cli::cli_text(cli::cli_ul(c(
        "[1] Continue with this response ---",
        "[2] Edit the response ---",
        "[3] Stop the workflow ---"
      )))

      repeat {
        answer <- readline("Your choice [1/2/3]: ")
        if (nzchar(answer) && answer %in% c("1", "2", "3")) break
        cli::cli_alert_warning("Invalid input. Please enter 1, 2, or 3.")
      }

      if (answer == "2") {
        new_response <- readline("(->) Enter edited response: ")
        step$response <- new_response
        step$edited_by_hitl <- TRUE
        cli::cli_alert_success("Response updated.")
      } else if (answer == "3") {
        cli::cli_alert_danger("Workflow manually stopped at step {step_index}.")
        cli::cli_abort("HITL: Execution stopped by user.")
      } else {
        cli::cli_alert_success("Continuing with original response.")
      }

      self$agents_interaction[[step_index]] <- step
    }
  )
)

#' Agent: A General-Purpose LLM Agent
#'
#' @description
#' The `Agent` class defines a modular LLM-based agent capable of responding to prompts using a defined role/instruction.
#' It wraps an OpenAI-compatible chat model via the [`ellmer`](https://github.com/llrs/ellmer) package.
#'
#' Each agent maintains its own message history and unique identity.
#'
#' @importFrom R6 R6Class
#' @importFrom uuid UUIDgenerate
#' @importFrom checkmate assert_string assert_flag assert_character assert_integerish assert_environment
#' @importFrom cli cli_abort cli_alert_success cli_alert_warning cli_alert_info cli_rule cli_text cli_ul
#' @importFrom glue glue
#' @importFrom ellmer type_object type_string type_number
#' @importFrom rlang parse_expr eval_bare caller_env

#' @export
Agent <- R6::R6Class(
  classname = "Agent",

  public = list(
    #' @description
    #' Initializes a new Agent with a specific role/instruction.
    #'
    #' @param name A short identifier for the agent (e.g. `"translator"`).
    #' @param instruction The system prompt that defines the agent's role.
    #' @param llm_object The LLM object generate by ellmer (eg. output of ellmer::chat_openai)
    #' @param budget Numerical value denoting the amount to set for the budget in US$ to a specific agent,
    #' if the budget is reached, an error will be thrown.
    #' @examples
    #'   # An API KEY is required in order to invoke the Agent
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'
    #'   polar_bear_researcher <- Agent$new(
    #'     name = "POLAR BEAR RESEARCHER",
    #'     instruction = paste0(
    #'     "You are an expert in polar bears, ",
    #'     "you task is to collect information about polar bears. Answer in 1 sentence max."
    #'     ),
    #'     llm_object = openai_4_1_mini
    #'   )
    #'
    initialize = function(name, instruction, llm_object, budget = NA) {
      checkmate::assert_string(name)
      checkmate::assert_string(instruction)

      if (!"Chat" %in% class(llm_object)) {
        cli::cli_abort("The llm_object must be generated by ellmer")
      }

      self$name <- name
      self$instruction <- instruction
      self$llm_object <- llm_object$clone(deep = TRUE)

      meta_data <- self$llm_object$get_provider()
      self$model_provider <- meta_data@name
      self$model_name <- meta_data@model

      self$llm_object$set_system_prompt(value = instruction)

      private$._messages <- list(
        list(role = "system", content = instruction)
      )

      self$agent_id <- uuid::UUIDgenerate()

      checkmate::assert_numeric(budget)

      self$budget <- budget

      self$budget_policy <- list(
        on_exceed = "abort",
        warn_at = 0.8
      )

      self$cost <- NA

    },

    #' @description
    #' Sends a user prompt to the agent and returns the assistant's response.
    #'
    #' @param prompt A character string prompt for the agent to respond to.
    #' @return The LLM-generated response as a character string.
    #' @examples \dontrun{
    #' # An API KEY is required in order to invoke the Agent
    #' openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #' )
    #' agent <- Agent$new(
    #'  name = "translator",
    #'  instruction = "You are an Algerian citizen",
    #'  llm_object = openai_4_1_mini
    #' )
    #' agent$invoke("Continue this sentence: 1 2 3 viva")
    #' }
    invoke = function(prompt) {

      checkmate::assert_string(prompt)

      if (!is.na(self$budget)) {
        private$.check_budget()
      }

      response <- self$llm_object$chat(prompt)

      cost <- self$llm_object$get_cost()

      if (!is.na(cost)) {
        self$cost <- as.numeric(round(cost, 4))
      }

      private$.set_messages_from_turns()

      return(response)
    },

    #' @description
    #' Generate R code from natural language descriptions and optionally validate/execute it
    #'
    #' @param code_description Character string describing the R code to generate
    #' @param validate Logical indicating whether to validate the generated code syntax
    #' @param execute Logical indicating whether to execute the generated code (use with caution)
    #' @param interactive Logical; if TRUE, ask for user confirmation before executing generated code
    #' @param env Environment in which to execute the code if execute = TRUE. Default to \code{globalenv}
    #' @return A list containing the generated code and validation/execution results
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' r_assistant <- Agent$new(
    #'   name = "R Code Assistant",
    #'   instruction = paste("You are an expert R programmer",
    #'   llm_object = openai_4_1_mini
    #' )
    #' # Generate code for data manipulation
    #' result <- r_assistant$generate_execute_r_code(
    #'   code_description = "Calculate the summary of the mtcars dataframe",
    #'   validate = TRUE,
    #'   execute = TRUE,
    #'   interactive = TRUE
    #' )
    #' print(result)
    #' }
    generate_execute_r_code = function(
    code_description,
    validate = FALSE,
    execute = FALSE,
    interactive = TRUE,
    env = globalenv()
    ) {

      checkmate::assert_string(code_description)
      checkmate::assert_flag(validate)
      checkmate::assert_flag(execute)
      checkmate::assert_environment(env)
      checkmate::assert_flag(interactive)

      code_prompt <- paste0(
        "Generate R code for the following task. Return ONLY the R code that will be executed without any explanations, ",
        "markdown formatting, or additional text:\n\n",
        "if you run several commands, use the ';' character to separate them. \n\n",
        "DO NOT add additional spaces that would be interpreted later as '\\n'. \n\n",
        "DO NOT enclose the code using Markdown syntax like R`\n\n",
        code_description
      )

      generated_code <- self$invoke(code_prompt)

      clean_code <- gsub("```\\{?r\\}?|```", "", generated_code)
      clean_code <- gsub("```\\{?r\\}?|```", "", generated_code)
      clean_code <- trimws(clean_code)

      result <- list(
        description = code_description,
        code = clean_code,
        validated = FALSE,
        validation_message = NA_character_,
        executed = FALSE,
        execution_result = NULL,
        execution_error = NULL
      )

      if (validate) {
        result <- private$.validate_r_code(r_code = clean_code, result = result)
      }

      if (execute) {
        if (!validate || !result$validated) {
          cli::cli_alert_warning("Code execution skipped: code must be validated first")
          return(result)
        }

        if (interactive) {
          cli::cli_h1("Generated code preview:")
          cat(paste0("\n", clean_code, "\n\n"))
          user_input <- readline(prompt = "Do you want to execute this code? [y/N]: ")
          if (tolower(user_input) != "y") {
            cli::cli_alert_info("Execution cancelled by user.")
            return(result)
          }
        }

        cli::cli_alert_info("Executing generated R code...")

        execution_result <- tryCatch({
          output <- capture.output({
            eval_result <- eval(parse(text = clean_code), envir = env)
          })

          result$executed <- TRUE
          result$execution_result <- list(
            value = eval_result,
            output = output
          )
          result$execution_error <- NULL

          cli::cli_alert_success("Code executed successfully")

        }, error = function(e) {
          result$executed <- FALSE
          result$execution_error <- e$message
          cli::cli_alert_danger(paste("Execution error:", e$message))
        })
      }

      return(result)
    },

    #' @description
    #' Set a budget to a specific agent, if the budget is reached, an error will be thrown
    #'
    #' @param amount_in_usd Numerical value denoting the amount to set for the budget,
    #' @examples \dontrun{
    #' # An API KEY is required in order to invoke the Agent
    #' openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #' )
    #' agent <- Agent$new(
    #'  name = "translator",
    #'  instruction = "You are an Algerian citizen",
    #'  llm_object = openai_4_1_mini
    #' )
    #' agent$set_budget(amount_in_usd = 10.5) # this is equivalent to 10.5$
    #' }
    set_budget = function(amount_in_usd) {

      checkmate::assert_number(amount_in_usd, lower = 0)

      self$budget <- amount_in_usd

      cli::cli_alert_success(glue::glue("Budget successfully set to {amount_in_usd}$"))
      cli::cli_alert_info(glue::glue("Budget policy: on_exceed='{self$budget_policy$on_exceed}', warn_at={self$budget_policy$warn_at}"))
      cli::cli_alert_info("Use the set_budget_policy() method to configure the budget policy.")
      invisible(self)
    },

    #' @description
    #' Configure how the agent behaves as it approaches or exceeds its budget.
    #' Use `warn_at` (0-1) to emit a one-time warning when spending reaches the
    #' specified fraction of the budget. When the budget is exceeded, `on_exceed`
    #' controls behavior: abort, warn and proceed, or ask interactively.
    #' @param on_exceed One of "abort", "warn", or "ask".
    #' @param warn_at Numeric in (0,1); fraction of budget to warn at. Default 0.8.
    #' @examples \dontrun{
    #' agent$set_budget(5)
    #' agent$set_budget_policy(on_exceed = "ask", warn_at = 0.9)
    #' }
    set_budget_policy = function(on_exceed = "abort", warn_at = 0.8) {

      checkmate::assert_choice(on_exceed, c("abort", "warn", "ask"))
      checkmate::assert_number(warn_at, lower = 0, upper = 1)

      self$budget_policy <- list(
        on_exceed = on_exceed,
        warn_at = warn_at
      )

      cli::cli_alert_success(glue::glue(
        "Budget policy set: on_exceed='{on_exceed}', warn_at={warn_at}"
      ))

      invisible(self)
    },

    #' @description
    #' Keep only the most recent `n` messages, discarding older ones while keeping
    #' the system prompt.
    #' @param n Number of most recent messages to keep.
    #' @examples \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "capital finder",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' agent$invoke("What is the capital of Algeria")
    #' agent$invoke("What is the capital of Germany")
    #' agent$invoke("What is the capital of Italy")
    #' agent$keep_last_n_messages(n = 2)
    #' }
    keep_last_n_messages = function(n = 2) {

      checkmate::assert_integerish(n, lower = 1)

      ln_messags <- length(self$messages)

      messages_to_keep <- self$messages[(ln_messags - n + 1):ln_messags]

      system_prompt <- self$llm_object$get_system_prompt()

      tmp_sp <- list(
        list(role = "system", content = system_prompt)
      )

      private$._messages <- append(tmp_sp, messages_to_keep)
      private$.set_turns_from_messages()

      cli::cli_alert_success("Conversation truncated to last {n} messages.")

      invisible(self)

    },

    #' @description
    #' Add a pre-formatted message to the conversation history
    #'
    #' @param role The role of the message ("user", "assistant", or "system")
    #' @param content The content of the message
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "AI assistant",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #')
    #' agent$add_message("user", "Hello, how are you?")
    #' agent$add_message("assistant", "I'm doing well, thank you!")
    #' }
    add_message = function(role, content) {
      checkmate::assert_string(role)
      checkmate::assert_string(content)
      checkmate::assert_choice(role, c("user", "assistant", "system"))

      private$.add_message(content, role)
      private$.set_turns_from_messages()

      cli::cli_alert_success("Added {role} message: {substr(content, 1, 50)}...")
      invisible(self)
    },

    #' @description
    #' Share the most recent conversation messages with another agent.
    #'
    #' The last \code{n} non-system messages from \code{self} are appended to the
    #' history of an other \code{agent}. This is useful when
    #' one agent needs to handoff context or continue a thread in a separate
    #' agent.
    #'
    #' @param agent An object of class \code{Agent} that will receive the
    #'   messages.
    #' @param n Number of recent messages to transfer (excluding the system prompt).
    #'   Defaults to \code{5}.
    #' @examples
    #' \dontrun{
    #' llm_object <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' a1 <- Agent$new("a1", "instr", llm_object)
    #' a2 <- Agent$new("a2", "instr", llm_object)
    #' a1$invoke("Hello")
    #' a1$invoke("How are you?")
    #' a1$share_context_with(a2, n = 2)
    #' a2$messages
    #' }
    share_context_with = function(agent, n = 5) {

      if (!inherits(agent, "Agent")) {
        cli::cli_abort("`agent` must be an Agent instance")
      }

      checkmate::assert_integerish(n, lower = 1)

      total <- length(self$messages)

      if (total <= 1) {
        cli::cli_alert_info("No conversation messages to share.")
        return(invisible(agent))
      }

      start_idx <- max(2, total - n + 1)
      msgs <- self$messages[start_idx:total]

      for (m in msgs) {
        agent$add_message(m$role, m$content)
      }

      cli::cli_alert_success(
        "Shared {length(msgs)} messages with agent '{agent$name}'."
      )

    },

    #' @description
    #' Summarises the agent's conversation history into a concise form and appends it
    #' to the system prompt. Unlike `update_instruction()`, this method does not override
    #' the existing instruction but augments it with a summary for future context.
    #'
    #' After creating the summary, the method clears the conversation history and
    #' retains only the updated system prompt. This ensures that subsequent interactions
    #' start fresh but with the summary preserved as context.
    #'
    #' @examples \dontrun{
    #'   # Requires an OpenAI-compatible LLM from `ellmer`
    #'   openai_4_1_mini <- ellmer::chat(
    #'     name = "openai/gpt-4.1-mini",
    #'     api_key = Sys.getenv("OPENAI_API_KEY"),
    #'     echo = "none"
    #'   )
    #'
    #'   agent <- Agent$new(
    #'     name = "summariser",
    #'     instruction = "You are a summarising assistant",
    #'     llm_object = openai_4_1_mini
    #'   )
    #'
    #'   agent$invoke("The quick brown fox jumps over the lazy dog.")
    #'   agent$invoke("This is another example sentence.")
    #'
    #'   # Summarises and resets history
    #'   agent$summarise_messages()
    #'
    #'   # Now only the system prompt (with summary) remains
    #'   agent$messages
    #' }
    #'

    clear_and_summarise_messages = function() {

      if (length(self$messages) <= 1) {
        cli::cli_alert_info("No conversation history to summarise.")
        return(invisible(NULL))
      }

      summary_prompt <- paste0(
        "Summarise the following conversation history in a concise paragraph:\n\n",
        paste(
          vapply(self$messages, function(m) {
            paste0(m$role, ": ", m$content)
          }, character(1)),
          collapse = " \n "
        )
      )

      summary <- self$llm_object$chat(summary_prompt)
      summary <- as.character(summary)

      new_system_prompt <- paste(
        self$instruction,
        "\n\n--- Conversation Summary ---\n",
        summary
      )

      self$llm_object$set_system_prompt(value = new_system_prompt)

      private$._messages <- list(
        list(role = "system", content = new_system_prompt)
      )
      private$.set_turns_from_messages()

      cli::cli_alert_success("Conversation history summarised and appended to system prompt.")
      cli::cli_alert_info("Summary: {substr(summary, 1, 100)}...")

      invisible(self)
    },

    #' @description
    #' Update the system prompt/instruction
    #' @param new_instruction New instruction to use. Not that the new instruction
    #' will override the old one
    #' @examples \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "assistant",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' agent$update_instruction("You are a concise assistant.")
    #' }
    update_instruction = function(new_instruction) {

      checkmate::assert_string(new_instruction)

      old_instruction <- self$instruction
      self$instruction <- new_instruction
      self$llm_object$set_system_prompt(value = new_instruction)

      private$._messages[[1]]$content <- new_instruction
      private$.set_turns_from_messages()

      cli::cli_alert_success("Instruction successfully updated")
      cli::cli_alert_info("Old: {substr(old_instruction, 1, 50)}...")
      cli::cli_alert_info("New: {substr(new_instruction, 1, 50)}...")

      invisible(self)
    },

    #' @description
    #' Get the current token count and estimated cost of the conversation
    #'
    #' @return A list with token counts and cost information
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "assistant",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' agent$set_budget(1)
    #' agent$invoke("What is the capital of Algeria?")
    #' stats <- agent$get_usage_stats()
    #' stats
    #' }
    get_usage_stats = function() {

      current_cost <- self$cost

      budget_remaining <- NA

      if (!is.na(self$budget)) {
        budget_remaining <- self$budget - as.numeric(current_cost)
      }

      llm_costs <- list(
        estimated_cost = current_cost,
        budget = round(self$budget, 4),
        budget_remaining = round(budget_remaining, 4)
      )

      llm_costs

    },

    #' @description
    #' Reset the agent's conversation history while keeping the system instruction
    #'
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "AI assistant",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #')
    #' agent$invoke("Hello, how are you?")
    #' agent$invoke("Tell me about machine learning")
    #' agent$reset_conversation_history()  # Clears all messages except system prompt
    #' }
    reset_conversation_history = function() {
      system_prompt <- self$llm_object$get_system_prompt()

      private$._messages <- list(
        list(role = "system", content = system_prompt)
      )
      private$.set_turns_from_messages()

      cli::cli_alert_success("Conversation history reset. System prompt preserved.")
      invisible(self)
    },

    #' @description
    #' Saves the agent's current conversation history as a JSON file on disk.
    #' @param file_path Character string specifying the file path where the JSON
    #' file should be saved. Defaults to a file named
    #' `"<agent_name>_messages.json"` in the current working directory.
    #'
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "capital_finder",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #')
    #' agent$invoke("What is the capital of Algeria")
    #' agent$invoke("What is the capital of Italy")
    #' agent$export_messages_history()
    #' }
    #'
    #' @seealso [load_messages_history()] for reloading a saved message history.
    #'
    export_messages_history = function(
    file_path = paste0(getwd(), "/", paste0(self$name, "_messages.json"))
    ) {

      checkmate::assert_string(file_path)

      jsonlite::write_json(
        self$messages,
        path = file_path,
        auto_unbox = TRUE,
        pretty = TRUE
      )

      cli::cli_alert_success(glue::glue("Conversation saved to {file_path}"))

    },

    #' @description
    #' Load an agent's conversation history as a JSON file from disk.
    #' @param file_path Character string specifying the file path where the JSON
    #' file is stored. Defaults to a file named
    #' `"<agent_name>_messages.json"` in the current working directory.
    #'
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "capital_finder",
    #'   instruction = "You are an assistant.",
    #'   llm_object = openai_4_1_mini
    #')
    #' # use the export_messages_history to save the interaction first
    #' agent$load_messages_history("path/to/messages.json")
    #' agent$messages
    #' agent$llm_object
    #' }
    #'
    #' @seealso [export_messages_history()] for exporting the messages object to json.
    #'
    load_messages_history = function(
    file_path = paste0(getwd(), "/", paste0(self$name, "_messages.json"))
    ) {

      checkmate::assert_string(file_path)

      if (!file.exists(file_path)) {
        cli::cli_abort("File does not exist.")
      }

      messages <- jsonlite::read_json(file_path, simplifyVector = FALSE)

      self$messages <- messages

      cli::cli_alert_success(glue::glue("Conversation history loaded from {file_path}"))

    },

    #' @description
    #' Validates an agent's response against custom criteria using LLM-based validation.
    #' This method uses the agent's LLM to evaluate whether a response meets specified
    #' validation criteria (e.g., accuracy, completeness, tone, format).
    #' @param prompt The prompt used to generate the response.
    #' @param response The response text to validate.
    #' @param validation_criteria The criteria for validation.
    #' (e.g., "The response should be accurate and complete", "The response must be under 100 words").
    #' @param validation_score A numeric from 0 to 1 denoting the score to consider for
    #' the evaluation. During the evaluation, the agent will provide a score from 0 to 1,
    #' a provided score greater or equal to the `validation_score` will result in a `valid` response.
    #' Defaults to 0.8.
    #' @return list object
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "fact_checker",
    #'   instruction = "You are a factual assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' prompt <- "What is the capital of Algeria?"
    #' response <- agent$invoke(prompt)
    #' validation <- agent$validate_response(
    #'   response = response,
    #'   prompt = prompt,
    #'   validation_criteria = "The response must be accurate and mention Algiers",
    #'   validation_score = 0.8
    #' )
    #' print(validation)
    #' }
    validate_response = function(
    prompt,
    response,
    validation_criteria,
    validation_score = 0.8
    ) {

      checkmate::assert_string(validation_criteria)
      checkmate::assert_string(response)
      checkmate::assert_string(prompt)
      checkmate::assert_numeric(validation_score, lower = 0, upper = 1)

      validation_prompt <- paste0(
        "You are a response validation assistant. Evaluate the following response ",
        "against the specified criteria.\n\n",
        "Validation Criteria: ", validation_criteria, "\n\n",
        "Original Prompt: ", prompt, "\n\n",
        "Response to Validate:\n", response, "\n\n",
        "Provide your evaluation in the following format:\n",
        "SCORE: [0-1] If the response is extremely valid it gets a score of 1.\n",
        "FEEDBACK: [detailed feedback explaining your evaluation]\n\n",
        "Return ONLY the three lines above, nothing else."
      )

      original_system_prompt <- self$llm_object$get_system_prompt()

      validation_system_prompt <- paste0(
        "You are a response validation assistant. ",
        "Your task is to evaluate responses against given criteria. ",
        "Be objective and thorough in your evaluation.",
        "Provide a score and a feedback. The score should evaluate the validitiy of the ",
        "response, it goes from 0 to 1 (1 is the best score and 0 the worst score)"
      )

      self$llm_object$set_system_prompt(value = validation_system_prompt)

      on.exit(
        self$llm_object$set_system_prompt(value = original_system_prompt),
        add = TRUE,
        after = FALSE
      )

      validation_result <- self$llm_object$chat_structured(
        validation_prompt,
        type = ellmer::type_object(
          score = ellmer::type_number(
            description = "The validity score. If a response is extremely valid, it gets a score of 1",
            required = TRUE
          ),
          feedback = ellmer::type_string(
            description = "Feedback concerning the evaluation of the response according to the prompt",
            required = TRUE
          )
        )
      )

      score <- validation_result$score

      if (score >= validation_score) {
        valid <- TRUE
      } else {
        valid <- FALSE
      }

      validation_result <- append(
        list(
          prompt = prompt,
          response = response,
          validation_criteria = validation_criteria,
          validation_score = validation_score,
          valid = valid
        ),
        validation_result
      )

      if (valid) {
        cli::cli_alert_success(glue::glue(
          "The response is considered valid with a score of {score}"
        ))
      } else {
        cli::cli_alert_warning(
          glue::glue(
            "The response is considered invalid with a score of {score}"
          )
        )
      }

      validation_result
    },

    #' @description
    #' Register one or more tools with the agent
    #' @param tools A list of ellmer tool objects or a single tool object
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "file_assistant",
    #'   instruction = "You are a file management assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' # Register predefined tools
    #' agent$register_tools(list(change_directory_tool(), list_files_tool()))
    #' }
    register_tools = function(tools) {
      if (!is.list(tools)) {
        tools <- list(tools)
      }

      for (tool in tools) {
        if (!inherits(tool, "ellmer::ToolDef")) {
          cli::cli_abort("All tools must be valid ellmer Tool objects")
        }

        tool_name <- tool@name
        self$tools[[tool_name]] <- tool
        cli::cli_alert_success("Registered tool: {tool_name}")
      }

      private$.update_llm_tools()

      invisible(self)
    },

    #' @description
    #' List all registered tools
    #' @return Character vector of tool names
    #' @examples
    #' \dontrun{
    #' agent$list_tools()
    #' }
    list_tools = function() {
      tool_names <- names(self$tools)

      if (length(tool_names) == 0) {
        cli::cli_alert_info("No tools registered")
        return(character(0))
      }

      tool_names
    },

    #' @description
    #' Remove tools from the agent
    #' @param tool_names Character vector of tool names to remove
    #' @examples
    #' \dontrun{
    #' agent$remove_tools(c("change_directory", "list_files"))
    #' }
    remove_tools = function(tool_names) {
      checkmate::assert_character(tool_names)

      for (tool_name in tool_names) {
        if (tool_name %in% names(self$tools)) {
          self$tools[[tool_name]] <- NULL
          cli::cli_alert_success("Removed tool: {tool_name}")
        } else {
          cli::cli_alert_warning("Tool not found: {tool_name}")
        }
      }

      private$.update_llm_tools()

      invisible(self)
    },

    #' @description
    #' Clear all registered tools
    #' @examples
    #' \dontrun{
    #' agent$clear_tools()
    #' }
    clear_tools = function() {
      tool_count <- length(self$tools)
      self$tools <- list()

      private$.update_llm_tools()

      cli::cli_alert_success("Cleared {tool_count} tool{?s}")
      invisible(self)
    },

    #' @description
    #' Generate and register a tool from a natural language description.
    #' This method uses the agent's LLM to create both the R function and the
    #' ellmer tool definition based on your description, then automatically registers it.
    #' @param description Character string describing what the tool should do
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "file_assistant",
    #'   instruction = "You are a helpful assistant.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' # Generate and register a tool from description
    #' agent$generate_and_register_tool(
    #'   description = "Create a tool that saves text content to a file on disk"
    #' )
    #' # Now the tool is available to use
    #' agent$list_tools()
    #' }
    generate_and_register_tool = function(description) {

      checkmate::assert_string(description)

      cli::cli_h2("Generating tool from description")
      cli::cli_alert_info("Description: {description}")

      generation_prompt <- glue::glue(
        "Generate R code to create an ellmer tool based on this description: <<<<<description>>>>>

          Requirements:
          1. Create a complete R function that implements the tool's functionality
          2. Wrap it with ellmer::tool() to create a tool definition
          3. Use appropriate type specifications: ellmer::type_string(), ellmer::type_number(), ellmer::type_integer(), ellmer::type_boolean(), ellmer::type_enum(), ellmer::type_array(), ellmer::type_object()
          4. Include clear descriptions for the tool and each parameter
          Format your response as valid R code that can be evaluated directly. Include ONLY the R code without markdown formatting, explanations, or comments. The code should define the function and create the tool in one block.
          5. Do not use parameters that do not exist, for example, there is no 'output' parameter in the ellmer::tool function
          Example format 1:
          save_file_function <- function(content, filepath) {{
              writeLines(content, filepath)
              paste('File saved to', filepath)
          }}
          ellmer::tool(
            save_file_function,
            name = 'save_file',
            description = 'Save text content to a file',
            arguments = list(
              content = ellmer::type_string(description = 'Text content to save'),
              filepath = ellmer::type_string(description = 'Path where file should be saved')
            )
          )

        Example format 2:

        get_current_time <- function(tz = 'UTC') {
          format(Sys.time(), tz = tz, usetz = TRUE)
        }

        get_current_time <- ellmer::tool(
          get_current_time,
          name = 'get_current_time',
          description = 'Returns the current time.',
          arguments = list(
           tz = ellmer::type_string(
           'Time zone to display the current time in. Defaults to `\"UTC\"`.',
           required = FALSE
          )
        )
        )
        "
      , .open = "<<<<<", .close = ">>>>>")

      # Store original system prompt
      original_system_prompt <- self$llm_object$get_system_prompt()
      tool_generation_prompt <- paste0(
        "You are an expert R programmer specializing in creating ellmer tools. ",
        "Generate clean, functional R code that creates tools for LLM use."
      )

      self$llm_object$set_system_prompt(value = tool_generation_prompt)

      # Restore original system prompt on exit
      on.exit(
        self$llm_object$set_system_prompt(value = original_system_prompt),
        add = TRUE,
        after = FALSE
      )

      # Generate the tool code
      generated_code <- self$llm_object$chat(generation_prompt)

      cli::cli_h2("The following tool will be registered")
      cli::cli_code(generated_code)

      # Clean the generated code
      clean_code <- gsub("```\\{?r\\}?|```", "", generated_code)
      clean_code <- trimws(clean_code)
      clean_code <- paste0("{", clean_code, "}")

      expr <- rlang::parse_expr(clean_code)

      tool_result <- rlang::eval_bare(
        expr,
        env = rlang::caller_env()
      )

      self$register_tools(tool_result)

      cli::cli_alert_success("Tool successfully generated and registered")
      cli::cli_alert_info("Call '<agent-name>$llm_object$get_tools()' to inspect the tools")
      cli::cli_alert_info("If satisfied, you can copy the tool and put in your corresponding R file")
    },

    #' @description
    #' Create a copy of the agent with the same instruction and configuration but a new unique ID.
    #' Useful for creating multiple instances of the same agent type.
    #'
    #' @param new_name Optional character string to assign a new name to the cloned agent.
    #' If NULL, the cloned agent retains the original name.
    #' @examples
    #' \dontrun{
    #' openai_4_1_mini <- ellmer::chat(
    #'   name = "openai/gpt-4.1-mini",
    #'   api_key = Sys.getenv("OPENAI_API_KEY"),
    #'   echo = "none"
    #' )
    #' agent <- Agent$new(
    #'   name = "translator",
    #'   instruction = "You are a translator.",
    #'   llm_object = openai_4_1_mini
    #' )
    #' # Clone with same name
    #' agent_copy <- agent$clone_agent()
    #' }
    clone_agent = function(new_name = NULL) {

      if (!is.null(new_name)) {
        checkmate::assert_string(new_name)
      }

      cloned_agent <- self$clone(deep = TRUE)

      cloned_agent$agent_id <- uuid::UUIDgenerate()

      if (!is.null(new_name)) {
        cloned_agent$name <- new_name
      }

      cli::cli_alert_success(glue::glue(
        "Agent cloned successfully. New ID: {cloned_agent$agent_id}"
      ))

      return(cloned_agent)
    },

    #' @field name The agent's name.
    name = NULL,
    #' @field instruction The agent's role/system prompt.
    instruction = NULL,
    #' @field llm_object The underlying `ellmer::chat_openai` object.
    llm_object = NULL,
    #' @field agent_id A UUID uniquely identifying the agent.
    agent_id = NULL,
    #'@field model_provider The name of the entity providing the model (eg. OpenAI)
    model_provider = NULL,
    #'@field model_name The name of the model to be used (eg. gpt-4.1-mini)
    model_name = NULL,
    #'@field broadcast_history A list of all past broadcast interactions.
    broadcast_history = list(),
    #'@field budget A budget in $ that the agent should not exceed.
    budget = NULL,
    #'@field budget_policy A list controlling budget behavior: on_exceed and warn_at.
    budget_policy = NULL,
    #'@field budget_warned Internal flag indicating whether warn_at notice was emitted.
    budget_warned = NULL,
    #'@field cost The current cost of the agent
    cost = NULL,
    #'@field tools A list of registered tools available to the agent
    tools = list()

  ),

  active = list(
    #' @field messages Public active binding for the conversation history.
    #' Assignment is validated automatically.
    messages = function(value) {
      if (missing(value)) {
        return(private$._messages)
      }

      if (!is.list(value)) {
        cli::cli_abort("messages must be a list of message objects")
      }

      for (msg in value) {
        if (!is.list(msg) || !all(c("role", "content") %in% names(msg))) {
          cli::cli_abort("Each message must be a list with 'role' and 'content'")
        }
        if (!msg$role %in% c("system", "user", "assistant")) {
          cli::cli_abort(paste0("Invalid role: ", msg$role))
        }
      }

      # SYNC DIRECTION 1: Messages → Turns (triggered by user assignment)
      # When user assigns: agent$messages <- new_list
      # This active binding stores the new messages and syncs them to ellmer turns.
      # The sync flag is TRUE here, so .set_turns_from_messages() will execute.
      private$._messages <- value
      private$.set_turns_from_messages()  # Sync the new messages to ellmer object

    }
  ),

  private = list(
    ._messages = NULL,
    # Synchronization control flag to prevent infinite recursion between
    # messages and turns. When TRUE, allows normal sync operations.
    # When FALSE, prevents .set_turns_from_messages() from executing.
    .sync_flag = TRUE,
    .add_message = function(message, type) {
      private$._messages[[length(private$._messages) + 1]] <- list(
        role = type,
        content = message
      )
    },

    .add_assistant_message = function(message, type = "assistant") {
      private$.add_message(message, type)
    },

    .add_user_message = function(message, type = "user") {
      private$.add_message(message, type)
    },
    .format_arguments = function(args) {
      parts <- mapply(
        function(name, value) {
          # Quote character values
          if (is.character(value)) {
            sprintf('%s = "%s"', name, value)
          } else {
            sprintf('%s = %s', name, value)
          }
        },
        names(args), args,
        USE.NAMES = FALSE
      )

      paste(parts, collapse = ", ")
    },

    .set_messages_from_turns = function() {

      turns <- self$llm_object$get_turns(include_system_prompt = TRUE)

      messages <- lapply(turns, function(turn) {

        role <- turn@role

        contents <- turn@contents


        content_strings <- vapply(contents, function(ct) {

          cls <- class(ct)[[1]]

          if (grepl("ContentText", cls, ignore.case = TRUE)) {

            msg <- ct@text

            return(msg)

          } else if (grepl("ContentToolRequest", cls, ignore.case = TRUE)) {

            call_id <- ct@id
            tool_name <- ct@name
            args <- private$.format_arguments(ct@arguments)

            msg <- sprintf(
              "[tool request id=%s]: %s(%s)",
              call_id,
              tool_name,
              args
            )

            return(msg)

          } else if (grepl("ContentToolResult", cls, ignore.case = TRUE)) {

            call_id <- ct@request@id
            result <- ct@value

            if (is.list(result)) {
              result <- glue::glue_collapse(
                glue::glue("{names(result)}: {lapply(result, toString)}"),
                sep = "     "
              )
            }

            msg <- sprintf(
              "[tool result id=%s]: %s",
              call_id,
              result
            )

            return(msg)
          } else {
            return(as.character(ct@text))
          }

        }, FUN.VALUE = character(1))

        content <- paste(content_strings, collapse = "\n")

        list(
          role = role,
          content = content
        )
      })

      first_message <- messages[[1]]

      if (!first_message$role == "system") {

        messages <- append(
          list(
            list(
              role = "system",
              content = self$instruction
            )
          ),
          messages
        )
      }

      # SYNC DIRECTION 2: Turns → Messages (prevents recursion)
      # Temporarily disable sync flag to prevent the messages active binding
      # from triggering .set_turns_from_messages() again when we update messages.
      # This breaks the potential infinite loop:
      # .set_messages_from_turns() -> messages assignment -> active binding -> .set_turns_from_messages()
      private$.sync_flag <- FALSE
      private$._messages <- messages  # Direct assignment bypasses active binding
      private$.sync_flag <- TRUE      # Re-enable sync for future operations
    },

    .set_turns_from_messages = function() {
      # SYNC DIRECTION 1: Messages → Turns (respects sync flag)
      # Check sync flag to prevent execution during recursive calls.
      # When .set_messages_from_turns() is running, this prevents infinite loops.
      if (!private$.sync_flag) {
        return(invisible(NULL))
      }

      messages <- self$messages
      turns <- list()
      content_tool_requests <- list()
      existing_turns <- self$llm_object$get_turns(
        include_system_prompt = TRUE
      )

      for (i in seq_along(messages)) {

        msg <- messages[[i]]

        if (grepl("[tool request id=", msg$content, fixed = TRUE)) {

          id <- sub(".*id=([^]]+)].*", "\\1", msg$content)
          func_name <- sub(".*]: ([a-zA-Z0-9_]+)\\(.*", "\\1", msg$content)
          params_str <- sub(".*\\((.*)\\).*", "\\1", msg$content)
          tool <- eval(parse(text = func_name))

          params_list <- eval(parse(text = paste0("list(", params_str, ")")))

          content_to_consider <- ellmer::ContentToolRequest(
            id = id,
            name = func_name,
            arguments = params_list,
            tool = tool
          )

          tmp_list <- list(
            content_to_consider
          )

          tmp_list <- setNames(tmp_list, id)

          content_tool_requests <- append(
            content_tool_requests,
            tmp_list
          )

        } else if (grepl("[tool result id=", msg$content, fixed = TRUE)) {

          id <- sub(".*id=([^]]+)].*", "\\1", msg$content)

          result <- sub(".*]:\\s*", "", msg$content)

          if (length(content_tool_requests) > 0 & !id %in% names(content_tool_requests)) {
            cli::cli_abort(
              glue::glue(
                "{id} tool result detected but no corresponding tool request found, please fix."
              )
            )
          }

          request <- content_tool_requests[[id]]

          content_to_consider <- ellmer::ContentToolResult(
            value = result,
            request = request
          )

        } else {
          content_to_consider <- ellmer::ContentText(msg$content)
        }

        tokens <- c(0, 0, 0)

        if (i <= length(existing_turns)) {
          turn_to_consider <- existing_turns[[i]]
          if (S7::prop_exists(turn_to_consider, "tokens")) {
            tokens <- as.vector(turn_to_consider@tokens)
          }
        }

        turn <- ellmer::Turn(
          role = msg$role,
          contents = list(content_to_consider),
          tokens = tokens
        )

        turns <- append(turns, list(turn))
      }

      self$llm_object$set_turns(turns)

    },

    .validate_r_code = function(r_code, result) {

      validation <- tryCatch({
        parsed <- parse(text = r_code)
        result$validated <- TRUE
        result$validation_message <- "Syntax is valid"
        return(result)
      }, error = function(e) {
        result$validated <- FALSE
        result$validation_message <- paste("Syntax error:", e$message)
        return(result)
      })

    },

    .check_budget = function() {

      current_cost <- as.numeric(self$llm_object$get_cost())

      warn_at <- self$budget_policy$warn_at
      ratio <- current_cost / as.numeric(self$budget)

      budget_exceeded <- current_cost > self$budget

      if (ratio >= warn_at && !budget_exceeded) {
        cli::cli_alert_warning(
          glue::glue(
            "{self$name} budget nearing limit: Cost {format(round(current_cost, 4), scientific = FALSE)} / . ",
            "Budget {format(round(self$budget, 4), scientific = FALSE)} ({round(ratio * 100, 1)}%)"
          ))
      }

      if (!budget_exceeded) {
        return(invisible(NULL))
      }

      policy <- self$budget_policy$on_exceed

      if (policy == "warn") {
        cli::cli_alert_warning(glue::glue(
          "{self$name} exceeded budget: Cost {format(round(current_cost,4), scientific = FALSE)} > ",
          "Budget {format(round(self$budget,4), scientific = FALSE)}. Proceeding per policy 'warn'."
        ))
      }

      if (policy == "ask") {
        user_input <- readline(prompt = glue::glue(
          "Budget exceeded (Cost {format(round(current_cost,4), scientific = FALSE)} > Budget {format(round(self$budget,4), scientific = FALSE)}). ",
          "Continue? [y/N]: "
        ))
        if (tolower(user_input) != "y") {
          cli::cli_abort(glue::glue(
            "{self$name} agent cancelled due to budget exceedance. ",
            "Cost: {format(round(current_cost,4), scientific = FALSE)}, Budget: {format(round(self$budget,4), scientific = FALSE)}"
          ))
          return(invisible(NULL))
        }
      }

      if (policy == "abort") {
        cli::cli_abort(glue::glue(
          "{self$name} agent has exceeded its budget. ",
          "Cost: {format(round(current_cost, 4), scientific = FALSE)}, Budget: {format(round(self$budget, 4), scientific = FALSE)}"
        ))
      }
    },

    .update_llm_tools = function() {
      if (length(self$tools) > 0) {
        tool_list <- unname(self$tools)
        suppressMessages({
          self$llm_object$register_tools(tool_list)
        })
      } else {
        suppressMessages({
          self$llm_object$set_tools(list())
        })
      }
    }
  )
)

library(testthat)
library(mini007)

providerclass <- setClass(
  "Provider",
  slots = c(
    name = "character",
    model  = "character"
  ))

provider <- new("Provider", name = "dummy", model = "v0")

# Helper: Simplified mock for the ellmer LLM object with minimal sync
DummyChat <- R6::R6Class(
  "Chat",
  public = list(
    roles = NULL,
    system_prompt = NULL,
    turns = list(),
    tools = list(),
    provider = provider,
    chat = function(prompt) paste("Echo:", prompt),
    get_cost = function() 2.0,
    get_tokens = function() data.frame(tokens_total = 10),
    set_system_prompt = function(value) { self$system_prompt <- value },
    get_system_prompt = function() if (!is.null(self$system_prompt)) self$system_prompt else "default",
    set_turns = function(turns) { self$turns <- turns },
    get_turns = function(include_system_prompt = FALSE) { return(self$turns) },
    register_tools = function(tools) { self$tools <- tools },
    get_provider = function() { self$provider }
  )
)

dummy_chat <- DummyChat$new()

# --- Agent initialization ---
test_that("Agent initializes with correct fields", {
  agent <- Agent$new(
    name = "TestAgent",
    instruction = "Test instruction.",
    llm_object = dummy_chat$clone()
  )
  expect_equal(agent$name, "TestAgent")
  expect_equal(agent$instruction, "Test instruction.")
  expect_equal(agent$llm_object$get_system_prompt(), "Test instruction.")
  expect_length(agent$messages, 1)
  expect_equal(agent$messages[[1]]$role, "system")
})

# --- Budget and policy ---
test_that("set_budget and set_budget_policy manipulate values", {
  agent <- Agent$new("B", "Instr", dummy_chat$clone())
  agent$set_budget(123)
  expect_equal(agent$budget, 123)
  agent$set_budget_policy("warn", warn_at = 0.3)
  expect_equal(agent$budget_policy$on_exceed, "warn")
  expect_equal(agent$budget_policy$warn_at, 0.3)
})

# --- add_message - simplified without sync ---
test_that("add_message appends messages to list", {
  agent <- Agent$new("M", "I", dummy_chat$clone())
  # Manually add to messages without triggering sync
  agent$messages <- c(agent$messages, list(
    list(role = "user", content = "hi"),
    list(role = "assistant", content = "hello"),
    list(role = "system", content = "info")
  ))
  expect_equal(length(agent$messages), 4)
  expect_equal(agent$messages[[4]]$role, "system")
  expect_equal(agent$messages[[4]]$content, "info")
})

# --- reset_conversation_history ---
test_that("reset_conversation_history preserves only system prompt", {
  agent <- Agent$new("N", "Sys", dummy_chat$clone())
  # Manually add messages
  agent$messages <- c(agent$messages, list(
    list(role = "user", content = "hi"),
    list(role = "assistant", content = "there")
  ))
  expect_equal(length(agent$messages), 3)
  agent$reset_conversation_history()
  expect_length(agent$messages, 1)
  expect_equal(agent$messages[[1]]$role, "system")
  expect_equal(agent$messages[[1]]$content, "Sys")
})

# --- export/load_messages_history ---
test_that("export and load history roundtrip", {
  tmpfile <- tempfile(fileext = ".json")
  agent <- Agent$new("Hist", "Testing history", dummy_chat$clone())
  # Manually set messages to avoid sync issues
  agent$messages <- list(
    list(role = "system", content = "Testing history"),
    list(role = "user", content = "hi"),
    list(role = "assistant", content = "response")
  )
  agent$export_messages_history(tmpfile)
  # start a new agent, load history
  agent2 <- Agent$new("Hist2", "Will get replaced", dummy_chat$clone())
  agent2$load_messages_history(tmpfile)
  expect_equal(
    vapply(agent2$messages, `[[`, "content", FUN.VALUE = character(1)),
    vapply(agent$messages, `[[`, "content", FUN.VALUE = character(1))
  )
  unlink(tmpfile)
})

# --- keep_last_n_messages ---
test_that("keep_last_n_messages keeps only system and N most recent", {
  agent <- Agent$new("S", "info", dummy_chat$clone())
  # Manually set up messages
  agent$messages <- list(
    list(role = "system", content = "info"),
    list(role = "user", content = "first"),
    list(role = "assistant", content = "hey first!"),
    list(role = "user", content = "second"),
    list(role = "assistant", content = "hey second!")
  )
  expect_equal(length(agent$messages), 5)
  agent$keep_last_n_messages(2)
  expect_equal(length(agent$messages), 3)
  roles <- vapply(agent$messages, `[[`, "role", FUN.VALUE = character(1))
  expect_equal(roles, c("system", "user", "assistant"))
})

# --- update_instruction ---
test_that("update_instruction changes both field and system prompt", {
  agent <- Agent$new("X", "Old", dummy_chat$clone())
  agent$update_instruction("New role")
  expect_equal(agent$instruction, "New role")
  expect_equal(agent$llm_object$get_system_prompt(), "New role")
})

# --- add_message errors ---
test_that("add_message rejects invalid role", {
  agent <- Agent$new("M2", "I2", dummy_chat$clone())
  expect_error(agent$add_message("bogus", "x"))
})

# --- messages active binding validation ---
test_that("messages active binding validates roles", {
  agent <- Agent$new("MB", "InstrMB", dummy_chat$clone())
  # invalid role in list
  expect_error({ agent$messages <- list(list(role = "bogus", content = "x")) })
})


# --- keep_last_n_messages edge case ---
test_that("keep_last_n_messages with n=1 keeps system and last message", {
  agent <- Agent$new("S2", "info2", dummy_chat$clone())
  agent$messages <- list(
    list(role = "system", content = "info2"),
    list(role = "user", content = "only")
  )
  agent$keep_last_n_messages(1)
  expect_equal(length(agent$messages), 2)
  roles <- vapply(agent$messages, `[[`, "role", FUN.VALUE = character(1))
  expect_equal(roles, c("system", "user"))
})

# --- share_context_with ---
test_that("share_context_with transfers last messages as user entries", {
  agent1 <- Agent$new("A1", "instr", dummy_chat$clone())
  agent2 <- Agent$new("A2", "instr", dummy_chat$clone())
  # create some messages in agent1
  agent1$messages <- c(agent1$messages,
                       list(list(role = "user", content = "hello"),
                            list(role = "assistant", content = "hi"),
                            list(role = "user", content = "how are you?")))
  # share two most recent messages
  agent1$share_context_with(agent2, n = 2)
  # agent2 should have received two additional user messages
  expect_true(length(agent2$messages) >= 3)
  last_roles <- vapply(tail(agent2$messages, 2), `[[`, "role", FUN.VALUE = character(1))
  expect_equal(last_roles, c("assistant", "user"))
  last_contents <- vapply(tail(agent2$messages, 2), `[[`, "content", FUN.VALUE = character(1))
  expect_equal(last_contents,
               c("hi", "how are you?"))
})

# --- get_usage_stats ---
test_that("get_usage_stats returns expected fields and values", {
  agent <- Agent$new("Stats", "Role", dummy_chat$clone())
  stats <- agent$get_usage_stats()
  expect_true(all(c("estimated_cost", "budget", "budget_remaining") %in% names(stats)))
  expect_true(is.na(stats$budget))
  expect_true(is.na(stats$budget_remaining))
})

# --- set_budget_policy validation ---
test_that("set_budget_policy validates inputs", {
  agent <- Agent$new("BP", "I", dummy_chat$clone())
  expect_error(agent$set_budget_policy(on_exceed = "nope"))
  expect_error(agent$set_budget_policy(on_exceed = "abort", warn_at = 1.5))
})


# --- load_messages_history errors ---
test_that("load_messages_history errors on missing file", {
  agent <- Agent$new("HistErr", "I", dummy_chat$clone())
  tmp <- tempfile(fileext = ".json")
  if (file.exists(tmp)) unlink(tmp)
  expect_error(agent$load_messages_history(tmp))
})

# --- Agent initialization validation ---
test_that("Agent initialization validates inputs correctly", {
  # Test invalid llm_object
  fake_object <- list(not = "a_chat_object")
  expect_error(Agent$new("Test", "Instruction", fake_object))

  # Test non-string inputs
  expect_error(Agent$new(123, "Instruction", dummy_chat$clone()))
  expect_error(Agent$new("Test", 456, dummy_chat$clone()))

  # Note: Empty strings might be allowed by checkmate::assert_string()
  # so we don't test for those as errors unless specifically required
})

# --- Budget functionality ---
test_that("set_budget validates inputs and sets values correctly", {
  agent <- Agent$new("BudgetTest", "I", dummy_chat$clone())

  # Test valid budget setting
  agent$set_budget(50.0)
  expect_equal(agent$budget, 50.0)

  # Test invalid budget values
  expect_error(agent$set_budget(-1))
  expect_error(agent$set_budget("not_a_number"))
})

test_that("budget policy configuration works correctly", {
  agent <- Agent$new("PolicyTest", "I", dummy_chat$clone())

  # Test valid policy settings
  agent$set_budget_policy("warn", 0.9)
  expect_equal(agent$budget_policy$on_exceed, "warn")
  expect_equal(agent$budget_policy$warn_at, 0.9)

  # Test default values
  agent$set_budget_policy("abort")
  expect_equal(agent$budget_policy$warn_at, 0.8)

  # Test invalid policy values
  expect_error(agent$set_budget_policy("invalid_policy"))
  expect_error(agent$set_budget_policy("warn", 1.5))
  expect_error(agent$set_budget_policy("warn", -0.1))
})

# --- Message validation and manipulation ---
test_that("add_message validates inputs correctly", {
  agent <- Agent$new("MessageTest", "I", dummy_chat$clone())

  # Test invalid role
  expect_error(agent$add_message("invalid_role", "content"))

  # Test invalid content type
  expect_error(agent$add_message("user", NULL))
  expect_error(agent$add_message("user", 123))
})

test_that("messages active binding validates message structure", {
  agent <- Agent$new("BindingTest", "I", dummy_chat$clone())

  # Test invalid message structure - missing fields
  expect_error(agent$messages <- list(list(role = "user"))) # missing content
  expect_error(agent$messages <- list(list(content = "test"))) # missing role

  # Test invalid message structure - not a list
  expect_error(agent$messages <- "not_a_list")
  expect_error(agent$messages <- c("user", "content"))
})

# --- Export/Import functionality ---
test_that("export_messages_history creates valid JSON file", {
  tmpfile <- tempfile(fileext = ".json")
  agent <- Agent$new("ExportTest", "Testing export", dummy_chat$clone())

  # Set some messages
  agent$messages <- list(
    list(role = "system", content = "Testing export"),
    list(role = "user", content = "Hello"),
    list(role = "assistant", content = "Hi there")
  )

  # Test export
  agent$export_messages_history(tmpfile)
  expect_true(file.exists(tmpfile))

  # Test file content is valid JSON
  loaded_data <- jsonlite::read_json(tmpfile, simplifyVector = FALSE)
  expect_equal(length(loaded_data), 3)
  expect_equal(loaded_data[[2]]$role, "user")
  expect_equal(loaded_data[[2]]$content, "Hello")

  unlink(tmpfile)
})

test_that("export_messages_history validates file path", {
  agent <- Agent$new("ExportValidation", "I", dummy_chat$clone())
  expect_error(agent$export_messages_history(NULL))
  expect_error(agent$export_messages_history(123))
})

# --- Agent cloning ---
test_that("clone_agent creates independent copy", {
  agent1 <- Agent$new("Original", "Original instruction", dummy_chat$clone())
  agent1$set_budget(100)

  # Clone without new name
  agent2 <- agent1$clone_agent()
  expect_equal(agent2$name, "Original")
  expect_equal(agent2$instruction, "Original instruction")
  expect_equal(agent2$budget, 100)
  expect_false(agent1$agent_id == agent2$agent_id) # Different IDs

  # Clone with new name
  agent3 <- agent1$clone_agent("NewName")
  expect_equal(agent3$name, "NewName")
  expect_false(agent1$agent_id == agent3$agent_id)
})

test_that("clone_agent validates new_name parameter", {
  agent <- Agent$new("CloneTest", "I", dummy_chat$clone())
  expect_error(agent$clone_agent(123))
  expect_error(agent$clone_agent(c("a", "b")))
})

# --- Instruction updates ---
test_that("update_instruction modifies system prompt correctly", {
  agent <- Agent$new("InstructionTest", "Old instruction", dummy_chat$clone())
  original_message_count <- length(agent$messages)

  agent$update_instruction("New instruction")

  expect_equal(agent$instruction, "New instruction")
  expect_equal(agent$llm_object$get_system_prompt(), "New instruction")
  expect_equal(agent$messages[[1]]$content, "New instruction")
  expect_equal(length(agent$messages), original_message_count) # Same number of messages
})

test_that("update_instruction validates input", {
  agent <- Agent$new("ValidationTest", "I", dummy_chat$clone())
  expect_error(agent$update_instruction(NULL))
  expect_error(agent$update_instruction(123))
  expect_error(agent$update_instruction(c("a", "b")))
})

# --- Keep last N messages functionality ---
test_that("keep_last_n_messages handles normal cases", {
  agent <- Agent$new("KeepTest", "System", dummy_chat$clone())

  # Set up multiple messages
  agent$messages <- list(
    list(role = "system", content = "System"),
    list(role = "user", content = "msg1"),
    list(role = "assistant", content = "resp1"),
    list(role = "user", content = "msg2"),
    list(role = "assistant", content = "resp2"),
    list(role = "user", content = "msg3"),
    list(role = "assistant", content = "resp3")
  )

  # Test keeping a reasonable number of messages
  agent$keep_last_n_messages(3)
  # Should keep system + 3 most recent messages
  expect_equal(length(agent$messages), 4) # system + 3

  # Test keeping 0 messages (should keep at least system)
  expect_error(agent$keep_last_n_messages(0))
})

test_that("keep_last_n_messages validates input", {
  agent <- Agent$new("KeepValidation", "I", dummy_chat$clone())
  expect_error(agent$keep_last_n_messages(-1))
  expect_error(agent$keep_last_n_messages("not_a_number"))
  # Note: checkmate::assert_integerish might accept vectors, so this test might not fail
  # expect_error(agent$keep_last_n_messages(c(1, 2)))
})

# --- Usage stats ---
test_that("get_usage_stats handles different cost scenarios", {
  agent <- Agent$new("StatsTest", "I", dummy_chat$clone())

  # Test with no budget set
  stats <- agent$get_usage_stats()
  expect_true(is.na(stats$budget))
  expect_true(is.na(stats$budget_remaining))

  # Test with budget but no cost
  agent$set_budget(10)
  stats <- agent$get_usage_stats()
  expect_equal(stats$budget, 10)
  expect_true(is.na(stats$budget_remaining)) # because cost is NA

  # Test with both budget and cost
  agent$cost <- 3.5
  stats <- agent$get_usage_stats()
  expect_equal(stats$estimated_cost, 3.5)
  expect_equal(stats$budget_remaining, 6.5)
})

# --- Field access and initialization ---
test_that("Agent fields are initialized correctly", {
  agent <- Agent$new("FieldTest", "Test instruction", dummy_chat$clone())

  expect_equal(agent$name, "FieldTest")
  expect_equal(agent$instruction, "Test instruction")
  expect_true(!is.null(agent$agent_id))
  expect_true(is.character(agent$agent_id))
  expect_equal(length(agent$agent_id), 1)
  expect_true(is.na(agent$budget))
  expect_equal(agent$budget_policy$on_exceed, "abort")
  expect_equal(agent$budget_policy$warn_at, 0.8)
  expect_true(is.na(agent$cost))
  expect_true(is.list(agent$broadcast_history))
  expect_equal(length(agent$broadcast_history), 0)
})

# --- Error handling in various methods ---
test_that("methods handle invalid input types gracefully", {
  agent <- Agent$new("ErrorTest", "I", dummy_chat$clone())

  # Test update_instruction with invalid types
  expect_error(agent$update_instruction(NULL))
  expect_error(agent$update_instruction(list("instruction")))

  # Test load_messages_history with invalid path
  expect_error(agent$load_messages_history(NULL))
  expect_error(agent$load_messages_history(123))
})

# --- Provider and model information ---
test_that("Agent captures provider information correctly", {
  agent <- Agent$new("ProviderTest", "I", dummy_chat$clone())

  expect_equal(agent$model_provider, "dummy")
  expect_equal(agent$model_name, "v0")
  expect_true(!is.null(agent$llm_object$get_provider()))
})

# --- Message list manipulation ---
test_that("messages can be accessed and modified correctly", {
  agent <- Agent$new("AccessTest", "I", dummy_chat$clone())

  # Test getter
  initial_messages <- agent$messages
  expect_equal(length(initial_messages), 1)
  expect_equal(initial_messages[[1]]$role, "system")

  # Test setter with valid messages
  new_messages <- list(
    list(role = "system", content = "New system"),
    list(role = "user", content = "User message"),
    list(role = "assistant", content = "Assistant response")
  )
  agent$messages <- new_messages

  retrieved_messages <- agent$messages
  expect_equal(length(retrieved_messages), 3)
  expect_equal(retrieved_messages[[2]]$content, "User message")
})

# --- Note: Tool management methods are not included in the current Agent implementation ---

# --- Additional edge case tests ---

test_that("Agent handles large numbers of messages correctly", {
  agent <- Agent$new("LargeTest", "System", dummy_chat$clone())

  # Create a large number of messages
  large_message_list <- list(list(role = "system", content = "System"))
  for (i in 1:100) {
    large_message_list <- c(large_message_list, list(
      list(role = "user", content = paste("Message", i)),
      list(role = "assistant", content = paste("Response", i))
    ))
  }

  # Test setting large message list
  expect_no_error(agent$messages <- large_message_list)
  expect_equal(length(agent$messages), 201) # 1 system + 200 user/assistant

  # Test keep_last_n_messages with large list
  expect_no_error(agent$keep_last_n_messages(10))
  expect_equal(length(agent$messages), 11) # system + 10 most recent
})

test_that("Agent budget calculations handle edge values", {
  agent <- Agent$new("BudgetEdge", "I", dummy_chat$clone())

  # Test with very small budget
  agent$set_budget(0.001)
  expect_equal(agent$budget, 0.001)

  # Test with zero cost
  agent$cost <- 0
  stats <- agent$get_usage_stats()
  expect_equal(stats$budget_remaining, 0.001)

  # Test with cost exceeding budget
  agent$cost <- 0.01
  stats <- agent$get_usage_stats()
  expect_true(stats$budget_remaining < 0)
})

test_that("Agent message export handles different file extensions", {
  agent <- Agent$new("ExportTypes", "Test", dummy_chat$clone())

  # Test with .json extension
  tmpfile_json <- tempfile(fileext = ".json")
  expect_no_error(agent$export_messages_history(tmpfile_json))
  expect_true(file.exists(tmpfile_json))
  unlink(tmpfile_json)

  # Test with no extension
  tmpfile_none <- tempfile()
  expect_no_error(agent$export_messages_history(tmpfile_none))
  expect_true(file.exists(tmpfile_none))
  unlink(tmpfile_none)
})

test_that("Agent cloning preserves all state correctly", {
  agent1 <- Agent$new("FullState", "Original", dummy_chat$clone())

  # Set up complex state
  agent1$set_budget(25.5)
  agent1$set_budget_policy("warn", 0.7)
  agent1$cost <- 5.0
  agent1$messages <- list(
    list(role = "system", content = "Original"),
    list(role = "user", content = "Test message"),
    list(role = "assistant", content = "Test response")
  )

  # Clone the agent
  agent2 <- agent1$clone_agent("ClonedState")

  # Verify all state is preserved
  expect_equal(agent2$name, "ClonedState") # New name
  expect_equal(agent2$instruction, "Original") # Original instruction
  expect_equal(agent2$budget, 25.5)
  expect_equal(agent2$budget_policy$on_exceed, "warn")
  expect_equal(agent2$budget_policy$warn_at, 0.7)
  expect_equal(agent2$cost, 5.0)
  expect_equal(length(agent2$messages), 3)

  # Verify independence (different IDs)
  expect_false(agent1$agent_id == agent2$agent_id)

  # Verify independence (modifying one doesn't affect the other)
  agent2$set_budget(50)
  expect_equal(agent1$budget, 25.5) # Original unchanged
  expect_equal(agent2$budget, 50)   # Clone changed
})

#' @title Workflow
#'
#' @description
#' An R6 class for building sequential multi-agent pipelines. A `Workflow`
#' is composed of **Stations** (processing units - an `Agent`, a
#' `WorkflowAgent`, or any plain R function) connected by **Routes** (directed
#' links, optionally gated by a condition function). Execution is always
#' sequential: the output of one Station becomes the input of the next.
#'
#' Stations whose results have already been computed can be retrieved from an
#' internal cache instead of being re-executed, which is useful when iterating
#' on later parts of the pipeline without paying the cost (or latency) of
#' earlier LLM calls.
#'
#' A finished `Workflow` can be wrapped as a `WorkflowAgent` via
#' \code{$as_agent()}, making it composable with `LeadAgent` or embeddable as
#' a Station inside another `Workflow`.
#'
#' @export
Workflow <- R6::R6Class(
  "Workflow",
  cloneable = FALSE,

  # -- public -----------------------------------------------------------------
  public = list(

    #' @field name Workflow identifier.
    name = NULL,

    #' @field description Optional human-readable description.
    description = NULL,

    #' @field use_cache Whether to cache and reuse Station results.
    use_cache = NULL,

    #' @field cache Environment used as a hash-map for cached Station outputs.
    cache = NULL,

    #' @field run_history List of records from every \code{$run()} call.
    run_history = NULL,

    #' @field hitl_steps Integer vector of step numbers at which execution pauses
    #'   for human review. Set via \code{$set_hitl()}.
    hitl_steps = NULL,

    # -- initialize ------------------------------------------------------------

    #' @description Create a new `Workflow`.
    #'
    #' @param name `[character(1)]` Workflow name.
    #' @param description `[character(1)]` Optional description.
    #' @param use_cache `[logical(1)]` Enable result caching (default `TRUE`).
    initialize = function(name, description = NULL, use_cache = TRUE) {
      checkmate::assert_string(name)
      checkmate::assert_flag(use_cache)
      if (!is.null(description)) checkmate::assert_string(description)

      self$name        <- name
      self$description <- description
      self$use_cache   <- use_cache
      self$cache       <- new.env(hash = TRUE, parent = emptyenv())
      self$run_history <- list()
    },

    # -- add_station -----------------------------------------------------------

    #' @description Add a Station to the workflow.
    #'
    #' A Station is a named processing unit. Its `handler` can be:
    #' \itemize{
    #'   \item An `Agent` or `WorkflowAgent` - the Station calls
    #'         \code{handler$invoke(input)}.
    #'   \item A plain R `function(input)` - the Station calls
    #'         \code{handler(input)} and coerces the return value to
    #'         `character`.
    #' }
    #'
    #' @param name `[character(1)]` Unique Station name within this workflow.
    #' @param handler An `Agent`, `WorkflowAgent`, or `function`.
    #' @param description `[character(1)]` Optional human-readable description
    #'   (shown in \code{$visualize()}).
    #'
    #' @return Invisibly returns `self` for method chaining.
    add_station = function(name, handler, description = NULL) {
      checkmate::assert_string(name)
      if (!is.null(description)) checkmate::assert_string(description)

      is_agent    <- inherits(handler, "Agent") || inherits(handler, "WorkflowAgent")
      is_function <- is.function(handler)

      if (!is_agent && !is_function) {
        cli::cli_abort(c(
          "{.arg handler} must be an {.cls Agent}, {.cls WorkflowAgent}, or a {.cls function}.",
          "i" = "Received: {.cls {class(handler)[[1L]]}}."
        ))
      }

      if (name %in% names(private$.stations)) {
        cli::cli_abort("A Station named {.val {name}} already exists in workflow {.val {self$name}}.")
      }

      private$.stations[[name]] <- list(
        name        = name,
        handler     = handler,
        description = description
      )

      cli::cli_alert_success("Station {.val {name}} added.")
      invisible(self)
    },

    # -- add_route -------------------------------------------------------------

    #' @description Add a Route between two Stations.
    #'
    #' Routes define the execution order. An optional `condition` function
    #' receives the output of the `from` Station and must return `TRUE` or
    #' `FALSE`. Conditional routes are evaluated first (in insertion order);
    #' the first matching one is followed. If none match, the first
    #' unconditional route from that Station is used as a default.
    #'
    #' @param from `[character(1)]` Name of the source Station.
    #' @param to `[character(1)]` Name of the destination Station.
    #' @param condition `[function | NULL]` A function \code{function(output)}
    #'   returning a single logical value. `NULL` means "always follow this
    #'   route" (i.e., unconditional).
    #'
    #' @return Invisibly returns `self` for method chaining.
    add_route = function(from, to, condition = NULL) {
      checkmate::assert_string(from)
      checkmate::assert_string(to)

      if (!from %in% names(private$.stations)) {
        cli::cli_abort("Station {.val {from}} not found. Add it first with {.fn add_station}.")
      }
      if (!to %in% names(private$.stations)) {
        cli::cli_abort("Station {.val {to}} not found. Add it first with {.fn add_station}.")
      }
      if (!is.null(condition) && !is.function(condition)) {
        cli::cli_abort(
          "{.arg condition} must be a {.cls function} returning {.val TRUE}/{.val FALSE}, or {.val NULL}."
        )
      }

      private$.routes <- c(
        private$.routes,
        list(list(from = from, to = to, condition = condition))
      )

      cli::cli_alert_success("Route {.val {from}} -> {.val {to}} registered.")
      invisible(self)
    },

    # -- set_entry -------------------------------------------------------------

    #' @description Set the entry Station where execution begins.
    #'
    #' If not called, \code{$run()} defaults to the first Station added.
    #'
    #' @param station_name `[character(1)]` Name of the entry Station.
    #'
    #' @return Invisibly returns `self` for method chaining.
    set_entry = function(station_name) {
      checkmate::assert_string(station_name)

      if (!station_name %in% names(private$.stations)) {
        cli::cli_abort("Station {.val {station_name}} not found. Add it with {.fn add_station}.")
      }

      private$.entry <- station_name
      cli::cli_alert_success("Entry Station set to {.val {station_name}}.")
      invisible(self)
    },

    # -- run -------------------------------------------------------------------

    #' @description Execute the workflow sequentially.
    #'
    #' The `input` string is passed to the entry Station. Each Station's
    #' output becomes the next Station's input. Execution stops when a Station
    #' has no outgoing Route (or no Route whose condition evaluates to `TRUE`).
    #'
    #' When \code{use_cache = TRUE}, a Station whose (name, input) pair has
    #' been seen before returns the cached output without re-invoking the
    #' handler. Use \code{$clear_cache()} to force re-execution.
    #'
    #' @param input `[character(1)]` The initial prompt / payload.
    #'
    #' @return `[character(1)]` The output of the last Station executed.
    run = function(input) {
      checkmate::assert_string(input)

      if (length(private$.stations) == 0L) {
        cli::cli_abort("No Stations defined. Use {.fn add_station} first.")
      }

      entry <- if (!is.null(private$.entry)) {
        private$.entry
      } else {
        names(private$.stations)[[1L]]
      }

      if (is.null(private$.entry)) {
        cli::cli_alert_info(
          "No entry Station set - defaulting to first Station: {.val {entry}}."
        )
      }

      cli::cli_rule(left = glue::glue("Workflow: {self$name}"))

      current       <- entry
      current_input <- input
      trace         <- list()
      steps_taken   <- 0L
      max_steps     <- 500L
      result        <- NULL

      repeat {
        steps_taken <- steps_taken + 1L

        if (steps_taken > max_steps) {
          cli::cli_abort(
            "Workflow exceeded {max_steps} steps. Possible cycle in Routes."
          )
        }

        station   <- private$.stations[[current]]
        cache_key <- private$.cache_key(current, current_input)

        if (self$use_cache && exists(cache_key, envir = self$cache, inherits = FALSE)) {
          cli::cli_alert_info("[cache] Station {.val {current}}.")
          result <- get(cache_key, envir = self$cache, inherits = FALSE)
        } else {
          cli::cli_text("  {cli::col_blue('->')} Station {.val {current}}")
          result <- private$.invoke_handler(station$handler, current_input)

          if (!is.null(self$hitl_steps) && steps_taken %in% self$hitl_steps) {
            result <- private$.human_confirm(steps_taken, current, current_input, result)
          }

          if (self$use_cache) {
            assign(cache_key, result, envir = self$cache)
          }
        }

        trace[[length(trace) + 1L]] <- list(
          step    = steps_taken,
          station = current,
          input   = current_input,
          output  = result
        )

        next_station <- private$.get_next_station(current, result)

        if (is.null(next_station)) break

        current       <- next_station
        current_input <- result
      }

      cli::cli_alert_success(
        "Workflow {.val {self$name}} completed in {steps_taken} step(s)."
      )

      self$run_history[[length(self$run_history) + 1L]] <- list(
        input  = input,
        output = result,
        steps  = steps_taken,
        trace  = trace
      )

      result
    },

    # -- set_hitl --------------------------------------------------------------

    #' @description Set Human-In-The-Loop (HITL) pause points.
    #'
    #' When execution reaches a step whose number is listed in `steps`, it
    #' pauses and presents the human with three choices:
    #' \enumerate{
    #'   \item Continue with the Station's original output.
    #'   \item Edit the output manually before the next Station receives it.
    #'   \item Stop the workflow immediately (raises an error).
    #' }
    #' HITL only fires on fresh Station executions - cache hits are skipped.
    #' Steps are numbered from 1 in execution order, matching the step counter
    #' shown in \code{$run()} output. You can set multiple steps at once:
    #' \code{wf$set_hitl(c(1, 3))}.
    #'
    #' @param steps `[integerish]` One or more step numbers (>= 1).
    #'
    #' @return Invisibly returns `self` for method chaining.
    set_hitl = function(steps) {
      checkmate::assert_integerish(steps, lower = 1L, any.missing = FALSE)
      self$hitl_steps <- unique(as.integer(steps))
      cli::cli_alert_success(
        "HITL enabled at step(s): {.val {toString(self$hitl_steps)}}."
      )
      invisible(self)
    },

    # -- clear_cache -----------------------------------------------------------

    #' @description Remove all cached Station results.
    #'
    #' @return Invisibly returns `self`.
    clear_cache = function() {
      cached <- ls(self$cache)
      if (length(cached) > 0L) rm(list = cached, envir = self$cache)
      cli::cli_alert_success("Cache cleared for workflow {.val {self$name}}.")
      invisible(self)
    },

    # -- as_agent --------------------------------------------------------------

    #' @description Wrap this Workflow as a `WorkflowAgent`.
    #'
    #' The returned `WorkflowAgent` exposes \code{$invoke(prompt)} and holds an
    #' `agent_id`, making it compatible with `LeadAgent$register_agents()` and
    #' usable as a Station handler inside another `Workflow`.
    #'
    #' @param name `[character(1)]` Name for the `WorkflowAgent`. Defaults to
    #'   \code{"\{workflow name\} (agent)"}.
    #' @param instruction `[character(1)]` Instruction string describing the
    #'   agent's role. Used by `LeadAgent` for task-agent matching.
    #'
    #' @return A `WorkflowAgent` object.
    as_agent = function(name = NULL, instruction = NULL) {
      agent_name <- if (!is.null(name)) {
        name
      } else {
        paste0(self$name, " (agent)")
      }

      agent_instruction <- if (!is.null(instruction)) {
        instruction
      } else {
        desc_part <- if (!is.null(self$description) && nchar(self$description) > 0L) {
          paste0(" ", self$description, ".")
        } else {
          ""
        }
        paste0(
          "You are a workflow agent named '", self$name, "'.",
          desc_part,
          " When invoked with a prompt, you run a sequential pipeline of",
          " specialized Stations and return the final result."
        )
      }

      wa <- WorkflowAgent$new(
        name        = agent_name,
        instruction = agent_instruction,
        workflow    = self
      )

      cli::cli_alert_success(
        "Workflow {.val {self$name}} wrapped as agent {.val {agent_name}}."
      )

      wa
    },

    # -- visualize -------------------------------------------------------------

    #' @description Render the workflow as a directed graph via DiagrammeR.
    #'
    #' Stations are shown as rounded boxes. Conditional Routes are shown as
    #' dashed arrows labelled "cond". The entry Station is marked with a filled
    #' circle labelled "START".
    #'
    #' @return A `DiagrammeR` / `htmlwidget` object.
    visualize = function() {
      if (length(private$.stations) == 0L) {
        cli::cli_abort("Nothing to visualize - no Stations have been added.")
      }

      nodes <- vapply(names(private$.stations), function(n) {
        s            <- private$.stations[[n]]
        handler_type <- if (inherits(s$handler, c("Agent", "WorkflowAgent"))) {
          class(s$handler)[[1L]]
        } else {
          "Function"
        }
        bot_label <- if (!is.null(s$description)) s$description else handler_type
        glue::glue(
          '  "{n}" [label="{n}\\n[{bot_label}]", shape=box,',
          ' style="filled,rounded", fillcolor="#AED6F1"]'
        )
      }, character(1L))

      entry_name <- if (!is.null(private$.entry)) {
        private$.entry
      } else {
        names(private$.stations)[[1L]]
      }

      entry_def <- c(
        '  "START" [label="START", shape=circle, style=filled, fillcolor="#82E0AA", width=0.5]',
        glue::glue('  "START" -> "{entry_name}"')
      )

      routes <- vapply(private$.routes, function(r) {
        attrs <- if (!is.null(r$condition)) {
          ' [label="  cond", style=dashed]'
        } else {
          ""
        }
        glue::glue('  "{r$from}" -> "{r$to}"{attrs}')
      }, character(1L))

      dot <- paste0(
        "digraph workflow {\n",
        '  graph [rankdir=TB, label="', self$name,
        '", labelloc=t, fontsize=18, fontname="Helvetica"]\n',
        "  node [fontname=\"Helvetica\", fontsize=12]\n",
        "  edge [fontname=\"Helvetica\", fontsize=10]\n",
        paste(nodes,     collapse = "\n"), "\n",
        paste(entry_def, collapse = "\n"), "\n",
        if (length(routes) > 0L) paste(routes, collapse = "\n") else "",
        "\n}"
      )

      DiagrammeR::grViz(dot)
    }
  ),

  # -- private ----------------------------------------------------------------
  private = list(
    .stations = list(),
    .routes   = list(),
    .entry    = NULL,

    # Resolve the next Station given the current one and its output.
    # Conditional Routes are tried first (in insertion order); the first
    # whose condition returns TRUE wins. If none match, the first
    # unconditional Route is used as default. Returns NULL if the workflow
    # should stop.
    .get_next_station = function(current, result) {
      from_here <- Filter(function(r) r$from == current, private$.routes)

      if (length(from_here) == 0L) return(NULL)

      cond_routes   <- Filter(function(r) !is.null(r$condition), from_here)
      uncond_routes <- Filter(function(r)  is.null(r$condition), from_here)

      for (route in cond_routes) {
        ok <- tryCatch(
          isTRUE(route$condition(result)),
          error = function(e) {
            cli::cli_alert_warning(
              "Condition on route {.val {route$from}} -> {.val {route$to}} errored: {e$message}"
            )
            FALSE
          }
        )
        if (ok) return(route$to)
      }

      if (length(uncond_routes) > 0L) return(uncond_routes[[1L]]$to)

      NULL
    },

    # Dispatch a handler call: Agent/WorkflowAgent use $invoke(); plain
    # functions are called directly and their output coerced to character.
    .invoke_handler = function(handler, input) {
      if (inherits(handler, c("Agent", "WorkflowAgent"))) {
        handler$invoke(input)
      } else {
        out <- handler(input)
        if (!is.character(out)) as.character(out) else out
      }
    },

    # Pause execution, show the station's input and output, then ask the human
    # what to do. Returns the result to continue with (original or edited).
    .human_confirm = function(step_index, station_name, input, result) {
      cli::cli_rule(left = glue::glue("HITL - Step {step_index}"))
      cli::cli_text("Station: {.strong {station_name}}")
      cli::cli_alert_info("Input:")
      cli::cli_verbatim(input)
      cli::cli_alert_info("Output:")
      cli::cli_verbatim(result)

      cli::cli_text(cli::cli_ul(c(
        "[1] Continue with this output",
        "[2] Edit the output",
        "[3] Stop the workflow"
      )))

      repeat {
        answer <- readline("Your choice [1/2/3]: ")
        if (nzchar(answer) && answer %in% c("1", "2", "3")) break
        cli::cli_alert_warning("Invalid input. Please enter 1, 2, or 3.")
      }

      if (answer == "2") {
        new_result <- readline("(->) Enter edited output: ")
        cli::cli_alert_success("Output updated.")
        return(new_result)
      } else if (answer == "3") {
        cli::cli_alert_danger("Workflow stopped by user at step {step_index}.")
        cli::cli_abort("HITL: Execution stopped by user.")
      } else {
        cli::cli_alert_success("Continuing with original output.")
        return(result)
      }
    },

    # Build a deterministic cache key from a Station name and its input.
    # The fixed separator is unlikely to appear in normal station names or prompts.
    .cache_key = function(station_name, input) {
      paste0(station_name, "|||", substr(input, 1L, 1024L))
    }
  )
)


# ==============================================================================

#' @title WorkflowAgent
#'
#' @description
#' A lightweight, Agent-compatible wrapper around a `Workflow`. Created via
#' \code{Workflow$as_agent()}.
#'
#' `WorkflowAgent` exposes the same interface that `LeadAgent` expects from
#' any agent (\code{name}, \code{instruction}, \code{agent_id},
#' \code{$invoke()}), so it can be:
#' \itemize{
#'   \item Registered with \code{LeadAgent$register_agents()}.
#'   \item Used as a Station handler inside another `Workflow`.
#' }
#'
#' Do not instantiate `WorkflowAgent` directly - use \code{Workflow$as_agent()}.
#'
#' @export
WorkflowAgent <- R6::R6Class(
  "WorkflowAgent",
  cloneable = FALSE,

  public = list(

    #' @field name Agent name (visible to `LeadAgent` for task matching).
    name = NULL,

    #' @field instruction System instruction describing the agent's role.
    instruction = NULL,

    #' @field agent_id Unique identifier (UUID).
    agent_id = NULL,

    #' @field workflow The underlying `Workflow` object.
    workflow = NULL,

    #' @field messages Conversation history as a list of
    #'   \code{list(role, content)} entries.
    messages = NULL,

    #' @description Create a new `WorkflowAgent`. Prefer \code{Workflow$as_agent()}.
    #'
    #' @param name `[character(1)]` Agent name.
    #' @param instruction `[character(1)]` Instruction / system prompt.
    #' @param workflow A `Workflow` object.
    initialize = function(name, instruction, workflow) {
      checkmate::assert_string(name)
      checkmate::assert_string(instruction)

      if (!inherits(workflow, "Workflow")) {
        cli::cli_abort("{.arg workflow} must be a {.cls Workflow} object.")
      }

      self$name        <- name
      self$instruction <- instruction
      self$agent_id    <- uuid::UUIDgenerate()
      self$workflow    <- workflow
      self$messages    <- list()
    },

    #' @description Run the underlying workflow with `prompt` as input.
    #'
    #' @param prompt `[character(1)]` The user prompt / input payload.
    #'
    #' @return `[character(1)]` The final output of the workflow.
    invoke = function(prompt) {
      checkmate::assert_string(prompt)

      self$messages <- c(
        self$messages,
        list(list(role = "user", content = prompt))
      )

      result <- self$workflow$run(prompt)

      self$messages <- c(
        self$messages,
        list(list(role = "assistant", content = result))
      )

      result
    },

    #' @description Clear the conversation history stored in \code{$messages}.
    #'
    #' @return Invisibly returns `self`.
    reset_conversation_history = function() {
      self$messages <- list()
      invisible(self)
    }
  )
)

library(testthat)
library(mini007)

# ── Helpers ───────────────────────────────────────────────────────────────────

# Minimal R6 mock that passes inherits(x, "Agent") check and records calls.
MockAgent <- R6::R6Class(
  classname = "Agent",
  public = list(
    call_count = 0L,
    response   = NULL,
    initialize = function(response = "mock output") {
      self$response <- response
    },
    invoke = function(input) {
      self$call_count <- self$call_count + 1L
      self$response
    }
  )
)

# ── Workflow: initialization ──────────────────────────────────────────────────

test_that("Workflow initializes with correct defaults", {
  wf <- Workflow$new("MyFlow")
  expect_equal(wf$name, "MyFlow")
  expect_null(wf$description)
  expect_true(wf$use_cache)
  expect_null(wf$hitl_steps)
  expect_length(wf$run_history, 0L)
  expect_true(is.environment(wf$cache))
})

test_that("Workflow initializes with description and use_cache = FALSE", {
  wf <- Workflow$new("F", description = "A test workflow", use_cache = FALSE)
  expect_equal(wf$description, "A test workflow")
  expect_false(wf$use_cache)
})

test_that("Workflow$new errors on non-string name", {
  expect_error(Workflow$new(123))
  expect_error(Workflow$new(NULL))
})

test_that("Workflow$new errors on non-logical use_cache", {
  expect_error(Workflow$new("W", use_cache = "yes"))
  expect_error(Workflow$new("W", use_cache = 1L))
})

test_that("Workflow$new errors on non-string description", {
  expect_error(Workflow$new("W", description = 123))
  expect_error(Workflow$new("W", description = TRUE))
})

# ── add_station ───────────────────────────────────────────────────────────────

test_that("add_station accepts an Agent handler", {
  wf <- Workflow$new("W")
  expect_no_error(wf$add_station("s1", MockAgent$new()))
})

test_that("add_station accepts a function handler", {
  wf <- Workflow$new("W")
  expect_no_error(wf$add_station("s1", toupper))
})

test_that("add_station returns self invisibly for chaining", {
  wf     <- Workflow$new("W")
  result <- wf$add_station("s1", toupper)
  expect_identical(result, wf)
})

test_that("add_station errors on duplicate station name", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_error(wf$add_station("s1", tolower), "already exists")
})

test_that("add_station errors when handler is neither Agent nor function", {
  wf <- Workflow$new("W")
  expect_error(wf$add_station("s1", 42L),    "handler")
  expect_error(wf$add_station("s1", list()), "handler")
  expect_error(wf$add_station("s1", TRUE),   "handler")
})

test_that("add_station errors on non-string name", {
  wf <- Workflow$new("W")
  expect_error(wf$add_station(123, toupper))
})

test_that("add_station accepts an optional description string", {
  wf <- Workflow$new("W")
  expect_no_error(wf$add_station("s1", toupper, description = "Uppercase step"))
})

test_that("add_station errors on non-string description", {
  wf <- Workflow$new("W")
  expect_error(wf$add_station("s1", toupper, description = 99))
})

# ── add_route ─────────────────────────────────────────────────────────────────

test_that("add_route registers a route and returns self invisibly", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  result <- wf$add_route("s1", "s2")
  expect_identical(result, wf)
})

test_that("add_route errors when from station does not exist", {
  wf <- Workflow$new("W")
  wf$add_station("s2", tolower)
  expect_error(wf$add_route("s1", "s2"), "not found")
})

test_that("add_route errors when to station does not exist", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_error(wf$add_route("s1", "s2"), "not found")
})

test_that("add_route accepts a valid condition function", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  expect_no_error(wf$add_route("s1", "s2", condition = function(x) nchar(x) > 5))
})

test_that("add_route errors when condition is not NULL and not a function", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  expect_error(wf$add_route("s1", "s2", condition = TRUE),      "condition")
  expect_error(wf$add_route("s1", "s2", condition = "always"),  "condition")
  expect_error(wf$add_route("s1", "s2", condition = 1L),        "condition")
})

test_that("add_route allows multiple routes from the same station", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$add_station("s3", function(x) x)
  expect_no_error({
    wf$add_route("s1", "s2", condition = function(x) TRUE)
    wf$add_route("s1", "s3")
  })
})

# ── set_entry ─────────────────────────────────────────────────────────────────

test_that("set_entry sets the entry station and returns self invisibly", {
  wf     <- Workflow$new("W")
  wf$add_station("s1", toupper)
  result <- wf$set_entry("s1")
  expect_identical(result, wf)
})

test_that("set_entry errors when station not found", {
  wf <- Workflow$new("W")
  expect_error(wf$set_entry("missing"), "not found")
})

# ── run: basic execution ──────────────────────────────────────────────────────

test_that("run errors when no stations are defined", {
  wf <- Workflow$new("W")
  expect_error(wf$run("hello"), "No Stations")
})

test_that("run errors on non-string input", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_error(wf$run(123))
  expect_error(wf$run(NULL))
  expect_error(wf$run(c("a", "b")))
})

test_that("run executes a single function station and returns its output", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_equal(wf$run("hello"), "HELLO")
})

test_that("run passes each station's output as the next station's input", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", function(x) paste("got:", x))
  wf$add_route("s1", "s2")
  expect_equal(wf$run("hello"), "got: HELLO")
})

test_that("run executes an Agent station via invoke()", {
  agent <- MockAgent$new("agent reply")
  wf    <- Workflow$new("W")
  wf$add_station("s1", agent)
  expect_equal(wf$run("anything"), "agent reply")
  expect_equal(agent$call_count, 1L)
})

test_that("run coerces non-character function output to character", {
  wf <- Workflow$new("W")
  wf$add_station("s1", function(x) 42L)
  result <- wf$run("input")
  expect_type(result, "character")
  expect_equal(result, "42")
})

test_that("run uses the explicitly set entry station", {
  wf <- Workflow$new("W")
  wf$add_station("s1", function(x) paste("s1:", x))
  wf$add_station("s2", function(x) paste("s2:", x))
  wf$set_entry("s2")
  expect_equal(wf$run("in"), "s2: in")
})

test_that("run defaults to first added station when no entry is set", {
  wf <- Workflow$new("W")
  wf$add_station("first",  function(x) paste("FIRST:",  x))
  wf$add_station("second", function(x) paste("SECOND:", x))
  expect_equal(wf$run("x"), "FIRST: x")
})

test_that("run stops at station with no outgoing route", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  # No route from s1 — workflow ends after s1
  expect_equal(wf$run("hello"), "HELLO")
})

test_that("run handles a three-station chain correctly", {
  wf <- Workflow$new("W")
  wf$add_station("a", toupper)
  wf$add_station("b", function(x) paste0("[", x, "]"))
  wf$add_station("c", function(x) paste("done:", x))
  wf$add_route("a", "b")
  wf$add_route("b", "c")
  expect_equal(wf$run("hi"), "done: [HI]")
})

# ── run: history and trace ────────────────────────────────────────────────────

test_that("run appends a record to run_history after each call", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$run("hello")
  wf$run("world")
  expect_length(wf$run_history, 2L)
  expect_equal(wf$run_history[[1L]]$input,  "hello")
  expect_equal(wf$run_history[[1L]]$output, "HELLO")
  expect_equal(wf$run_history[[2L]]$input,  "world")
  expect_equal(wf$run_history[[2L]]$output, "WORLD")
})

test_that("run history trace contains per-step station and I/O info", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$add_route("s1", "s2")
  wf$run("Hello")
  trace <- wf$run_history[[1L]]$trace
  expect_length(trace, 2L)
  expect_equal(trace[[1L]]$step,    1L)
  expect_equal(trace[[1L]]$station, "s1")
  expect_equal(trace[[1L]]$input,   "Hello")
  expect_equal(trace[[1L]]$output,  "HELLO")
  expect_equal(trace[[2L]]$step,    2L)
  expect_equal(trace[[2L]]$station, "s2")
  expect_equal(trace[[2L]]$input,   "HELLO")
  expect_equal(trace[[2L]]$output,  "hello")
})

test_that("run history steps count is correct", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$add_station("s3", function(x) x)
  wf$add_route("s1", "s2")
  wf$add_route("s2", "s3")
  wf$run("hi")
  expect_equal(wf$run_history[[1L]]$steps, 3L)
})

# ── run: caching ──────────────────────────────────────────────────────────────

test_that("run caches output so repeated identical runs skip re-execution", {
  calls <- 0L
  wf <- Workflow$new("W", use_cache = TRUE)
  wf$add_station("s1", function(x) { calls <<- calls + 1L; toupper(x) })
  wf$run("hello")
  wf$run("hello")
  expect_equal(calls, 1L)
})

test_that("run re-executes when input differs even with cache enabled", {
  calls <- 0L
  wf <- Workflow$new("W", use_cache = TRUE)
  wf$add_station("s1", function(x) { calls <<- calls + 1L; toupper(x) })
  wf$run("hello")
  wf$run("world")
  expect_equal(calls, 2L)
})

test_that("run with use_cache = FALSE always re-executes", {
  calls <- 0L
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) { calls <<- calls + 1L; toupper(x) })
  wf$run("hello")
  wf$run("hello")
  expect_equal(calls, 2L)
})

test_that("cached run still returns the correct output", {
  wf <- Workflow$new("W", use_cache = TRUE)
  wf$add_station("s1", toupper)
  wf$run("hello")
  expect_equal(wf$run("hello"), "HELLO")
})

# ── clear_cache ───────────────────────────────────────────────────────────────

test_that("clear_cache forces re-execution on the next run", {
  calls <- 0L
  wf <- Workflow$new("W", use_cache = TRUE)
  wf$add_station("s1", function(x) { calls <<- calls + 1L; toupper(x) })
  wf$run("hello")
  wf$clear_cache()
  wf$run("hello")
  expect_equal(calls, 2L)
})

test_that("clear_cache returns self invisibly", {
  wf     <- Workflow$new("W")
  result <- wf$clear_cache()
  expect_identical(result, wf)
})

test_that("clear_cache is safe when cache is already empty", {
  wf <- Workflow$new("W")
  expect_no_error(wf$clear_cache())
})

# ── conditional routes ────────────────────────────────────────────────────────

test_that("matching conditional route is followed", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "long enough string")
  wf$add_station("s2", function(x) paste("LONG:", x))
  wf$add_station("s3", function(x) paste("SHORT:", x))
  wf$add_route("s1", "s2", condition = function(x) nchar(x) >= 10)
  wf$add_route("s1", "s3")
  expect_equal(wf$run("go"), "LONG: long enough string")
})

test_that("unconditional fallback route is used when no condition matches", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "hi")
  wf$add_station("s2", function(x) paste("LONG:", x))
  wf$add_station("s3", function(x) paste("SHORT:", x))
  wf$add_route("s1", "s2", condition = function(x) nchar(x) >= 100)
  wf$add_route("s1", "s3")
  expect_equal(wf$run("go"), "SHORT: hi")
})

test_that("first matching conditional route wins when multiple conditions are true", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "abc")
  wf$add_station("s2", function(x) "route-A")
  wf$add_station("s3", function(x) "route-B")
  wf$add_route("s1", "s2", condition = function(x) TRUE)
  wf$add_route("s1", "s3", condition = function(x) TRUE)
  expect_equal(wf$run("go"), "route-A")
})

test_that("erroring condition is treated as FALSE without crashing", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "out")
  wf$add_station("s2", function(x) paste("next:", x))
  wf$add_route("s1", "s2", condition = function(x) stop("boom"))
  # Condition errors → treated as FALSE, no unconditional fallback → workflow stops
  expect_equal(wf$run("in"), "out")
})

test_that("workflow stops when no outgoing route matches", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "short")
  wf$add_station("s2", function(x) "never reached")
  wf$add_route("s1", "s2", condition = function(x) nchar(x) > 100)
  # No unconditional fallback
  expect_equal(wf$run("go"), "short")
})

# ── set_hitl ─────────────────────────────────────────────────────────────────

test_that("set_hitl sets hitl_steps and returns self invisibly", {
  wf     <- Workflow$new("W")
  result <- wf$set_hitl(1L)
  expect_equal(wf$hitl_steps, 1L)
  expect_identical(result, wf)
})

test_that("set_hitl accepts multiple steps and deduplicates", {
  wf <- Workflow$new("W")
  wf$set_hitl(c(1L, 2L, 2L, 3L))
  expect_equal(wf$hitl_steps, c(1L, 2L, 3L))
})

test_that("set_hitl coerces numeric to integer", {
  wf <- Workflow$new("W")
  wf$set_hitl(c(1, 2))
  expect_type(wf$hitl_steps, "integer")
})

test_that("set_hitl errors on non-integerish steps", {
  wf <- Workflow$new("W")
  expect_error(wf$set_hitl("one"))
  expect_error(wf$set_hitl(list(1)))
})

test_that("set_hitl errors on steps lower than 1", {
  wf <- Workflow$new("W")
  expect_error(wf$set_hitl(0L))
  expect_error(wf$set_hitl(-1L))
})

test_that("HITL choice 1 continues with original station output", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "original")
  wf$set_hitl(1L)
  local_mocked_bindings(
    readline = function(prompt) "1",
    .package = "base"
  )
  expect_equal(wf$run("input"), "original")
})

test_that("HITL choice 2 replaces station output with edited text", {
  calls <- 0L
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "original")
  wf$set_hitl(1L)
  local_mocked_bindings(
    readline = function(prompt) {
      calls <<- calls + 1L
      if (calls == 1L) "2" else "edited output"
    },
    .package = "base"
  )
  expect_equal(wf$run("input"), "edited output")
})

test_that("HITL choice 3 aborts the workflow with an error", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) "output")
  wf$set_hitl(1L)
  local_mocked_bindings(
    readline = function(prompt) "3",
    .package = "base"
  )
  expect_error(wf$run("input"), "[Ss]topped by user")
})

test_that("HITL fires only on the specified step, not others", {
  hitl_fired <- FALSE
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", function(x) paste("A:", x))
  wf$add_station("s2", function(x) paste("B:", x))
  wf$add_route("s1", "s2")
  wf$set_hitl(2L)  # only step 2
  local_mocked_bindings(
    readline = function(prompt) { hitl_fired <<- TRUE; "1" },
    .package = "base"
  )
  result <- wf$run("in")
  expect_true(hitl_fired)
  expect_equal(result, "B: A: in")
})

test_that("HITL is skipped on cache hit", {
  hitl_called <- FALSE
  wf <- Workflow$new("W", use_cache = TRUE)
  wf$add_station("s1", function(x) "out")
  wf$set_hitl(1L)
  # First run — HITL fires
  local_mocked_bindings(
    readline = function(prompt) "1",
    .package = "base"
  )
  wf$run("hello")
  # Second run — cache hit, HITL must NOT fire
  local_mocked_bindings(
    readline = function(prompt) { hitl_called <<- TRUE; "1" },
    .package = "base"
  )
  wf$run("hello")
  expect_false(hitl_called)
})

# ── as_agent ─────────────────────────────────────────────────────────────────

test_that("as_agent returns a WorkflowAgent", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_s3_class(wf$as_agent(), "WorkflowAgent")
})

test_that("as_agent defaults name to '<workflow name> (agent)'", {
  wf <- Workflow$new("MyFlow")
  wf$add_station("s1", toupper)
  expect_equal(wf$as_agent()$name, "MyFlow (agent)")
})

test_that("as_agent uses a custom name when provided", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_equal(wf$as_agent(name = "CustomAgent")$name, "CustomAgent")
})

test_that("as_agent assigns a non-null character UUID", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_type(wa$agent_id, "character")
  expect_false(is.null(wa$agent_id))
  expect_gt(nchar(wa$agent_id), 0L)
})

test_that("as_agent keeps a reference to the underlying workflow", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_identical(wa$workflow, wf)
})

test_that("as_agent uses a custom instruction when provided", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent(instruction = "Do the thing.")
  expect_equal(wa$instruction, "Do the thing.")
})

test_that("as_agent embeds workflow description in default instruction", {
  wf <- Workflow$new("W", description = "does something special")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_true(grepl("does something special", wa$instruction))
})

test_that("as_agent default instruction includes the workflow name", {
  wf <- Workflow$new("UniqueFlowName")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_true(grepl("UniqueFlowName", wa$instruction))
})

test_that("two as_agent() calls produce distinct agent_ids", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa1 <- wf$as_agent()
  wa2 <- wf$as_agent()
  expect_false(identical(wa1$agent_id, wa2$agent_id))
})

# ── WorkflowAgent ─────────────────────────────────────────────────────────────

test_that("WorkflowAgent initializes empty messages list", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_length(wa$messages, 0L)
})

test_that("WorkflowAgent$new errors when workflow is not a Workflow object", {
  expect_error(
    WorkflowAgent$new("A", "I", workflow = list()),
    "Workflow"
  )
})

test_that("WorkflowAgent$new errors on non-string name", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_error(WorkflowAgent$new(123, "I", wf))
})

test_that("WorkflowAgent$new errors on non-string instruction", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  expect_error(WorkflowAgent$new("A", 123, wf))
})

test_that("WorkflowAgent$invoke runs the workflow and returns its output", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  expect_equal(wa$invoke("hello"), "HELLO")
})

test_that("WorkflowAgent$invoke appends one user and one assistant message", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  wa$invoke("hello")
  expect_length(wa$messages, 2L)
  expect_equal(wa$messages[[1L]]$role,    "user")
  expect_equal(wa$messages[[1L]]$content, "hello")
  expect_equal(wa$messages[[2L]]$role,    "assistant")
  expect_equal(wa$messages[[2L]]$content, "HELLO")
})

test_that("WorkflowAgent$invoke accumulates messages across multiple calls", {
  wf <- Workflow$new("W", use_cache = FALSE)
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  wa$invoke("hello")
  wa$invoke("world")
  expect_length(wa$messages, 4L)
  expect_equal(wa$messages[[3L]]$content, "world")
  expect_equal(wa$messages[[4L]]$content, "WORLD")
})

test_that("WorkflowAgent$reset_conversation_history clears all messages", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa <- wf$as_agent()
  wa$invoke("hello")
  wa$reset_conversation_history()
  expect_length(wa$messages, 0L)
})

test_that("WorkflowAgent$reset_conversation_history returns self invisibly", {
  wf     <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wa     <- wf$as_agent()
  result <- wa$reset_conversation_history()
  expect_identical(result, wa)
})

test_that("WorkflowAgent used as a station handler inside another Workflow", {
  inner_wf <- Workflow$new("inner")
  inner_wf$add_station("s1", toupper)
  inner_agent <- inner_wf$as_agent()

  outer_wf <- Workflow$new("outer")
  outer_wf$add_station("step1", inner_agent)
  expect_equal(outer_wf$run("hello"), "HELLO")
})

# ── visualize ─────────────────────────────────────────────────────────────────

test_that("visualize errors when no stations are defined", {
  wf <- Workflow$new("W")
  expect_error(wf$visualize(), "Nothing to visualize")
})

test_that("visualize returns an htmlwidget / grViz object", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  g <- wf$visualize()
  expect_true(inherits(g, "htmlwidget") || inherits(g, "grViz"))
})

test_that("visualize DOT output contains all station names", {
  wf <- Workflow$new("W")
  wf$add_station("alpha", toupper)
  wf$add_station("beta",  tolower)
  wf$add_route("alpha", "beta")
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_true(grepl("alpha", dot))
  expect_true(grepl("beta",  dot))
})

test_that("visualize DOT output contains a START node", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_true(grepl("START", dot))
})

test_that("visualize marks conditional routes as dashed", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$add_route("s1", "s2", condition = function(x) TRUE)
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_true(grepl("dashed", dot))
})

test_that("visualize unconditional routes have no dashed style", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$add_route("s1", "s2")
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_false(grepl("dashed", dot))
})

test_that("visualize includes the workflow name as graph label", {
  wf <- Workflow$new("SpecialFlowName")
  wf$add_station("s1", toupper)
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_true(grepl("SpecialFlowName", dot))
})

test_that("visualize uses explicit entry station for START arrow", {
  wf <- Workflow$new("W")
  wf$add_station("s1", toupper)
  wf$add_station("s2", tolower)
  wf$set_entry("s2")
  g   <- wf$visualize()
  dot <- g$x$diagram
  expect_true(grepl('START.*->.*"s2"', dot) || grepl('"START".*->.*"s2"', dot))
})

# ── method chaining ───────────────────────────────────────────────────────────

test_that("full method-chaining pipeline works end-to-end", {
  result <- Workflow$new("Chain")$
    add_station("s1", toupper)$
    add_station("s2", function(x) paste0("[", x, "]"))$
    add_route("s1", "s2")$
    run("hello")
  expect_equal(result, "[HELLO]")
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LeadAgent.R
\name{LeadAgent}
\alias{LeadAgent}
\title{LeadAgent: A Multi-Agent Orchestration Coordinator}
\description{
`LeadAgent` extends `Agent` to coordinate a group of specialized agents.
It decomposes complex prompts into subtasks using LLMs and assigns each subtask to the most suitable registered agent.
The lead agent handles response chaining, where each agent can consider prior results.
}
\details{
This class builds intelligent multi-agent workflows by delegating sub-tasks using `delegate_prompt()`,
executing them with `invoke()`, and storing the results in the `agents_interaction` list.
}
\examples{

## ------------------------------------------------
## Method `LeadAgent$new`
## ------------------------------------------------


  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )



## ------------------------------------------------
## Method `LeadAgent$clear_agents`
## ------------------------------------------------

  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = paste0(
   "You are an agent designed to summarise ",
   "a given text into 3 distinct bullet points."
   ),
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )
 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents

 lead_agent$clear_agents()

 lead_agent$agents


## ------------------------------------------------
## Method `LeadAgent$remove_agents`
## ------------------------------------------------

  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )


 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents

 # deleting the translator agent

 id_translator_agent <- translator$agent_id

 lead_agent$remove_agents(id_translator_agent)

 lead_agent$agents


## ------------------------------------------------
## Method `LeadAgent$register_agents`
## ------------------------------------------------

  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents

## ------------------------------------------------
## Method `LeadAgent$invoke`
## ------------------------------------------------

\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed ",
   "and accurate information. Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$invoke(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}

## ------------------------------------------------
## Method `LeadAgent$generate_plan`
## ------------------------------------------------

\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. Your job is to answer factual questions ",
   "with detailed and accurate information. Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$generate_plan(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}

## ------------------------------------------------
## Method `LeadAgent$broadcast`
## ------------------------------------------------

\dontrun{
 # An API KEY is required in order to invoke the agents
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

openai_4_1_agent <- Agent$new(
  name = "openai_4_1_agent",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

openai_4_1_nano_agent <- Agent$new(
  name = "openai_4_1_nano_agent",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1_nano
  )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

lead_agent$register_agents(c(openai_4_1_agent, openai_4_1_nano_agent))
lead_agent$broadcast(
  prompt = paste0(
    "If I were Algerian, which song would I like to sing ",
    "when running under the rain? how about a flower?"
  )
  )
}

## ------------------------------------------------
## Method `LeadAgent$set_hitl`
## ------------------------------------------------

\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
    "You are a research assistant. ",
    "Your job is to answer factual questions with detailed and accurate information. ",
    "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = paste0(
   "You are agent designed to summarise a give text ",
   "into 3 distinct bullet points."
   ),
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 # setting a human in the loop in step 2
 lead_agent$set_hitl(1)

 # The execution will stop at step 2 and a human will be able
 # to either accept the answer, modify it or stop the execution of
 # the workflow

 lead_agent$invoke(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}

## ------------------------------------------------
## Method `LeadAgent$judge_and_choose_best_response`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

stylist <- Agent$new(
  name = "stylist",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

stylist2 <- Agent$new(
  name = "stylist2",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1_nano
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(stylist, stylist2))

lead_agent$judge_and_choose_best_response("what's the best way to war a kalvin klein shirt?")

}

## ------------------------------------------------
## Method `LeadAgent$agents_dialog`
## ------------------------------------------------

\dontrun{
# An API KEY is required in order to invoke the agents
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

creative_writer <- Agent$new(
  name = "creative_writer",
  instruction = "You are a creative writer. Focus on engaging storytelling.",
  llm_object = openai_4_1_nano
)

editor <- Agent$new(
  name = "editor",
  instruction = "You are an editor. Focus on clarity and conciseness.",
  llm_object = openai_4_1_mini
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(creative_writer, editor))

result <- lead_agent$agents_dialog(
  prompt = "Write a compelling opening sentence for a sci-fi novel.",
  agent_1_id = creative_writer$agent_id,
  agent_2_id = editor$agent_id,
  max_iterations = 3
)

# Access the final response
result$final_response

# View the dialog history
result$dialog_history
}
}
\section{Super class}{
\code{\link[mini007:Agent]{mini007::Agent}} -> \code{LeadAgent}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{agents}}{A named list of registered sub-agents (by UUID).}

\item{\code{agents_interaction}}{A list of delegated task history with agent IDs, prompts, and responses.}

\item{\code{plan}}{A list containing the most recently generated task plan.}

\item{\code{hitl_steps}}{The steps where the workflow should be stopped in order to allow for a human interaction}

\item{\code{prompt_for_plan}}{The prompt used to generate the plan.}

\item{\code{agents_for_plan}}{The agents used for the plan}

\item{\code{dialog_history}}{A list storing the history of agent dialogs}

\item{\code{broadcast_history}}{A list storing the history of broadcast interactions}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-LeadAgent-new}{\code{LeadAgent$new()}}
\item \href{#method-LeadAgent-clear_agents}{\code{LeadAgent$clear_agents()}}
\item \href{#method-LeadAgent-remove_agents}{\code{LeadAgent$remove_agents()}}
\item \href{#method-LeadAgent-register_agents}{\code{LeadAgent$register_agents()}}
\item \href{#method-LeadAgent-visualize_plan}{\code{LeadAgent$visualize_plan()}}
\item \href{#method-LeadAgent-invoke}{\code{LeadAgent$invoke()}}
\item \href{#method-LeadAgent-generate_plan}{\code{LeadAgent$generate_plan()}}
\item \href{#method-LeadAgent-broadcast}{\code{LeadAgent$broadcast()}}
\item \href{#method-LeadAgent-set_hitl}{\code{LeadAgent$set_hitl()}}
\item \href{#method-LeadAgent-judge_and_choose_best_response}{\code{LeadAgent$judge_and_choose_best_response()}}
\item \href{#method-LeadAgent-agents_dialog}{\code{LeadAgent$agents_dialog()}}
\item \href{#method-LeadAgent-clone}{\code{LeadAgent$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="add_message"><a href='../../mini007/html/Agent.html#method-Agent-add_message'><code>mini007::Agent$add_message()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="clear_and_summarise_messages"><a href='../../mini007/html/Agent.html#method-Agent-clear_and_summarise_messages'><code>mini007::Agent$clear_and_summarise_messages()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="clear_tools"><a href='../../mini007/html/Agent.html#method-Agent-clear_tools'><code>mini007::Agent$clear_tools()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="clone_agent"><a href='../../mini007/html/Agent.html#method-Agent-clone_agent'><code>mini007::Agent$clone_agent()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="export_messages_history"><a href='../../mini007/html/Agent.html#method-Agent-export_messages_history'><code>mini007::Agent$export_messages_history()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="generate_and_register_tool"><a href='../../mini007/html/Agent.html#method-Agent-generate_and_register_tool'><code>mini007::Agent$generate_and_register_tool()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="generate_execute_r_code"><a href='../../mini007/html/Agent.html#method-Agent-generate_execute_r_code'><code>mini007::Agent$generate_execute_r_code()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="get_usage_stats"><a href='../../mini007/html/Agent.html#method-Agent-get_usage_stats'><code>mini007::Agent$get_usage_stats()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="keep_last_n_messages"><a href='../../mini007/html/Agent.html#method-Agent-keep_last_n_messages'><code>mini007::Agent$keep_last_n_messages()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="list_tools"><a href='../../mini007/html/Agent.html#method-Agent-list_tools'><code>mini007::Agent$list_tools()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="load_messages_history"><a href='../../mini007/html/Agent.html#method-Agent-load_messages_history'><code>mini007::Agent$load_messages_history()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="register_tools"><a href='../../mini007/html/Agent.html#method-Agent-register_tools'><code>mini007::Agent$register_tools()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="remove_tools"><a href='../../mini007/html/Agent.html#method-Agent-remove_tools'><code>mini007::Agent$remove_tools()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="reset_conversation_history"><a href='../../mini007/html/Agent.html#method-Agent-reset_conversation_history'><code>mini007::Agent$reset_conversation_history()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="set_budget"><a href='../../mini007/html/Agent.html#method-Agent-set_budget'><code>mini007::Agent$set_budget()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="set_budget_policy"><a href='../../mini007/html/Agent.html#method-Agent-set_budget_policy'><code>mini007::Agent$set_budget_policy()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="share_context_with"><a href='../../mini007/html/Agent.html#method-Agent-share_context_with'><code>mini007::Agent$share_context_with()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="update_instruction"><a href='../../mini007/html/Agent.html#method-Agent-update_instruction'><code>mini007::Agent$update_instruction()</code></a></span></li>
<li><span class="pkg-link" data-pkg="mini007" data-topic="Agent" data-id="validate_response"><a href='../../mini007/html/Agent.html#method-Agent-validate_response'><code>mini007::Agent$validate_response()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-new"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-new}{}}}
\subsection{Method \code{new()}}{
Initializes the LeadAgent with a built-in task-decomposition prompt.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$new(name, llm_object)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{A short name for the coordinator (e.g. `"lead"`).}

\item{\code{llm_object}}{The LLM object generate by ellmer (eg. output of ellmer::chat_openai)}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{
  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )


}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-clear_agents"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-clear_agents}{}}}
\subsection{Method \code{clear_agents()}}{
Clear out the registered Agents
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$clear_agents()}\if{html}{\out{</div>}}
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = paste0(
   "You are an agent designed to summarise ",
   "a given text into 3 distinct bullet points."
   ),
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )
 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents

 lead_agent$clear_agents()

 lead_agent$agents

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-remove_agents"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-remove_agents}{}}}
\subsection{Method \code{remove_agents()}}{
Remove registered agents by IDs
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$remove_agents(agent_ids)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{agent_ids}}{The Agent ID to remove from the registered Agents}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )


 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents

 # deleting the translator agent

 id_translator_agent <- translator$agent_id

 lead_agent$remove_agents(id_translator_agent)

 lead_agent$agents

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-register_agents"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-register_agents}{}}}
\subsection{Method \code{register_agents()}}{
Register one or more agents for delegation.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$register_agents(agents)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{agents}}{A vector of `Agent` objects to register.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{  # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed and accurate information. ",
   "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$agents
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-visualize_plan"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-visualize_plan}{}}}
\subsection{Method \code{visualize_plan()}}{
Visualizes the orchestration plan
Each agent node is shown in sequence (left → right), with tooltips showing
the actual prompt delegated to that agent.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$visualize_plan()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-invoke"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-invoke}{}}}
\subsection{Method \code{invoke()}}{
Executes the full prompt pipeline: decomposition → delegation → invocation.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$invoke(prompt, force_regenerate_plan = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{The complex user instruction to process.}

\item{\code{force_regenerate_plan}}{If TRUE, regenerate a plan even if one exists,
defaults to FALSE.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The final response (from the last agent in the sequence).
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. ",
   "Your job is to answer factual questions with detailed ",
   "and accurate information. Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$invoke(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-generate_plan"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-generate_plan}{}}}
\subsection{Method \code{generate_plan()}}{
Generates a task execution plan without executing the subtasks.
It returns a structured list containing the subtask, the selected agent, and metadata.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$generate_plan(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{A complex instruction to be broken into subtasks.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list of lists containing agent_id, agent_name, model_name, model_provider, and the assigned prompt.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
   "You are a research assistant. Your job is to answer factual questions ",
   "with detailed and accurate information. Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = "You are agent designed to summarise a given text into 3 distinct bullet points.",
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 lead_agent$generate_plan(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-broadcast"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-broadcast}{}}}
\subsection{Method \code{broadcast()}}{
Broadcasts a prompt to all registered agents and collects their responses.
This does not affect the main agent orchestration logic or history.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$broadcast(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{A user prompt to send to all agents.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list of responses from all agents.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
 # An API KEY is required in order to invoke the agents
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

openai_4_1_agent <- Agent$new(
  name = "openai_4_1_agent",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

openai_4_1_nano_agent <- Agent$new(
  name = "openai_4_1_nano_agent",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1_nano
  )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

lead_agent$register_agents(c(openai_4_1_agent, openai_4_1_nano_agent))
lead_agent$broadcast(
  prompt = paste0(
    "If I were Algerian, which song would I like to sing ",
    "when running under the rain? how about a flower?"
  )
  )
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-set_hitl"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-set_hitl}{}}}
\subsection{Method \code{set_hitl()}}{
Set Human In The Loop (HITL) interaction at determined steps within the workflow
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$set_hitl(steps)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{steps}}{At which steps the Human In The Loop is required?}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list of responses from all agents.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
 # An API KEY is required in order to invoke the agents
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )
 researcher <- Agent$new(
   name = "researcher",
   instruction = paste0(
    "You are a research assistant. ",
    "Your job is to answer factual questions with detailed and accurate information. ",
    "Do not answer with more than 2 lines"
   ),
   llm_object = openai_4_1_mini
 )

 summarizer <- Agent$new(
   name = "summarizer",
   instruction = paste0(
   "You are agent designed to summarise a give text ",
   "into 3 distinct bullet points."
   ),
   llm_object = openai_4_1_mini
 )

 translator <- Agent$new(
   name = "translator",
   instruction = "Your role is to translate a text from English to German",
   llm_object = openai_4_1_mini
 )

 lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
 )

 lead_agent$register_agents(c(researcher, summarizer, translator))

 # setting a human in the loop in step 2
 lead_agent$set_hitl(1)

 # The execution will stop at step 2 and a human will be able
 # to either accept the answer, modify it or stop the execution of
 # the workflow

 lead_agent$invoke(
 paste0(
  "Describe the economic situation in Algeria in 3 sentences. ",
  "Answer in German"
  )
 )
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-judge_and_choose_best_response"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-judge_and_choose_best_response}{}}}
\subsection{Method \code{judge_and_choose_best_response()}}{
The Lead Agent send a prompt to its registered agents and choose the best response
from the agents' responses
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$judge_and_choose_best_response(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{The prompt to send to the registered agents}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list of responses from all agents, including the chosen response
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
openai_4_1 <- ellmer::chat(
  name = "openai/gpt-4.1",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

stylist <- Agent$new(
  name = "stylist",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1
)

openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

stylist2 <- Agent$new(
  name = "stylist2",
  instruction = "You are an AI assistant. Answer in 1 sentence max.",
  llm_object = openai_4_1_nano
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(stylist, stylist2))

lead_agent$judge_and_choose_best_response("what's the best way to war a kalvin klein shirt?")

}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-agents_dialog"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-agents_dialog}{}}}
\subsection{Method \code{agents_dialog()}}{
Facilitates a collaborative dialog between two agents to refine a response.
The agents take turns building on each other's responses until they reach
consensus or the maximum iterations are reached. Agents can signal consensus
by starting their response with "CONSENSUS:". If max iterations is reached
without consensus, the lead agent synthesizes a final response.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$agents_dialog(prompt, agent_1_id, agent_2_id, max_iterations = 5)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{The initial task or question for the agents to discuss.}

\item{\code{agent_1_id}}{The ID of the first agent to participate in the dialog.}

\item{\code{agent_2_id}}{The ID of the second agent to participate in the dialog.}

\item{\code{max_iterations}}{Maximum number of back-and-forth exchanges (default: 5).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing the final response, consensus status, and complete dialog history.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
# An API KEY is required in order to invoke the agents
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
openai_4_1_nano <- ellmer::chat(
  name = "openai/gpt-4.1-nano",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)

creative_writer <- Agent$new(
  name = "creative_writer",
  instruction = "You are a creative writer. Focus on engaging storytelling.",
  llm_object = openai_4_1_nano
)

editor <- Agent$new(
  name = "editor",
  instruction = "You are an editor. Focus on clarity and conciseness.",
  llm_object = openai_4_1_mini
)

lead_agent <- LeadAgent$new(
  name = "Leader",
  llm_object = openai_4_1_mini
)

lead_agent$register_agents(c(creative_writer, editor))

result <- lead_agent$agents_dialog(
  prompt = "Write a compelling opening sentence for a sci-fi novel.",
  agent_1_id = creative_writer$agent_id,
  agent_2_id = editor$agent_id,
  max_iterations = 3
)

# Access the final response
result$final_response

# View the dialog history
result$dialog_history
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LeadAgent-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LeadAgent-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{LeadAgent$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Workflow.R
\name{Workflow}
\alias{Workflow}
\title{Workflow}
\description{
An R6 class for building sequential multi-agent pipelines. A `Workflow`
is composed of **Stations** (processing units - an `Agent`, a
`WorkflowAgent`, or any plain R function) connected by **Routes** (directed
links, optionally gated by a condition function). Execution is always
sequential: the output of one Station becomes the input of the next.

Stations whose results have already been computed can be retrieved from an
internal cache instead of being re-executed, which is useful when iterating
on later parts of the pipeline without paying the cost (or latency) of
earlier LLM calls.

A finished `Workflow` can be wrapped as a `WorkflowAgent` via
\code{$as_agent()}, making it composable with `LeadAgent` or embeddable as
a Station inside another `Workflow`.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{name}}{Workflow identifier.}

\item{\code{description}}{Optional human-readable description.}

\item{\code{use_cache}}{Whether to cache and reuse Station results.}

\item{\code{cache}}{Environment used as a hash-map for cached Station outputs.}

\item{\code{run_history}}{List of records from every \code{$run()} call.}

\item{\code{hitl_steps}}{Integer vector of step numbers at which execution pauses
for human review. Set via \code{$set_hitl()}.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Workflow-new}{\code{Workflow$new()}}
\item \href{#method-Workflow-add_station}{\code{Workflow$add_station()}}
\item \href{#method-Workflow-add_route}{\code{Workflow$add_route()}}
\item \href{#method-Workflow-set_entry}{\code{Workflow$set_entry()}}
\item \href{#method-Workflow-run}{\code{Workflow$run()}}
\item \href{#method-Workflow-set_hitl}{\code{Workflow$set_hitl()}}
\item \href{#method-Workflow-clear_cache}{\code{Workflow$clear_cache()}}
\item \href{#method-Workflow-as_agent}{\code{Workflow$as_agent()}}
\item \href{#method-Workflow-visualize}{\code{Workflow$visualize()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-new"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-new}{}}}
\subsection{Method \code{new()}}{
Create a new `Workflow`.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$new(name, description = NULL, use_cache = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{`[character(1)]` Workflow name.}

\item{\code{description}}{`[character(1)]` Optional description.}

\item{\code{use_cache}}{`[logical(1)]` Enable result caching (default `TRUE`).}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-add_station"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-add_station}{}}}
\subsection{Method \code{add_station()}}{
Add a Station to the workflow.

A Station is a named processing unit. Its `handler` can be:
\itemize{
  \item An `Agent` or `WorkflowAgent` - the Station calls
        \code{handler$invoke(input)}.
  \item A plain R `function(input)` - the Station calls
        \code{handler(input)} and coerces the return value to
        `character`.
}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$add_station(name, handler, description = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{`[character(1)]` Unique Station name within this workflow.}

\item{\code{handler}}{An `Agent`, `WorkflowAgent`, or `function`.}

\item{\code{description}}{`[character(1)]` Optional human-readable description
(shown in \code{$visualize()}).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Invisibly returns `self` for method chaining.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-add_route"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-add_route}{}}}
\subsection{Method \code{add_route()}}{
Add a Route between two Stations.

Routes define the execution order. An optional `condition` function
receives the output of the `from` Station and must return `TRUE` or
`FALSE`. Conditional routes are evaluated first (in insertion order);
the first matching one is followed. If none match, the first
unconditional route from that Station is used as a default.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$add_route(from, to, condition = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{from}}{`[character(1)]` Name of the source Station.}

\item{\code{to}}{`[character(1)]` Name of the destination Station.}

\item{\code{condition}}{`[function | NULL]` A function \code{function(output)}
returning a single logical value. `NULL` means "always follow this
route" (i.e., unconditional).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Invisibly returns `self` for method chaining.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-set_entry"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-set_entry}{}}}
\subsection{Method \code{set_entry()}}{
Set the entry Station where execution begins.

If not called, \code{$run()} defaults to the first Station added.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$set_entry(station_name)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{station_name}}{`[character(1)]` Name of the entry Station.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Invisibly returns `self` for method chaining.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-run"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-run}{}}}
\subsection{Method \code{run()}}{
Execute the workflow sequentially.

The `input` string is passed to the entry Station. Each Station's
output becomes the next Station's input. Execution stops when a Station
has no outgoing Route (or no Route whose condition evaluates to `TRUE`).

When \code{use_cache = TRUE}, a Station whose (name, input) pair has
been seen before returns the cached output without re-invoking the
handler. Use \code{$clear_cache()} to force re-execution.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$run(input)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{input}}{`[character(1)]` The initial prompt / payload.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
`[character(1)]` The output of the last Station executed.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-set_hitl"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-set_hitl}{}}}
\subsection{Method \code{set_hitl()}}{
Set Human-In-The-Loop (HITL) pause points.

When execution reaches a step whose number is listed in `steps`, it
pauses and presents the human with three choices:
\enumerate{
  \item Continue with the Station's original output.
  \item Edit the output manually before the next Station receives it.
  \item Stop the workflow immediately (raises an error).
}
HITL only fires on fresh Station executions - cache hits are skipped.
Steps are numbered from 1 in execution order, matching the step counter
shown in \code{$run()} output. You can set multiple steps at once:
\code{wf$set_hitl(c(1, 3))}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$set_hitl(steps)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{steps}}{`[integerish]` One or more step numbers (>= 1).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Invisibly returns `self` for method chaining.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-clear_cache"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-clear_cache}{}}}
\subsection{Method \code{clear_cache()}}{
Remove all cached Station results.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$clear_cache()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Invisibly returns `self`.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-as_agent"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-as_agent}{}}}
\subsection{Method \code{as_agent()}}{
Wrap this Workflow as a `WorkflowAgent`.

The returned `WorkflowAgent` exposes \code{$invoke(prompt)} and holds an
`agent_id`, making it compatible with `LeadAgent$register_agents()` and
usable as a Station handler inside another `Workflow`.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$as_agent(name = NULL, instruction = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{`[character(1)]` Name for the `WorkflowAgent`. Defaults to
\code{"\{workflow name\} (agent)"}.}

\item{\code{instruction}}{`[character(1)]` Instruction string describing the
agent's role. Used by `LeadAgent` for task-agent matching.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A `WorkflowAgent` object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Workflow-visualize"></a>}}
\if{latex}{\out{\hypertarget{method-Workflow-visualize}{}}}
\subsection{Method \code{visualize()}}{
Render the workflow as a directed graph via DiagrammeR.

Stations are shown as rounded boxes. Conditional Routes are shown as
dashed arrows labelled "cond". The entry Station is marked with a filled
circle labelled "START".
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Workflow$visualize()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A `DiagrammeR` / `htmlwidget` object.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Agent.R
\name{Agent}
\alias{Agent}
\title{Agent: A General-Purpose LLM Agent}
\description{
The `Agent` class defines a modular LLM-based agent capable of responding to prompts using a defined role/instruction.
It wraps an OpenAI-compatible chat model via the [`ellmer`](https://github.com/llrs/ellmer) package.

Each agent maintains its own message history and unique identity.
}
\examples{

## ------------------------------------------------
## Method `Agent$new`
## ------------------------------------------------

  # An API KEY is required in order to invoke the Agent
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

  polar_bear_researcher <- Agent$new(
    name = "POLAR BEAR RESEARCHER",
    instruction = paste0(
    "You are an expert in polar bears, ",
    "you task is to collect information about polar bears. Answer in 1 sentence max."
    ),
    llm_object = openai_4_1_mini
  )


## ------------------------------------------------
## Method `Agent$invoke`
## ------------------------------------------------

\dontrun{
# An API KEY is required in order to invoke the Agent
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
)
agent <- Agent$new(
 name = "translator",
 instruction = "You are an Algerian citizen",
 llm_object = openai_4_1_mini
)
agent$invoke("Continue this sentence: 1 2 3 viva")
}

## ------------------------------------------------
## Method `Agent$generate_execute_r_code`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
r_assistant <- Agent$new(
  name = "R Code Assistant",
  instruction = paste("You are an expert R programmer",
  llm_object = openai_4_1_mini
)
# Generate code for data manipulation
result <- r_assistant$generate_execute_r_code(
  code_description = "Calculate the summary of the mtcars dataframe",
  validate = TRUE,
  execute = TRUE,
  interactive = TRUE
)
print(result)
}

## ------------------------------------------------
## Method `Agent$set_budget`
## ------------------------------------------------

\dontrun{
# An API KEY is required in order to invoke the Agent
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
)
agent <- Agent$new(
 name = "translator",
 instruction = "You are an Algerian citizen",
 llm_object = openai_4_1_mini
)
agent$set_budget(amount_in_usd = 10.5) # this is equivalent to 10.5$
}

## ------------------------------------------------
## Method `Agent$set_budget_policy`
## ------------------------------------------------

\dontrun{
agent$set_budget(5)
agent$set_budget_policy(on_exceed = "ask", warn_at = 0.9)
}

## ------------------------------------------------
## Method `Agent$keep_last_n_messages`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("What is the capital of Algeria")
agent$invoke("What is the capital of Germany")
agent$invoke("What is the capital of Italy")
agent$keep_last_n_messages(n = 2)
}

## ------------------------------------------------
## Method `Agent$add_message`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "AI assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$add_message("user", "Hello, how are you?")
agent$add_message("assistant", "I'm doing well, thank you!")
}

## ------------------------------------------------
## Method `Agent$share_context_with`
## ------------------------------------------------

\dontrun{
llm_object <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
a1 <- Agent$new("a1", "instr", llm_object)
a2 <- Agent$new("a2", "instr", llm_object)
a1$invoke("Hello")
a1$invoke("How are you?")
a1$share_context_with(a2, n = 2)
a2$messages
}

## ------------------------------------------------
## Method `Agent$clear_and_summarise_messages`
## ------------------------------------------------

\dontrun{
  # Requires an OpenAI-compatible LLM from `ellmer`
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

  agent <- Agent$new(
    name = "summariser",
    instruction = "You are a summarising assistant",
    llm_object = openai_4_1_mini
  )

  agent$invoke("The quick brown fox jumps over the lazy dog.")
  agent$invoke("This is another example sentence.")

  # Summarises and resets history
  agent$summarise_messages()

  # Now only the system prompt (with summary) remains
  agent$messages
}


## ------------------------------------------------
## Method `Agent$update_instruction`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$update_instruction("You are a concise assistant.")
}

## ------------------------------------------------
## Method `Agent$get_usage_stats`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$set_budget(1)
agent$invoke("What is the capital of Algeria?")
stats <- agent$get_usage_stats()
stats
}

## ------------------------------------------------
## Method `Agent$reset_conversation_history`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "AI assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("Hello, how are you?")
agent$invoke("Tell me about machine learning")
agent$reset_conversation_history()  # Clears all messages except system prompt
}

## ------------------------------------------------
## Method `Agent$export_messages_history`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital_finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("What is the capital of Algeria")
agent$invoke("What is the capital of Italy")
agent$export_messages_history()
}


## ------------------------------------------------
## Method `Agent$load_messages_history`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital_finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
# use the export_messages_history to save the interaction first
agent$load_messages_history("path/to/messages.json")
agent$messages
agent$llm_object
}


## ------------------------------------------------
## Method `Agent$validate_response`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "fact_checker",
  instruction = "You are a factual assistant.",
  llm_object = openai_4_1_mini
)
prompt <- "What is the capital of Algeria?"
response <- agent$invoke(prompt)
validation <- agent$validate_response(
  response = response,
  prompt = prompt,
  validation_criteria = "The response must be accurate and mention Algiers",
  validation_score = 0.8
)
print(validation)
}

## ------------------------------------------------
## Method `Agent$register_tools`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "file_assistant",
  instruction = "You are a file management assistant.",
  llm_object = openai_4_1_mini
)
# Register predefined tools
agent$register_tools(list(change_directory_tool(), list_files_tool()))
}

## ------------------------------------------------
## Method `Agent$list_tools`
## ------------------------------------------------

\dontrun{
agent$list_tools()
}

## ------------------------------------------------
## Method `Agent$remove_tools`
## ------------------------------------------------

\dontrun{
agent$remove_tools(c("change_directory", "list_files"))
}

## ------------------------------------------------
## Method `Agent$clear_tools`
## ------------------------------------------------

\dontrun{
agent$clear_tools()
}

## ------------------------------------------------
## Method `Agent$generate_and_register_tool`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "file_assistant",
  instruction = "You are a helpful assistant.",
  llm_object = openai_4_1_mini
)
# Generate and register a tool from description
agent$generate_and_register_tool(
  description = "Create a tool that saves text content to a file on disk"
)
# Now the tool is available to use
agent$list_tools()
}

## ------------------------------------------------
## Method `Agent$clone_agent`
## ------------------------------------------------

\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "translator",
  instruction = "You are a translator.",
  llm_object = openai_4_1_mini
)
# Clone with same name
agent_copy <- agent$clone_agent()
}
}
\seealso{
[load_messages_history()] for reloading a saved message history.

[export_messages_history()] for exporting the messages object to json.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{name}}{The agent's name.}

\item{\code{instruction}}{The agent's role/system prompt.}

\item{\code{llm_object}}{The underlying `ellmer::chat_openai` object.}

\item{\code{agent_id}}{A UUID uniquely identifying the agent.}

\item{\code{model_provider}}{The name of the entity providing the model (eg. OpenAI)}

\item{\code{model_name}}{The name of the model to be used (eg. gpt-4.1-mini)}

\item{\code{broadcast_history}}{A list of all past broadcast interactions.}

\item{\code{budget}}{A budget in $ that the agent should not exceed.}

\item{\code{budget_policy}}{A list controlling budget behavior: on_exceed and warn_at.}

\item{\code{budget_warned}}{Internal flag indicating whether warn_at notice was emitted.}

\item{\code{cost}}{The current cost of the agent}

\item{\code{tools}}{A list of registered tools available to the agent}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{messages}}{Public active binding for the conversation history.
Assignment is validated automatically.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Agent-new}{\code{Agent$new()}}
\item \href{#method-Agent-invoke}{\code{Agent$invoke()}}
\item \href{#method-Agent-generate_execute_r_code}{\code{Agent$generate_execute_r_code()}}
\item \href{#method-Agent-set_budget}{\code{Agent$set_budget()}}
\item \href{#method-Agent-set_budget_policy}{\code{Agent$set_budget_policy()}}
\item \href{#method-Agent-keep_last_n_messages}{\code{Agent$keep_last_n_messages()}}
\item \href{#method-Agent-add_message}{\code{Agent$add_message()}}
\item \href{#method-Agent-share_context_with}{\code{Agent$share_context_with()}}
\item \href{#method-Agent-clear_and_summarise_messages}{\code{Agent$clear_and_summarise_messages()}}
\item \href{#method-Agent-update_instruction}{\code{Agent$update_instruction()}}
\item \href{#method-Agent-get_usage_stats}{\code{Agent$get_usage_stats()}}
\item \href{#method-Agent-reset_conversation_history}{\code{Agent$reset_conversation_history()}}
\item \href{#method-Agent-export_messages_history}{\code{Agent$export_messages_history()}}
\item \href{#method-Agent-load_messages_history}{\code{Agent$load_messages_history()}}
\item \href{#method-Agent-validate_response}{\code{Agent$validate_response()}}
\item \href{#method-Agent-register_tools}{\code{Agent$register_tools()}}
\item \href{#method-Agent-list_tools}{\code{Agent$list_tools()}}
\item \href{#method-Agent-remove_tools}{\code{Agent$remove_tools()}}
\item \href{#method-Agent-clear_tools}{\code{Agent$clear_tools()}}
\item \href{#method-Agent-generate_and_register_tool}{\code{Agent$generate_and_register_tool()}}
\item \href{#method-Agent-clone_agent}{\code{Agent$clone_agent()}}
\item \href{#method-Agent-clone}{\code{Agent$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-new"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-new}{}}}
\subsection{Method \code{new()}}{
Initializes a new Agent with a specific role/instruction.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$new(name, instruction, llm_object, budget = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{A short identifier for the agent (e.g. `"translator"`).}

\item{\code{instruction}}{The system prompt that defines the agent's role.}

\item{\code{llm_object}}{The LLM object generate by ellmer (eg. output of ellmer::chat_openai)}

\item{\code{budget}}{Numerical value denoting the amount to set for the budget in US$ to a specific agent,
if the budget is reached, an error will be thrown.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{  # An API KEY is required in order to invoke the Agent
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

  polar_bear_researcher <- Agent$new(
    name = "POLAR BEAR RESEARCHER",
    instruction = paste0(
    "You are an expert in polar bears, ",
    "you task is to collect information about polar bears. Answer in 1 sentence max."
    ),
    llm_object = openai_4_1_mini
  )

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-invoke"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-invoke}{}}}
\subsection{Method \code{invoke()}}{
Sends a user prompt to the agent and returns the assistant's response.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$invoke(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{A character string prompt for the agent to respond to.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The LLM-generated response as a character string.
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
# An API KEY is required in order to invoke the Agent
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
)
agent <- Agent$new(
 name = "translator",
 instruction = "You are an Algerian citizen",
 llm_object = openai_4_1_mini
)
agent$invoke("Continue this sentence: 1 2 3 viva")
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-generate_execute_r_code"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-generate_execute_r_code}{}}}
\subsection{Method \code{generate_execute_r_code()}}{
Generate R code from natural language descriptions and optionally validate/execute it
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$generate_execute_r_code(
  code_description,
  validate = FALSE,
  execute = FALSE,
  interactive = TRUE,
  env = globalenv()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{code_description}}{Character string describing the R code to generate}

\item{\code{validate}}{Logical indicating whether to validate the generated code syntax}

\item{\code{execute}}{Logical indicating whether to execute the generated code (use with caution)}

\item{\code{interactive}}{Logical; if TRUE, ask for user confirmation before executing generated code}

\item{\code{env}}{Environment in which to execute the code if execute = TRUE. Default to \code{globalenv}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing the generated code and validation/execution results
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
r_assistant <- Agent$new(
  name = "R Code Assistant",
  instruction = paste("You are an expert R programmer",
  llm_object = openai_4_1_mini
)
# Generate code for data manipulation
result <- r_assistant$generate_execute_r_code(
  code_description = "Calculate the summary of the mtcars dataframe",
  validate = TRUE,
  execute = TRUE,
  interactive = TRUE
)
print(result)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-set_budget"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-set_budget}{}}}
\subsection{Method \code{set_budget()}}{
Set a budget to a specific agent, if the budget is reached, an error will be thrown
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$set_budget(amount_in_usd)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{amount_in_usd}}{Numerical value denoting the amount to set for the budget,}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
# An API KEY is required in order to invoke the Agent
openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
)
agent <- Agent$new(
 name = "translator",
 instruction = "You are an Algerian citizen",
 llm_object = openai_4_1_mini
)
agent$set_budget(amount_in_usd = 10.5) # this is equivalent to 10.5$
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-set_budget_policy"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-set_budget_policy}{}}}
\subsection{Method \code{set_budget_policy()}}{
Configure how the agent behaves as it approaches or exceeds its budget.
Use `warn_at` (0-1) to emit a one-time warning when spending reaches the
specified fraction of the budget. When the budget is exceeded, `on_exceed`
controls behavior: abort, warn and proceed, or ask interactively.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$set_budget_policy(on_exceed = "abort", warn_at = 0.8)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{on_exceed}}{One of "abort", "warn", or "ask".}

\item{\code{warn_at}}{Numeric in (0,1); fraction of budget to warn at. Default 0.8.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
agent$set_budget(5)
agent$set_budget_policy(on_exceed = "ask", warn_at = 0.9)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-keep_last_n_messages"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-keep_last_n_messages}{}}}
\subsection{Method \code{keep_last_n_messages()}}{
Keep only the most recent `n` messages, discarding older ones while keeping
the system prompt.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$keep_last_n_messages(n = 2)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{n}}{Number of most recent messages to keep.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("What is the capital of Algeria")
agent$invoke("What is the capital of Germany")
agent$invoke("What is the capital of Italy")
agent$keep_last_n_messages(n = 2)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-add_message"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-add_message}{}}}
\subsection{Method \code{add_message()}}{
Add a pre-formatted message to the conversation history
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$add_message(role, content)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{role}}{The role of the message ("user", "assistant", or "system")}

\item{\code{content}}{The content of the message}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "AI assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$add_message("user", "Hello, how are you?")
agent$add_message("assistant", "I'm doing well, thank you!")
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-share_context_with"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-share_context_with}{}}}
\subsection{Method \code{share_context_with()}}{
Share the most recent conversation messages with another agent.

The last \code{n} non-system messages from \code{self} are appended to the
history of an other \code{agent}. This is useful when
one agent needs to handoff context or continue a thread in a separate
agent.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$share_context_with(agent, n = 5)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{agent}}{An object of class \code{Agent} that will receive the
messages.}

\item{\code{n}}{Number of recent messages to transfer (excluding the system prompt).
Defaults to \code{5}.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
llm_object <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
a1 <- Agent$new("a1", "instr", llm_object)
a2 <- Agent$new("a2", "instr", llm_object)
a1$invoke("Hello")
a1$invoke("How are you?")
a1$share_context_with(a2, n = 2)
a2$messages
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-clear_and_summarise_messages"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-clear_and_summarise_messages}{}}}
\subsection{Method \code{clear_and_summarise_messages()}}{
Summarises the agent's conversation history into a concise form and appends it
to the system prompt. Unlike `update_instruction()`, this method does not override
the existing instruction but augments it with a summary for future context.

After creating the summary, the method clears the conversation history and
retains only the updated system prompt. This ensures that subsequent interactions
start fresh but with the summary preserved as context.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$clear_and_summarise_messages()}\if{html}{\out{</div>}}
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
  # Requires an OpenAI-compatible LLM from `ellmer`
  openai_4_1_mini <- ellmer::chat(
    name = "openai/gpt-4.1-mini",
    api_key = Sys.getenv("OPENAI_API_KEY"),
    echo = "none"
  )

  agent <- Agent$new(
    name = "summariser",
    instruction = "You are a summarising assistant",
    llm_object = openai_4_1_mini
  )

  agent$invoke("The quick brown fox jumps over the lazy dog.")
  agent$invoke("This is another example sentence.")

  # Summarises and resets history
  agent$summarise_messages()

  # Now only the system prompt (with summary) remains
  agent$messages
}

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-update_instruction"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-update_instruction}{}}}
\subsection{Method \code{update_instruction()}}{
Update the system prompt/instruction
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$update_instruction(new_instruction)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{new_instruction}}{New instruction to use. Not that the new instruction
will override the old one}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$update_instruction("You are a concise assistant.")
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-get_usage_stats"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-get_usage_stats}{}}}
\subsection{Method \code{get_usage_stats()}}{
Get the current token count and estimated cost of the conversation
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$get_usage_stats()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A list with token counts and cost information
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$set_budget(1)
agent$invoke("What is the capital of Algeria?")
stats <- agent$get_usage_stats()
stats
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-reset_conversation_history"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-reset_conversation_history}{}}}
\subsection{Method \code{reset_conversation_history()}}{
Reset the agent's conversation history while keeping the system instruction
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$reset_conversation_history()}\if{html}{\out{</div>}}
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "AI assistant",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("Hello, how are you?")
agent$invoke("Tell me about machine learning")
agent$reset_conversation_history()  # Clears all messages except system prompt
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-export_messages_history"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-export_messages_history}{}}}
\subsection{Method \code{export_messages_history()}}{
Saves the agent's current conversation history as a JSON file on disk.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$export_messages_history(
  file_path = paste0(getwd(), "/", paste0(self$name, "_messages.json"))
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file_path}}{Character string specifying the file path where the JSON
file should be saved. Defaults to a file named
`"<agent_name>_messages.json"` in the current working directory.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital_finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
agent$invoke("What is the capital of Algeria")
agent$invoke("What is the capital of Italy")
agent$export_messages_history()
}

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-load_messages_history"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-load_messages_history}{}}}
\subsection{Method \code{load_messages_history()}}{
Load an agent's conversation history as a JSON file from disk.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$load_messages_history(
  file_path = paste0(getwd(), "/", paste0(self$name, "_messages.json"))
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file_path}}{Character string specifying the file path where the JSON
file is stored. Defaults to a file named
`"<agent_name>_messages.json"` in the current working directory.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "capital_finder",
  instruction = "You are an assistant.",
  llm_object = openai_4_1_mini
)
# use the export_messages_history to save the interaction first
agent$load_messages_history("path/to/messages.json")
agent$messages
agent$llm_object
}

}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-validate_response"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-validate_response}{}}}
\subsection{Method \code{validate_response()}}{
Validates an agent's response against custom criteria using LLM-based validation.
This method uses the agent's LLM to evaluate whether a response meets specified
validation criteria (e.g., accuracy, completeness, tone, format).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$validate_response(
  prompt,
  response,
  validation_criteria,
  validation_score = 0.8
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{The prompt used to generate the response.}

\item{\code{response}}{The response text to validate.}

\item{\code{validation_criteria}}{The criteria for validation.
(e.g., "The response should be accurate and complete", "The response must be under 100 words").}

\item{\code{validation_score}}{A numeric from 0 to 1 denoting the score to consider for
the evaluation. During the evaluation, the agent will provide a score from 0 to 1,
a provided score greater or equal to the `validation_score` will result in a `valid` response.
Defaults to 0.8.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
list object
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "fact_checker",
  instruction = "You are a factual assistant.",
  llm_object = openai_4_1_mini
)
prompt <- "What is the capital of Algeria?"
response <- agent$invoke(prompt)
validation <- agent$validate_response(
  response = response,
  prompt = prompt,
  validation_criteria = "The response must be accurate and mention Algiers",
  validation_score = 0.8
)
print(validation)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-register_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-register_tools}{}}}
\subsection{Method \code{register_tools()}}{
Register one or more tools with the agent
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$register_tools(tools)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tools}}{A list of ellmer tool objects or a single tool object}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "file_assistant",
  instruction = "You are a file management assistant.",
  llm_object = openai_4_1_mini
)
# Register predefined tools
agent$register_tools(list(change_directory_tool(), list_files_tool()))
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-list_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-list_tools}{}}}
\subsection{Method \code{list_tools()}}{
List all registered tools
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$list_tools()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character vector of tool names
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
agent$list_tools()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-remove_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-remove_tools}{}}}
\subsection{Method \code{remove_tools()}}{
Remove tools from the agent
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$remove_tools(tool_names)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{tool_names}}{Character vector of tool names to remove}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
agent$remove_tools(c("change_directory", "list_files"))
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-clear_tools"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-clear_tools}{}}}
\subsection{Method \code{clear_tools()}}{
Clear all registered tools
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$clear_tools()}\if{html}{\out{</div>}}
}

\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
agent$clear_tools()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-generate_and_register_tool"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-generate_and_register_tool}{}}}
\subsection{Method \code{generate_and_register_tool()}}{
Generate and register a tool from a natural language description.
This method uses the agent's LLM to create both the R function and the
ellmer tool definition based on your description, then automatically registers it.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$generate_and_register_tool(description)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{description}}{Character string describing what the tool should do}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "file_assistant",
  instruction = "You are a helpful assistant.",
  llm_object = openai_4_1_mini
)
# Generate and register a tool from description
agent$generate_and_register_tool(
  description = "Create a tool that saves text content to a file on disk"
)
# Now the tool is available to use
agent$list_tools()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-clone_agent"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-clone_agent}{}}}
\subsection{Method \code{clone_agent()}}{
Create a copy of the agent with the same instruction and configuration but a new unique ID.
Useful for creating multiple instances of the same agent type.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$clone_agent(new_name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{new_name}}{Optional character string to assign a new name to the cloned agent.
If NULL, the cloned agent retains the original name.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
openai_4_1_mini <- ellmer::chat(
  name = "openai/gpt-4.1-mini",
  api_key = Sys.getenv("OPENAI_API_KEY"),
  echo = "none"
)
agent <- Agent$new(
  name = "translator",
  instruction = "You are a translator.",
  llm_object = openai_4_1_mini
)
# Clone with same name
agent_copy <- agent$clone_agent()
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Agent-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Agent-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Agent$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Workflow.R
\name{WorkflowAgent}
\alias{WorkflowAgent}
\title{WorkflowAgent}
\description{
A lightweight, Agent-compatible wrapper around a `Workflow`. Created via
\code{Workflow$as_agent()}.

`WorkflowAgent` exposes the same interface that `LeadAgent` expects from
any agent (\code{name}, \code{instruction}, \code{agent_id},
\code{$invoke()}), so it can be:
\itemize{
  \item Registered with \code{LeadAgent$register_agents()}.
  \item Used as a Station handler inside another `Workflow`.
}

Do not instantiate `WorkflowAgent` directly - use \code{Workflow$as_agent()}.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{name}}{Agent name (visible to `LeadAgent` for task matching).}

\item{\code{instruction}}{System instruction describing the agent's role.}

\item{\code{agent_id}}{Unique identifier (UUID).}

\item{\code{workflow}}{The underlying `Workflow` object.}

\item{\code{messages}}{Conversation history as a list of
\code{list(role, content)} entries.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-WorkflowAgent-new}{\code{WorkflowAgent$new()}}
\item \href{#method-WorkflowAgent-invoke}{\code{WorkflowAgent$invoke()}}
\item \href{#method-WorkflowAgent-reset_conversation_history}{\code{WorkflowAgent$reset_conversation_history()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkflowAgent-new"></a>}}
\if{latex}{\out{\hypertarget{method-WorkflowAgent-new}{}}}
\subsection{Method \code{new()}}{
Create a new `WorkflowAgent`. Prefer \code{Workflow$as_agent()}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkflowAgent$new(name, instruction, workflow)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{`[character(1)]` Agent name.}

\item{\code{instruction}}{`[character(1)]` Instruction / system prompt.}

\item{\code{workflow}}{A `Workflow` object.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkflowAgent-invoke"></a>}}
\if{latex}{\out{\hypertarget{method-WorkflowAgent-invoke}{}}}
\subsection{Method \code{invoke()}}{
Run the underlying workflow with `prompt` as input.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkflowAgent$invoke(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{`[character(1)]` The user prompt / input payload.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
`[character(1)]` The final output of the workflow.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkflowAgent-reset_conversation_history"></a>}}
\if{latex}{\out{\hypertarget{method-WorkflowAgent-reset_conversation_history}{}}}
\subsection{Method \code{reset_conversation_history()}}{
Clear the conversation history stored in \code{$messages}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkflowAgent$reset_conversation_history()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Invisibly returns `self`.
}
}
}
